;;; `Lousy' --- Kana Kanji front end for emacs/mule

;; Normal people do not usee lkk.  Use skk instead.

;;
;; Authour: Seiichi Namba <sn@asahi-net.email.ne.jp>
;; http://www.asahi-net.or.jp/~pi9s-nnb/
;;
;; `Lousy' is FleaSoftware. Copy freely, and of course there is no warranty.
;;

(provide 'lkk)

;; Include this sort of codes in ~/.emacs to use lkk: 
;; (define-key global-map "\C-xJ" 'lkk-mode)
;; (autoload 'lkk-mode "lkk" nil t)
;;
;; Setup for windows.el/revive.el, to keep dict / work buffer for wipe/restore.
;; (setq revive:ignore-buffer-pattern "^ \\*.skk\\|^ \\*.*lkk")

;; Version 0
;; Wed Feb  2 03:47:53 2000 Dict->kanji converter primitives
;; Fri Feb  4 21:26:22 2000 Romaji-Kana input started.
;; Sat Feb  5 10:10:07 2000 Refine Kana input.
;; Wed Feb  9 20:51:13 2000 Roma-kana input optimized with hashed vector tables.
;; Sat Feb 12 16:08:01 2000 lkk-l-mode, lkk-L-mode, as minor mode.
;; Sun Feb 20 15:57:29 2000 lkk-J-mode (conversion) routines started.

(defconst lkk-version "0")
(defun lkk-version ()
  (interactive)
  (if (interactive-p)
      (message "Lousy version %s" lkk-version)
    lkk-version))

(setq enable-recursive-minibuffers t)

(or (fboundp 'defsubst)	; for emacs-18 ? (but I don't have it anymore...)
    (fset 'defsubst 'defun))
(or (fboundp 'buffer-substring-no-properties)
    (fset 'buffer-substring-no-properties 'buffer-substring))

;; Buffer local variable definition macro
;; We *NEVER* call this with non-nil default-value (to make things simple).
;; Skelton from viper-util.el as in skk-10, modified to make what it does clear.
;; Viper-util's deflocalvar is not so bad if programmer knows it is equivalent to
;; (progn (setq-default var val)
;;        (defvar var val "xxx")
;;        (make-variable-buffer-local))
;; though... See also skk-set-marker comments in skk-foreword.el (skk10.49).
(defmacro lkk-deflocalvar (var default-value &optional documentation)
  (` (progn
       (make-variable-buffer-local '(, var)) ; same effect (setq-default var nil)
       (defvar (, var) nil
	       (, (format "%s\n\(buffer local\)" documentation)))
       (if (, default-value)
	   (setq-default (, var) (, default-value)))
     )))

;;;
;;; Variables, Primitive Tables (user customizable maybe, or maybe not)
;;;

;; Cursor colors. No facility to detect background. Specify directly.
(defvar lkk-j-mode-cursor-color "dodgerblue")
(defvar lkk-J-mode-cursor-color lkk-j-mode-cursor-color)
(defvar lkk-L-mode-cursor-color "gold")
(defvar lkk-l-mode-cursor-color "slateblue2")
(defvar lkk-q-mode-cursor-color "palegreen")

(defvar lkk-dict "~/.lkk-jisyo" "lkk jisyo filename")
(defvar lkk-dict-buffer-name " *lkk-jisyo*" "lkk dictionary buffer name")

;; Some tables

(defvar lkk-r2k-alist nil
  "Romaji -- Kana conversion table.  Setup online.")
(defvar lkk-r2k-alist-C
  '(("cha" . ("" . "")) ("chi" . ("" . ""))
    ("chu" . ("" . "")) ("che" . ("" . ""))
    ("cho" . ("" . ""))

    ("sha" . ("" . "")) ("shi" . ("" . ""))
    ("shu" . ("" . "")) ("she" . ("" . ""))
    ("sho" . ("" . "")) ("sha" . ("" . ""))
    ("shi" . ("" . ""))     ("shu" . ("" . ""))
    ("she" . ("" . "")) ("sho" . ("" . ""))

    ("kya" . ("" . "")) ("kyu" . ("" . ""))
    ("kyo" . ("" . ""))

    ("sya" . ("" . "")) ("syu" . ("" . ""))
    ("syo" . ("" . "")) ("tya" . ("" . ""))

    ("tyu" . ("" . "")) ("tyo" . ("" . ""))

    ("nya" . ("ˤ" . "˥")) ("nyu" . ("ˤ" . "˥"))

    ("nyo" . ("ˤ" . "˥")) ("hya" . ("Ҥ" . "ҥ"))

    ("hyu" . ("Ҥ" . "ҥ")) ("hyo" . ("Ҥ" . "ҥ"))

    ("mya" . ("ߤ" . "ߥ")) ("myu" . ("ߤ" . "ߥ"))
    ("myo" . ("ߤ" . "ߥ")) ("rya" . ("" . ""))

    ("ryu" . ("" . "")) ("ryo" . ("" . ""))

    ("gya" . ("" . "")) ("gyu" . ("" . ""))
    ("gyo" . ("" . ""))

    ("zya" . ("" . "")) ("zyu" . ("" . ""))
    ("zyo" . ("" . ""))

    ("dya" . ("¤" . "¥")) ("dyu" . ("¤" . "¥"))
    ("dyo" . ("¤" . "¥"))

    ("bya" . ("Ӥ" . "ӥ")) ("byu" . ("Ӥ" . "ӥ"))
    ("byo" . ("Ӥ" . "ӥ"))

    ("pya" . ("Ԥ" . "ԥ")) ("pyu" . ("Ԥ" . "ԥ"))
    ("pyo" . ("Ԥ" . "ԥ"))

    ("jya" . ("" . "")) ("jyu" . ("" . ""))
    ("jyo" . ("" . ""))

    ("tsu" . ("" . ""))

    ("mm" . ("" . ""))
    ("nn" . ("" . ""))
    ("m'" . ("" . ""))
    ("n'" . ("" . ""))

    ("fa" . ("դ" . "ե")) ("fi" . ("դ" . "ե"))
    ("fu" . ("" . ""))     ("fe" . ("դ" . "ե"))
    ("fo" . ("դ" . "ե"))

    ("ka" . ("" . "")) ("ki" . ("" . ""))
    ("ku" . ("" . "")) ("ke" . ("" . ""))
    ("ko" . ("" . ""))
    ("sa" . ("" . "")) ("si" . ("" . ""))
    ("su" . ("" . "")) ("se" . ("" . ""))
    ("so" . ("" . ""))
    ("ta" . ("" . "")) ("ti" . ("" . ""))
    ("tu" . ("" . "")) ("te" . ("" . ""))
    ("to" . ("" . ""))
    ("na" . ("" . "")) ("ni" . ("" . ""))
    ("nu" . ("" . "")) ("ne" . ("" . ""))
    ("no" . ("" . ""))
    ("ha" . ("" . "")) ("hi" . ("" . ""))
    ("hu" . ("" . "")) ("he" . ("" . ""))
    ("ho" . ("" . ""))
    ("ma" . ("" . "")) ("mi" . ("" . ""))
    ("mu" . ("" . "")) ("me" . ("" . ""))
    ("mo" . ("" . ""))
    ("ya" . ("" . "")) ("yu" . ("" . ""))
    ("yo" . ("" . ""))
    ("ra" . ("" . "")) ("ri" . ("" . ""))
    ("ru" . ("" . "")) ("re" . ("" . ""))
    ("ro" . ("" . ""))
    ("wa" . ("" . "")) ("wi" . ("" . ""))
    ("we" . ("" . "")) ("wo" . ("" . ""))

    ("ga" . ("" . "")) ("gi" . ("" . ""))
    ("gu" . ("" . "")) ("ge" . ("" . ""))
    ("go" . ("" . ""))
    ("za" . ("" . "")) ("zi" . ("" . ""))
    ("zu" . ("" . "")) ("ze" . ("" . ""))
    ("zo" . ("" . ""))
    ("da" . ("" . "")) ("di" . ("" . ""))
    ("du" . ("" . "")) ("de" . ("" . ""))
    ("do" . ("" . ""))
    ("ba" . ("" . "")) ("bi" . ("" . ""))
    ("bu" . ("" . "")) ("be" . ("" . ""))
    ("bo" . ("" . ""))
    ("pa" . ("" . "")) ("pi" . ("" . ""))
    ("pu" . ("" . "")) ("pe" . ("" . ""))
    ("po" . ("" . ""))

    ("ja" . ("" . "")) ("ji" . ("" . ""))
    ("ju" . ("" . "")) ("je" . ("" . ""))
    ("jo" . ("" . ""))

    ("va" . ("" . "")) ("vi" . ("" . ""))
    ("vu" . ("" . "")) ("ve" . ("" . ""))
    ("vo" . ("" . ""))

    ("xa" . ("" . ""))
    ("xka" . ("" . ""))
    ("xwa" . ("" . ""))
    ("xya" . ("" . ""))
    ("xi" . ("" . ""))
    ("xwi" . ("" . ""))
    ("xu" . ("" . ""))
    ("xtu" . ("" . ""))
    ("xtsu" . ("" . ""))
    ("xtu" . ("" . ""))
    ("xyu" . ("" . ""))
    ("xe" . ("" . ""))
    ("xke" . ("" . ""))
    ("xwe" . ("" . ""))
    ("xo" . ("" . ""))
    ("xyo" . ("" . ""))
    )
  "Romaji -- Kana basic table 1 -- all yomi beginning from consonant.
If you are modified this alist, eval:
 \(lkk-setup-r2k-internal\)
to reflect the changes.")

(defconst lkk-r2k-alist-V
  '(("a" . ("" . ""))
    ("i" . ("" . ""))
    ("u" . ("" . ""))
    ("e" . ("" . ""))
    ("o" . ("" . "")))
  "Romaji -- Kana basic table 0 -- yomi includes vowels only.")

(defconst lkk-input-vector
  [nil  nil  nil  nil  nil  nil  nil  nil
   nil  nil  nil  nil  nil  nil  nil  nil 
   nil  nil  nil  nil  nil  nil  nil  nil 
   nil  nil  nil  nil  nil  nil  nil  nil 
   nil  "" nil  nil  nil  nil  nil  nil
   nil  nil  nil  nil  "" "" "" nil
   nil  nil  nil  nil  nil  nil  nil  nil 
   nil  nil  "" "" nil  nil  nil  ""
   nil  nil  nil  nil  nil  nil  nil  nil 
   nil  nil  nil  nil  nil  nil  nil  nil 
   nil  nil  nil  nil  nil  nil  nil  nil 
   nil  nil  nil  "" nil  "" nil  nil
   nil  nil  nil  nil  nil  nil  nil  nil 
   nil  nil  nil  nil  nil  nil  nil  nil 
   nil  nil  nil  nil  nil  nil  nil  nil 
   nil  nil  nil  nil  nil  nil  nil  nil]
  "Characters to be inserted directly in hiragana mode.")

;;(lkk-dump-chars ?  (1+ (- 127 ? )) ?\x20 16)
;;; (defconst lkk-ascii-vector
;;;   [nil  nil  nil  nil  nil  nil  nil  nil
;;;    nil  nil  nil  nil  nil  nil  nil  nil 
;;;    nil  nil  nil  nil  nil  nil  nil  nil 
;;;    nil  nil  nil  nil  nil  nil  nil  nil 
;;;    " "  "!"  "\"" "#"  "$"  "%"  "&"  "'"
;;;    "("  ")"  "*"  "+"  ","  "-"  "."  "/"
;;;    "0"  "1"  "2"  "3"  "4"  "5"  "6"  "7" 
;;;    "8"  "9"  ":"  ";"  "<"  "="  ">"  "?" 
;;;    "@"  "A"  "B"  "C"  "D"  "E"  "F"  "G" 
;;;    "H"  "I"  "J"  "K"  "L"  "M"  "N"  "O" 
;;;    "P"  "Q"  "R"  "S"  "T"  "U"  "V"  "W" 
;;;    "X"  "Y"  "Z"  "["  "\\"  "]"  "^"  "_" 
;;;    "\`"  "a"  "b"  "c"  "d"  "e"  "f"  "g" 
;;;    "h"  "i"  "j"  "k"  "l"  "m"  "n"  "o" 
;;;    "p"  "q"  "r"  "s"  "t"  "u"  "v"  "w" 
;;;    "x"  "y"  "z"  "{"  "|"  "}"  "~"  nil]
;;;   "Ascii characters -- mainly for `l' mode.")

(defconst lkk-zenkaku-vector
  [nil  nil  nil  nil  nil  nil  nil  nil
   nil  nil  nil  nil  nil  nil  nil  nil 
   nil  nil  nil  nil  nil  nil  nil  nil 
   nil  nil  nil  nil  nil  nil  nil  nil 
   ""  "" "" "" "" "" "" ""
   "" "" "" "" "" "" "" ""
   "" "" "" "" "" "" "" ""
   "" "" "" "" "" "" "" ""
   "" "" "" "" "" "" "" ""
   "" "" "" "" "" "" "" ""
   "" "" "" "" "" "" "" ""
   "" "" "" "" "" "" "" ""
   "" "" "" "" "" "" "" ""
   "" "" "" "" "" "" "" ""
   "" "" "" "" "" "" "" ""
   "" "" "" "" "" "" "" nil]
  "Zenkaku characters -- mainly for `L' mode.")

;;;
;;; Table build and Utilities referring the tables
;;;

;;
;; Romaji -- Kana Table build, and
;; Romaji -- Kana conversion handling subroutines
;;
(defvar lkk-r2k-terminator-list nil
  "List of possible string of romaji yomi terminator. Internal use only.")
(defvar lkk-valid-consonant-list nil
  "List of possible consonant part of romaji yomi. Internal use only.")
(defvar lkk-valid-consonant-vector nil
  "Vector of valid consonant list hashed by 1st char. Internal use only.")
(defvar lkk-r2k-alist-vector nil
  "Vector of romaji-kana alist hashed by 1st char of yomi. Internal use only.")

(defun lkk-build-valid-consonant-list (alis)
  "Return all consnant part of romaji yomi as uniqued list."
  (let ((uniqued nil))
    (mapcar (lambda (c) ; Uniqueing squeezes the list down to 1/3 to 1/4.
	      (if (and (null (member c uniqued))
		       (> (length c) 0))
		  (setq uniqued (cons c uniqued))))
	    (mapcar (lambda (c) (substring c 0 -1)) ; all consonant part
		    (mapcar (lambda (c) (car c)) ; all yomi
			    alis)))
    uniqued))

(defun lkk-build-valid-consonant-vector (xlis)
  "Return vector of valid consonant lists built out of XLIST.
The vector is hashed by 1st char."
  (let ((vec (make-vector 128 nil))	; Always start with fresh table.
	ch tem)
    (mapcar
     (lambda (consonant)
       (setq ch (string-to-char (substring consonant 0 1))
	     tem (aref vec ch))
       (aset vec ch (append tem (list consonant))) )
     xlis)
    vec))

(defun lkk-build-r2k-alist-vector (xlis)
  "Return a hashed vector of alists built from XLIST (alist maybe).
The returned vector contains alists hashed by 1st char."
  (let ((vec (make-vector 128 nil))	; Always start with fresh table.
	ch tem)
    (mapcar
     (lambda (cell)
       (setq ch (string-to-char (substring (car cell) 0 1))
	     tem (aref vec ch))
       (aset vec ch (append tem (list cell))) )
     xlis)
    vec))

(defun lkk-build-r2k-terminator-list (&optional table)
  (let (tem lastchar (table (or table lkk-r2k-alist-C)))
    (mapcar (lambda (c)
	      (setq lastchar (substring (car c) -1))
	      (or (member lastchar tem)
		  (setq tem (cons lastchar tem))))
	    table)
    (nreverse tem)))

(defun lkk-setup-r2k-internal ()
  "Setup lkk's internal alist/list/vectors for Romaji conversion."
  ;; setq order is very significant.
  (setq lkk-r2k-alist
	(append 
	 lkk-r2k-alist-C
	 lkk-r2k-alist-V))
  (setq lkk-r2k-terminator-list
	(lkk-build-r2k-terminator-list lkk-r2k-alist))
  (setq lkk-valid-consonant-list
	(lkk-build-valid-consonant-list lkk-r2k-alist))
  (setq lkk-valid-consonant-vector
	(lkk-build-valid-consonant-vector lkk-valid-consonant-list))
  (setq lkk-r2k-alist-vector
	(lkk-build-r2k-alist-vector lkk-r2k-alist)))

;; Setup roma -- kana table on loading this program.
(lkk-setup-r2k-internal)
;;(length lkk-r2k-alist)

(defsubst lkk-r2k-terminator-p (str)
  "Return non-nil if STR is member of lkk-r2k-terminator-list."
  (member str lkk-r2k-terminator-list))

(defsubst lkk-valid-roma-p0 (str &optional lis)
  "Predicate if STR is head part of valid Romaji yomi.
Optional 2nd arg LIST is used as list of valid strings, or
lkk-valid-consonant-list directly used if optional 2nd arg is nil.
This function is not so fast. Used by faster one, so don't purge this function."
  (let ((xlis (or lis lkk-valid-consonant-list))
	(ret nil))
    (mapcar (lambda (c)
	      (if (string-match str (concat "^" c))
		  (setq ret t))) xlis)
    ret))

(defsubst lkk-valid-roma-p (str)
  "Predicate if STR is head part of valid Romaji yomi."
  (let* ((table (aref lkk-valid-consonant-vector
		      (string-to-char (substring str 0 1)))))
    (if table (lkk-valid-roma-p0 str table) nil) ))

(defsubst lkk-get-kana0 (str &optional alis)
  "Return kana corresponding to romaji in STR.  Slow primitive subroutine."
  (let ((cell (assoc str (or alis lkk-r2k-alist))))
    (if cell 
	(if lkk-q-mode
	    (cdr (cdr cell))
	  (car (cdr cell)))
      nil)))

(defsubst lkk-get-kana (str)
  "Return kana corresponding to romaji in STR."
  (let ((table (aref lkk-r2k-alist-vector
		      (string-to-char (substring str 0 1)))))
    (if table (lkk-get-kana0 str table) nil) ))

;;
;; ascii case, h2k, k2h conversion.  Very lousy.
;;

(defsubst lkk-tolower (x)
  (cond ((stringp x) 
	 (char-to-string (+ (string-to-char x) 32))) ; (- ?a ?A) 32
	(t (+ x 32))			; else x must be char
	 ))

(defsubst lkk-toupper (x)
  (cond ((stringp x) 
	 (char-to-string (- (string-to-char x) 32))) ; (- ?a ?A) 32
	(t (- x 32))			; else x must be char
	))

(defconst lkk-j-bytes (length ""))
;;(defconst lkk-j-bytes (char-bytes (string-to-char "")))
;;(defconst lkk-j-bytes
;;  (let (p) (with-temp-buffer
;;	    (erase-buffer) (insert lkk-insignia0  "")
;;	    (goto-char 1)  (forward-char 1) (- (point) 1))))

;; sref() is alias of aref in emacs-20.xx, while
;; sref() is special and different from aref() in mule-2.3@emacs-19.xx
;; char-bytes() still exists in emacs-20.xx, and always returns 1.
;; See mule-util.el:
;; string-to-char-list() is in mule-util.el of 19.xx (poem.el has it too).
;; string-to-list() is in mule-util.el of 20.xx.

(defun lkk-string-to-char-list (str)
  ;; Code is stolen from mule-util.el in mule-2.3@emacs-19.34.
  (let ((len (length str))
	(idx 0)
	c l)
    (while (< idx len)
      (setq c (sref str idx))		; Must be sref in emacs-19.xx
      (setq idx (+ idx (char-bytes c)))
      (setq l (cons c l)))
    (nreverse l)))

;;; Unused. Works only on emacs-20.
;;;(defsubst lkk-string-to-char-list (str)
;;;  "Hopefully equivalent to string-to-char-list in mule-2, and emacs-20.x."
;;;  (mapcar (lambda (c) c) str))
;;;(defsubst lkk-string-to-char (str)
;;;  "Hopefully equivalent to string-to-char in mule-2, and emacs-20.x."
;;;  (car (lkk-string-to-char-list str)))

(defvar lkk-h2k-alist nil "Alist for hiragana to katakana conversion.")
(defvar lkk-k2h-alist nil "Alist for katakana to hiragana conversion.")

;; Lousy ! "" can't be supported.
(defun lkk-build-h2k-k2h-alist ()
  (setq lkk-h2k-alist nil lkk-k2h-alist nil) ; reset
  (let ((hc ?) (kc ?) (hs) (ks))
    (while (>= ? hc)
      (setq hs (char-to-string hc)
	    ks (char-to-string kc)
	    lkk-h2k-alist
	    (append lkk-h2k-alist
		    (list (cons hs ks)))
	    lkk-k2h-alist
	    (append lkk-k2h-alist
		    (list (cons ks hs))) )
      (setq hc (1+ hc)
	    kc (1+ kc)) )))

(lkk-build-h2k-k2h-alist)

;; Lousy ! "" cann't be supported.
(defun lkk-h2k (hstr)
  "Convert HIRAGANA string to katakana."
  (mapconcat
   (lambda (c) 
     (let ((tem (assoc (char-to-string c) lkk-h2k-alist)))
       (if tem (cdr tem) (char-to-string c))))
   (lkk-string-to-char-list hstr) ""))

(defun lkk-k2h (kstr)
  "Convert KATAKANA string to hiragana."
  (mapconcat
   (lambda (c)
     (let ((tem (assoc (char-to-string c) lkk-k2h-alist)))
       (if tem (cdr tem) (char-to-string c))))
   (lkk-string-to-char-list kstr) ""))

(defun lkk-h2k-region (rb re)
  (interactive "*r")
  (let* ((str (buffer-substring-no-properties rb re))
	 (result (lkk-h2k str)))
    (goto-char rb)
    (delete-region rb re)
    (insert result)))
(fset 'lkk-katakana-region 'lkk-h2k-region)

(defun lkk-k2h-region (rb re)
  (interactive "*r")
  (let* ((str (buffer-substring-no-properties rb re))
	 (result (lkk-k2h str)))
    (goto-char rb)
    (delete-region rb re)
    (insert result)))
(fset 'lkk-hiragana-region 'lkk-k2h-region)


;;;
;;; LKK Automaton status variables (includes minor-mode variable definitions)
;;;

;; For debug use. Guess it's right to make it buffer-local.
(lkk-deflocalvar lkk-debug-hook nil "lkk debugger.")

(defconst lkk-valid-minor-minor-modes '(lkk-j-mode
					lkk-J-mode
					lkk-l-mode
					lkk-L-mode
					lkk-q-mode)
  "List of valid submode variable symbols.")

(lkk-deflocalvar lkk-j-mode nil "`j' mode. Kana mode.")
(lkk-deflocalvar lkk-J-mode nil "`J' mode. Turned on while kanji conversion.")
(lkk-deflocalvar lkk-l-mode nil "`l' mode. Direct ascii mode.")
(lkk-deflocalvar lkk-L-mode nil "`L' mode. Direct zenkaku mode.")
(lkk-deflocalvar lkk-q-mode nil "`q' mode. Katakana mode.")
(lkk-deflocalvar lkk-mode nil "lkk basic -- lowest layer mode.")

(defvar lkk-insignia0 "" "Insignia shown while conversion.") 
(defvar lkk-insignia1 "" "Insignia shown on kakutei status.")
;;(defvar lkk-insignia1 "" "Insignia shown while conversion.") 
;;(defvar lkk-insignia1 "" "Insignia shown on kakutei status.")

;; Dummy for authour only.
(defvar lkk-entry "
;; okuri-ari entries.
b //[//]/[//]/
;; okuri-nasi entries.
 ///
 ///ܺ/
skk# /#1˵ܤ/
" "Dummy dict entry.")

;; r2k conversion

(lkk-deflocalvar lkk-j-beginning nil
		 "lkk internal. Point/marker pointing kana start.")
(lkk-deflocalvar lkk-j-numarg nil
		 "lkk internal. Arg passed to Kana sequence.")

;; k2k conversion

(lkk-deflocalvar lkk-conv-beginning nil "lkk internal. Conversion start point.")

(lkk-deflocalvar lkk-/-converting nil "lkk internal. t while / conversion.")

;; Unused. Equivalent to lkk-J-mode.
;;(lkk-deflocalvar lkk-converting nil "Non-nil while kanji conversion.")
(lkk-deflocalvar lkk-current-record nil
		 "Accommodates entry under conversion.")
(lkk-deflocalvar lkk-kanji-index nil
		 "Indexes kanji under conversion. Starts from 0.")

;;;
;;; Minor mode definitions and utils
;;;

;; Several utils related to mode fiddles.

(defconst lkk-exit-key
  (or 
   (car (where-is-internal 'lkk-mode '(keymap))) ; search only global-map
   "\C-xJ")
  "User defined key to exit lkk-mode.
Normally that key is bound in ~/.emacs...
Used to exit directly from lkk-L/l-mode.")

(defun lkk-get-cursor-color ()
  "Get current cursor color -- mainly for getting color before lkk is enabeled."
  (cdr (assoc 'cursor-color (frame-parameters (selected-frame)))))

(defvar lkk-default-cursor-color nil
  "Initial cursor color used when lkk is turned on.")

(lkk-deflocalvar lkk-old-cursor-color nil "Old cursor color backed up.")

;; I don't know why sometimes this won't work
(defun lkk-set-right-cursor-color ()
  (if lkk-mode
      (mapcar
       (lambda (m) 
	 (if (eval m)	      
	     (lkk-set-cursor-color
	      (eval (intern (format "%s-%s" m "cursor-color"))))))
       lkk-valid-minor-minor-modes)
    (lkk-set-cursor-color lkk-default-cursor-color)))

(defun lkk-set-cursor-color (color)
  (condition-case nil
      (set-cursor-color color)
    (error
     (set-cursor-color lkk-default-cursor-color)
     ;; color exhausted, or just invalid color string ?
     (message "Cursor color `%s' can't be set." color)
     )))

;; I don't know what I am doing... Just stole code from skk-10.49.

(or (fboundp 'set-cursor-color)
    (defun set-cursor-color (&rest) nil)) ; NOP

(if (null (fboundp 'lkk-defadvice))
    ()
  (defadvice hilit-recenter (after lkk-ad activate)
    "Set lkk's right cursor color."
    (and lkk-mode (lkk-set-right-cursor-color)))

  (defadvice hilit-yank (after lkk-ad activate)
    "Set lkk's right cursor color."
    (and lkk-mode (lkk-set-right-cursor-color)))

  (defadvice hilit-yank-pop (after lkk-ad activate)
    "Set lkk's right cursor color."
    (and lkk-mode (lkk-set-right-cursor-color)))

  (defadvice yank (after lkk-ad activate)
    "Set lkk's right cursor color."
    (and lkk-mode (lkk-set-right-cursor-color)))

  (defadvice yank-pop (after lkk-ad activate)
    "Set lkk's right cursor color."
    (and lkk-mode (lkk-set-right-cursor-color))) )

(defsubst lkk-setup-minor-mode (modesym &optional cursorcolor no-sit-back)
  "Set lkk's minor minor-mode to MODESYM.
Passing 'lkk-default as MODESYM turns off all minor minor-mode's,
and sets cursor color to lkk-default-cursor-color.
If optional 2nd arg CURSORCOLOR is specified, use it."
;;; 'lkk-default fiddle may be tricky, and I may be confused (in the future).
;;; 'lkk-default matches none of them.
  (mapcar (lambda (mode)
	    (if (eq modesym 'lkk-J-mode)
		(if (eq modesym mode) (set mode t)) ; just turn on `modesym'.
	      (set mode (eq modesym mode)) ; This turns off other mode.
	      ))
	  lkk-valid-minor-minor-modes)

  (or (eq modesym 'lkk-J-mode) ; Can't sit-back if conversion fired allready.
      no-sit-back	       ; or no-sit-back is explicitly spacified.
      (lkk-sit-back))

  (cond
   ((member modesym lkk-valid-minor-minor-modes)
    (or lkk-default-cursor-color
	(progn
	  (setq lkk-default-cursor-color (lkk-get-cursor-color))
	  (add-hook 'kill-buffer-hook
		    (lambda () (set-cursor-color lkk-default-cursor-color)))))
    (setq lkk-old-cursor-color (lkk-get-cursor-color))
    (run-hooks (intern (format "%s-%s" modesym "hook")))))

  ;; Not in the `cond' above, because caller use 'lkk-default to exit mode.
  (lkk-set-cursor-color (or cursorcolor
			    (eval (intern
			       (format "%s-%s" modesym "cursor-color")))))
  (force-mode-line-update))

(defsubst lkk-all-minor-mode-disable ()
  "Turn off lkk's minor minor-mode passing 'lkk-default to lkk-setup-minor-mode."
  (lkk-setup-minor-mode 'lkk-default lkk-default-cursor-color))

;;
;; lkk-j-mode as ordinary minor mode, 
;;

(defvar lkk-j-mode-map nil  "lkk-j-mode-map")

(defun lkk-setup-j-mode-map (&optional remap)
  (or remap
      (setq lkk-j-mode-map (make-sparse-keymap)))
  (let (keylist tem)
    ;; Fill with zenkaku-vector. Dirty.
    (setq tem 0)
    (while (< tem 128)
	(if (aref lkk-zenkaku-vector tem)
	    (define-key lkk-j-mode-map (char-to-string tem) 'lkk-L-self-insert))
      (setq tem (1+ tem)))

    ;; get romaji key
    (mapcar
     (lambda (cell)
       (setq tem (substring (car cell) 0 1))
       ;; Squeeze dup's with or(). length squeezed from 136 -> 34.
       (or (member tem keylist)
	   (setq keylist (cons tem keylist))))
     (append lkk-r2k-alist-C lkk-r2k-alist-V))

    (mapcar
     (lambda (kstr)
       (define-key lkk-j-mode-map kstr 'lkk-j-input)
       (define-key lkk-j-mode-map (lkk-toupper kstr) 'lkk-start-yomi))
     keylist)

    (define-key lkk-j-mode-map "'" 'lkk-j-input) ; one exception

    (define-key lkk-j-mode-map "," 'lkk-insert-j-comma)
    (define-key lkk-j-mode-map "." 'lkk-insert-j-period)
    (define-key lkk-j-mode-map "-" 'lkk-insert-j-minus-char)
    (define-key lkk-j-mode-map "Q" 'lkk-set-conv-beginning)
    (define-key lkk-j-mode-map "q" 'lkk-toggle-kana)

    (define-key lkk-j-mode-map "L" 'lkk-L-mode)
    (define-key lkk-j-mode-map "l" 'lkk-l-mode)

    (define-key lkk-j-mode-map " " 'lkk-convert)
    (define-key lkk-j-mode-map "/" 'lkk-start-q-convert)

    ;; Not yet complete. These two are almost the same.
    (define-key lkk-j-mode-map "\t" 'lkk-completion-convert)
    (define-key lkk-j-mode-map "\e " 'lkk-completion-convert)

    (define-key lkk-j-mode-map "\C-j" 'lkk-sit-back)
    ))

(defvar lkk-j-mode-hook nil
  "Normal hook run when starting to lkk-j-mode.")

(defun lkk-j-mode (&optional arg)
  "\

LousyKanaKanji interface.

Keybindings:

\\{lkk-j-mode-map}"
  (interactive "P")
  (cond
   ((and arg
	 (if (> (prefix-numeric-value arg) 0)
	     lkk-j-mode (not lkk-j-mode)))
    ())					; Do nothing if already OK.
   (lkk-j-mode (lkk-j-mode-disable))
   (t (lkk-j-mode-enable))))

(defun lkk-j-mode-enable ()
  "Turn on lkk mode."
;;;  Not an ordinary minor mode.
;;;  (if (<= (string-to-int emacs-version) 19)
;;;      (add-hook 'change-major-mode-hook 'lkk-j-mode-disable nil)
;;;    (make-local-hook 'change-major-mode-hook)
;;;    (add-hook 'change-major-mode-hook 'lkk-j-mode-disable nil t))
  (lkk-setup-minor-mode 'lkk-j-mode))

(defun lkk-j-mode-disable ()
  "Turn off lkk mode."
;;;  Not an ordinary minor mode.
;;;  (if (> (string-to-int emacs-version) 19)
;;;      (remove-hook 'change-major-mode-hook 'lkk-j-mode-disable t)
;;;    (remove-hook 'change-major-mode-hook 'lkk-j-mode-disable))
  (interactive)
  ;;(lkk-sit-back);; lkk-all-minor-mode-disable does this (I believe).
  (lkk-all-minor-mode-disable))

(fset 'lkk-mode-all-exit 'lkk-j-mode-disable)

;;
;; lkk-J-mode -- kanji conversion mode as ordinary minor mode, 
;;
(defvar lkk-J-mode-map nil  "lkk-J-mode-map")

(defun lkk-setup-J-mode-map ()
  (setq lkk-J-mode-map (make-sparse-keymap))
  (let (keylist tem)

    (define-key lkk-J-mode-map " " 'lkk-reconvert)
;;; -> Lousy! should change this part
;;;    (define-key lkk-J-mode-map "a" 'lkk-select-cand)
;;;    (define-key lkk-J-mode-map "s" 'lkk-select-cand)
;;;    (define-key lkk-J-mode-map "d" 'lkk-select-cand)
;;;    (define-key lkk-J-mode-map "f" 'lkk-select-cand)
;;;    (define-key lkk-J-mode-map "j" 'lkk-select-cand)
;;;    (define-key lkk-J-mode-map "k" 'lkk-select-cand)
;;;    (define-key lkk-J-mode-map "l" 'lkk-select-cand)
;;;    (define-key lkk-J-mode-map ";" 'lkk-select-cand)
;;; -> Lousy! should change this part 

    (define-key lkk-J-mode-map "x" 'lkk-prev-cand)
    (define-key lkk-J-mode-map "X" 'lkk-perge-it)

    (define-key lkk-J-mode-map "\C-j" 'lkk-sit-back)
    ))

(defsubst lkk-J-mode-exit ()
  (setq lkk-J-mode nil)
  (force-mode-line-update))

;;
;; `q' -- aka katakana -- mode.
;; Uses also lkk-j-mode-map.
;;

(defvar lkk-q-mode-hook nil
  "Normal hook run when starting to lkk-q-mode.")

(defun lkk-q-mode (&optional arg)
  "\

LousyKanaKanji interface.

Keybindings:

\\{lkk-j-mode-map}"
  (interactive "P")
  (lkk-setup-minor-mode 'lkk-q-mode))

(defun lkk-q-mode-exit ()
  "Exit from `L' mode to old lkk mode."
  (interactive)
  (lkk-sit-back)
  (lkk-j-mode-enable))

;;
;; `l' mode (lkk-l-mode) as (minor-)minor-mode.
;; Almost the same as lkk-mode below, except mode name string.
;;

(defvar lkk-l-mode-map nil)

(defun lkk-setup-l-mode-map ()
  ;; Just two key only.
  (setq lkk-l-mode-map (make-sparse-keymap))
;;;   (let ((i 0))
;;;     (while (< i 128)
;;;       (if (aref lkk-ascii-vector i)
;;; 	  (define-key lkk-l-mode-map (char-to-string i) 'lkk-l-self-insert) )
;;;       (setq i (1+ i))))
  (define-key lkk-l-mode-map "\C-j" 'lkk-l-mode-exit)
  ;;(define-key lkk-l-mode-map lkk-exit-key 'lkk-mode-all-exit) ; wrong.
  (define-key lkk-l-mode-map lkk-exit-key 'lkk-mode) ; This disables all minor.
  )

;;; (defun lkk-l-self-insert (arg)
;;;   "`l' mode insert.  Direct input without conversion."
;;;   (interactive "*p")
;;;   (let ((key last-command-char))
;;;     (while (> arg 0)
;;;       (insert key) (setq arg (1- arg)))))

(defvar lkk-l-mode-hook nil
  "Normal hook run when starting to lkk-l-mode.")

(defun lkk-l-mode ()
  "\

`l' mode in LousyKanaKanji interface.

Keybindings:

\\{lkk-l-mode-map}"
  (interactive)
  (lkk-setup-minor-mode 'lkk-l-mode))

(defun lkk-l-mode-exit ()
  "Exit from `l' mode to old lkk mode."
  (interactive)
  (lkk-sit-back)
  (lkk-j-mode-enable))

;;
;; `L' mode (lkk-L-mode) as (minor-)minor-mode.
;;

(defvar lkk-L-mode-map nil "Keymap for L mode.")

(defun lkk-setup-zenkaku-map ()
  (setq lkk-L-mode-map (make-sparse-keymap))
  (let ((i 0))
    (while (< i 128)
      (if (aref lkk-zenkaku-vector i)
	  (define-key lkk-L-mode-map (char-to-string i) 'lkk-L-self-insert) )
      (setq i (1+ i)))
    (define-key lkk-L-mode-map "\C-j" 'lkk-L-mode-exit)
    ;;(define-key lkk-L-mode-map lkk-exit-key 'lkk-mode-all-exit) ; wrong.
    (define-key lkk-L-mode-map lkk-exit-key 'lkk-mode) ; This disables all minor.
    ))

(defun lkk-L-self-insert (arg &optional vec)
  "`L' mode insert. Uses lkk-zenkaku-vector."
  (interactive "*p")
  (let* ((key (aref (or vec lkk-zenkaku-vector) last-command-char)))
    (while (> arg 0)
      (insert key) (setq arg (1- arg)))))
(fset 'lkk-self-insert 'lkk-L-self-insert)

(defvar lkk-L-mode-hook nil
  "Normal hook run when starting to lkk-L-mode.")

(defun lkk-L-mode ()
  "\

`L' mode in LousyKanaKanji interface.

Keybindings:

\\{lkk-L-mode-map}"
  (interactive)
  (lkk-setup-minor-mode 'lkk-L-mode))

(defun lkk-L-mode-exit ()
  "Exit from `L' mode to old lkk mode."
  (interactive)
  (lkk-sit-back)
  (lkk-j-mode-enable))

(defvar lkk-j-space-char "")
(defvar lkk-j-comma-char "")
(defvar lkk-j-period-char "")
(defvar lkk-j-minus-char "")

;; These special functions must be eliminated in future version.

(defun lkk-insert-j-space (arg)
  "Insert lkk-space-char by ARG times."
  (interactive "*p")
  (lkk-sit-back)
  (while (> arg 0)
    (insert lkk-j-space-char) (setq arg (1- arg))))

(defun lkk-insert-j-comma (arg)
  "Insert lkk-comma-char by ARG times."
  (interactive "*p")
  (lkk-sit-back)
  (while (> arg 0)
    (insert lkk-j-comma-char) (setq arg (1- arg))))

(defun lkk-insert-j-period (arg)
  "Insert lkk-period-char by ARG times."
  (interactive "*p")
  (lkk-sit-back)
  (while (> arg 0)
    (insert lkk-j-period-char) (setq arg (1- arg))))
  
(defun lkk-insert-j-minus-char (arg)
  "Insert lkk-period-char by ARG times."
  (interactive "*p")
  (lkk-sit-back)
  (while (> arg 0)
    (insert lkk-j-minus-char) (setq arg (1- arg))))

(defun lkk-toggle-kana ()
  "Toggles hiragana / katakana input mode."
  (interactive)
  (if lkk-q-mode
      (lkk-q-mode-exit)
    (lkk-q-mode)))

;;
;; lkk-mode: very basic mode. 
;; Almost NOP except setting header part of minor mode string, and invoking
;; entity of lkk-initial-mode.
;; Uses no key map. 
;;

(defvar lkk-initial-mode 'lkk-j-mode
  "Initial mode. Defaults 'lkk-j-mode. Maybe you can speciry lkk-l-mode etc.")

(defvar lkk-mode-hook nil
  "Normal hook run when starting to lkk mode.")

(defun lkk-mode (&optional arg)
  "\

LousyKanaKanji interface basic layer.

lkk-mode: very basic mode. 
Almost NOP except setting header part of minor mode string, and invoking
entity of lkk-initial-mode.
Uses no key map. 
"
  (interactive "P")
  (cond
   ((and arg
	 (if (> (prefix-numeric-value arg) 0)
	     lkk-mode (not lkk-mode)))
    ())					; Do nothing if already OK.
   (lkk-mode
    (setq lkk-mode nil)
    (lkk-mode-all-exit)
    (force-mode-line-update))
   (t 
    (setq lkk-mode t)
    (force-mode-line-update)
    (run-hooks 'lkk-mode-hook)

    (lkk-j-mode-enable) ; Call this, then lkk-initial-mode (dirty).
    (cond ((eq lkk-initial-mode 'lkk-l-mode) (lkk-l-mode))
	  ((eq lkk-initial-mode 'lkk-L-mode) (lkk-L-mode))
	  ((eq lkk-initial-mode 'lkk-q-mode) (lkk-q-mode))
	  ((eq lkk-initial-mode 'lkk-j-mode) (lkk-j-mode-enable))
	  ))))

;;;
;;; Setup minor-mode alists here. Order is very significant.
;;;
(progn
  ;; j-mode
  (or lkk-j-mode-map
      (lkk-setup-j-mode-map))
  (or (assq 'lkk-j-mode minor-mode-alist)
      (setq minor-mode-alist
	    (cons '(lkk-j-mode " ") minor-mode-alist)))
  (or (assq 'lkk-j-mode minor-mode-map-alist)
      (setq minor-mode-map-alist
	    (cons (cons 'lkk-j-mode lkk-j-mode-map) minor-mode-map-alist)))
  ;; J-mode
  (or lkk-J-mode-map
      (lkk-setup-J-mode-map))
  (or (assq 'lkk-J-mode minor-mode-alist)
      (setq minor-mode-alist
	    (cons '(lkk-J-mode "*") minor-mode-alist)))

  ;; lkk-J-mode-map *MUST* precede lkk-j-mode-map in minor-mode-map-alist, or
  ;; we can't override lkk-j-mode-map.
  (or (assq 'lkk-J-mode minor-mode-map-alist)
      (setq minor-mode-map-alist
	    (cons (cons 'lkk-J-mode lkk-J-mode-map) minor-mode-map-alist)))
  ;; q-mode
  (or (assq 'lkk-q-mode minor-mode-alist)
      (setq minor-mode-alist
	    (cons '(lkk-q-mode " ") minor-mode-alist)))

  (or (assq 'lkk-q-mode minor-mode-map-alist)
      (setq minor-mode-map-alist
	    (cons (cons 'lkk-q-mode lkk-j-mode-map) minor-mode-map-alist)))
  ;; l-mode
  (or lkk-l-mode-map
      (lkk-setup-l-mode-map))

  (or (assq 'lkk-l-mode minor-mode-alist)
      (setq minor-mode-alist
	    (cons '(lkk-l-mode " ASCII") minor-mode-alist)))

  (or (assq 'lkk-l-mode minor-mode-map-alist)
      (setq minor-mode-map-alist
	    (cons (cons 'lkk-l-mode lkk-l-mode-map) minor-mode-map-alist)))
  ;; L-mode
  (or lkk-L-mode-map
      (lkk-setup-zenkaku-map))
  (or (assq 'lkk-L-mode minor-mode-alist)
      (setq minor-mode-alist
	    (cons '(lkk-L-mode " ") minor-mode-alist)))
  (or (assq 'lkk-L-mode minor-mode-map-alist)
      (setq minor-mode-map-alist
	    (cons (cons 'lkk-L-mode lkk-L-mode-map) minor-mode-map-alist)))
 ;; lkk-mode (basic)
  (or (assq 'lkk-mode minor-mode-alist)
      (setq minor-mode-alist
	    (cons '(lkk-mode " ˦ʦ") minor-mode-alist)))
  )

;;;
;;; Dictionary handlers
;;;
(defvar lkk-okuri-ari-dict-top nil)
(defvar lkk-okuri-nasi-dict-top nil)

(defun lkk-read-dict ()
  "Read in lkk-dict destructively.  Return lkk-dict-buffer-name.
Low level. Don't call directly. Use lkk-get-dict instead."
  (let ((buf (get-buffer-create lkk-dict-buffer-name)))
    (save-excursion
      (set-buffer buf) (erase-buffer)
      (or (file-readable-p lkk-dict) (error "Can't read %s" lkk-dict))
      (message "Reading lkk dictionary %s..." lkk-dict)
      (insert-file-contents lkk-dict)
      (message "")
      (goto-char (point-min))
      (save-excursion (search-forward ";; okuri-ari" (point-max) t)
		      (forward-line 1)
		      (setq lkk-okuri-ari-dict-top (point-marker)))
      (save-excursion (search-forward "\n;; okuri-nasi" (point-max) t)
		      (forward-line 1)
		      (setq lkk-okuri-nasi-dict-top (point-marker))))
    buf))

(defsubst lkk-get-dict ()
  "Return dict buffer as buffer obj."
  (or (get-buffer lkk-dict-buffer-name)
      (lkk-read-dict)))

(defun lkk-dict-search (yomi &optional start loose)
  (save-excursion
    (set-buffer (lkk-get-dict))
    (goto-char (if start start (point-min)))
    (if (search-forward (concat "\n" yomi (if loose "" " ")) (point-max) t)
;;; Using "^" slows down search very much.
;;;   600ms -> 2.1s for 300kb dict (search for last entry "󤭤礯")
;;;     (re-search-forward (concat "^" yomi (if strict " " "")) (point-max) t)
	(lkk-get-current-record)
      nil)))

;;(lkk-read-dict) ;; Now it's done by lkk-get-dict().

;;
;; Henkan Subroutines
;; Not yet complete at all.
;;

(defsubst lkk-get-yomi (entry)
  (car entry))
(defsubst lkk-get-kanji-list (entry)
  (cdr entry))

(defun lkk-sit-back (&optional arg)
  "Confirm all partial status, and sit back."
  (interactive "*P")

  ;; Debugger
  (if lkk-debug-hook (run-hooks 'lkk-debug-hook))

  ;; Kakutei proc must be done here.
  (cond (lkk-J-mode
	 (lkk-J-mode-exit)))

  (cond (lkk-conv-beginning
	 (setq lkk-current-record nil lkk-kanji-index nil)
	 (save-excursion
	   (goto-char lkk-conv-beginning)
	   (if (looking-at lkk-insignia0)
	       (delete-char 1)))
	 (setq lkk-conv-beginning nil)))
  
  (cond (lkk-j-beginning
	 (let ((char-left
		(buffer-substring-no-properties lkk-j-beginning (point))))
	   (cond ((string-match "^[nm]+$" char-left)
		  (delete-region lkk-j-beginning (point))
		  (insert (if lkk-q-mode "" "")))
		 ((string-match "\\([a-zA-Z]\\)\\1$" char-left)
		  (delete-region lkk-j-beginning (point))
		  (insert (if lkk-q-mode "" "")))))))
  (setq lkk-/-converting nil
	lkk-j-beginning nil
	lkk-j-numarg 1)
  ;; conversion status cleanup here.
  )
(fset 'lkk-kakutei 'lkk-sit-back)

(defun lkk-start-yomi (arg)
  "Normally bound to uppercase char in lkk-j-mode."
  (interactive "*p")
  ;; While partial kana input or not converting yet,
  ;; uppercase is just converted into lower, and do no point fiddle.
  (or lkk-j-beginning
      ;;(null lkk-J-mode)    ;; This part will be difficult -- okuri etc.
      (progn (lkk-set-conv-beginning)
	     (setq lkk-j-beginning (point-marker))))
  (let ((last-command-char (lkk-tolower last-command-char)))
    (lkk-j-input arg)))
  
(defun lkk-set-conv-beginning (&optional arg no-sit-back)
  "Normally bound to `Q' in lkk-j-mode-map."
  (interactive "*P\nP")			; arg unused yet
  (or no-sit-back
      (if lkk-conv-beginning (lkk-sit-back)))
  (setq lkk-conv-beginning (point-marker))
  (insert-and-inherit lkk-insignia0))

(defun lkk-convert (&optional arg)
  "Henkan command. Normally bound to SPC in lkk-j-mode-map.
Explicit numeric ARG such as M-3, C-4 etc. converts such chars before point.
If ARG is just C-u, converts current region RB RE is converted."
  (interactive "*P")

  ;; Debugger
  (if lkk-debug-hook (run-hooks 'lkk-debug-hook))

  (setq lkk-/-converting nil)
  (lkk-setup-minor-mode 'lkk-J-mode)

  (cond
   ((and (null lkk-conv-beginning) (null arg))
    (lkk-insert-j-space (prefix-numeric-value arg)))
   ((and arg (listp arg) 		; Just C-u only
	 (condition-case nil (region-end) (error nil))) ; nil if no mark at all.
    (lkk-convert-region (region-beginning) (region-end) 1))
   ((numberp arg)			; lkk-henkan-with-arg equivalent.
    (setq arg (prefix-numeric-value arg))
    (let ((rb (save-excursion (backward-char arg) (point)))
	  (re (point)))
      (lkk-convert-region rb re 1)))

   ((and lkk-conv-beginning
	 (null (lkk-same-line-p lkk-conv-beginning (point))))
    (lkk-sit-back))

   (t
    (cond
     (lkk-conv-beginning
	  (lkk-convert-region (+ lkk-j-bytes lkk-conv-beginning) (point) 1)
	))
    )))

(defun lkk-completion-convert (arg)
  "Normally bound to TAB in lkk-j-mode.
Next completion facility is not yet implemented."
  (interactive "*P")
  (cond 
   (lkk-conv-beginning			; else NOP
    (setq lkk-/-converting nil)
    (lkk-setup-minor-mode 'lkk-J-mode)

    (lkk-convert-region (+ lkk-j-bytes lkk-conv-beginning) (point) 1 'loose))))

(defun lkk-convert-region (rb re &optional arg loose) ; arg should be "p"
  (let* ((dtop lkk-okuri-nasi-dict-top)	; Only non-okuri yet.
	 (seed (buffer-substring-no-properties rb re))
	 (entry (lkk-dict-search seed dtop loose)))
    (cond (entry
	   (or lkk-conv-beginning
	       (save-excursion (goto-char rb) (lkk-set-conv-beginning nil)))
	   (setq lkk-current-record entry
		 lkk-kanji-index 1)
	   (lkk-relpace-buffer-substring (+ lkk-j-bytes lkk-conv-beginning)
					 (point)
					 (nth lkk-kanji-index entry) arg))
	  (t nil))))

;; Must add code to control range.
(defun lkk-reconvert (arg)
  "Henkan-again command. Normally bound to SPC in lkk-J-mode-map."
  (interactive "*p")

  ;; Debugger
  (if lkk-debug-hook (run-hooks 'lkk-debug-hook))

  (cond (lkk-conv-beginning
	 (setq lkk-kanji-index (1+ lkk-kanji-index))
	 (lkk-relpace-buffer-substring
	  (+ lkk-j-bytes lkk-conv-beginning) (point)
	  (nth lkk-kanji-index lkk-current-record))))
  )

(defun lkk-start-q-convert (arg)
  "Normally bound to / in lkk-j-mode-map."
  (interactive "*p")
  (lkk-set-conv-beginning)
  (setq lkk-/-converting t) ; lkk-j-input() refers, and reset by lkk-j-*convert*
;;; Quick and dity -- use read-char().
;;;  (let ((c))
;;;    (while (and (> (setq c (read-char)) ? )
;;;		(<= c ?~))		; C-j, Lousy.
;;;      (insert-and-inherit c))
;;;    (cond ((= c ? ) (lkk-convert nil))
;;;	  (t        (lkk-sit-back))))
  )

;; Must add code to control range.
(defun lkk-prev-cand (arg)
  (interactive "*p")

  ;; Debugger
  (if lkk-debug-hook (run-hooks 'lkk-debug-hook))

  (cond (lkk-conv-beginning
	 (setq lkk-kanji-index (1- lkk-kanji-index))
	 (lkk-relpace-buffer-substring
	  (+ lkk-j-bytes lkk-conv-beginning) (point)
	  (nth lkk-kanji-index lkk-current-record)))))

;;;;; Bare Test versions hereunder.

(defun lkk-convert-str (seed)
  (lkk-dict-search seed nil t))

;;; Bare Test version. Conversion to 1st candidate only.
;;;(defun lkk-henkan-with-arg (arg)
;;;  (interactive "*P")
;;;  (if arg
;;;      (cond 
;;;       ((numberp arg)
;;;	(setq arg (prefix-numeric-value arg))
;;;	(let ((rb (save-excursion (backward-char arg) (point)))
;;;	      (re (point)))
;;;	  (lkk-convert-region rb re)))
;;;       ((listp arg)
;;;	(lkk-convert-region (region-beginning) (region-end)))
;;;       (t nil))
;;;    nil))

;; from ~/elisp/filter.el
(defun lkk-str-to-list (str &optional delim)
  "Split STR into word list. Default delimiter is space or tab.
Optional 2nd arg DELIM is used if it's non-nil."
  (let ((n (string-match (or delim "[ \t]") str)))
    (cond ((= 0 (length str)) nil)
	  (n (append (list (substring str 0 n))
		    (lkk-str-to-list (substring str (1+ n)))))
	  (t (list str)))))

(defun lkk-get-current-record-string ()
  "Return current entry -- current line string."
  (buffer-substring-no-properties
   (save-excursion (beginning-of-line) (point))
   (save-excursion (end-of-line)       (point))))

(defun lkk-get-current-line ()
  "Return current entry as list."
  (lkk-str-to-list 
   (substring
    (lkk-replace-in-string  " */" " " (lkk-get-current-record-string)) 0 -1)))
(fset 'lkk-get-current-record 'lkk-get-current-line)

(defun lkk-replace-in-string (regexp newtext string)
  "Replace REGEXP with NEWTEXT all over STRING. NEWTEXT is taken literally."
  (let ((result "") (start 0))
    (while (string-match regexp string start)
      (setq result
	    (concat result (substring string start (match-beginning 0)) newtext)
       start (match-end 0)))
    (concat result (substring string start))))

(defun lkk-henkan (yomi &optional dic-entry dic-index)
  "Return kanji for YOMI.
Uses optional 2nd and 3rd arg DIC-ENTRY and DIC-INDEX if they are both non-nil."
  (cond ((and dic-entry dic-index)
      (if (string= yomi (lkk-get-yomi dic-entry))
	  (nth dic-index dic-entry)))
	(t
	 (if (string= yomi (lkk-get-yomi lkk-current-record))
	     (nth lkk-kanji-index lkk-current-record)))))

;;(setq lkk-current-record '("" "" ""))
;;(setq lkk-kanji-index 3)
;;(setq lkk-kanji-index 2)
;;(setq lkk-kanji-index 1)
;;
;;(lkk-henkan "")
;;""
;;(setq x (lkk-get-current-record))
;;(lkk-henkan "" x 1)
;;""

;;; 
;;; Japanese input automaton
;;;

(defsubst lkk-relpace-buffer-substring (rb re str &optional arg)
  "Replace region RB to RE with string STR. Opt. 2nd ARG : times of STR insert."
  ;; (goto-char rb)
  (delete-region rb re)
  (or arg (setq arg 1))
  (while (> arg 0)
    (insert str)
    (setq arg (1- arg))))

(defsubst lkk-same-line-p (pt0 pt1)
  ;; Non-nil if pt0 =< pt1 and no newline between them.
  ;; The idea was in skk-5.30.
  (and (<= pt0 pt1)
       (save-excursion
	 (goto-char pt1)
	 (beginning-of-line 1)
	 (<= (point) pt0))))

;;
;; r2k input
;;
;; `C-u 4 ccha' -- do not generate "ääää" like skk, but
;; this just suppresses "äää".  More decent one is in ./experimental/*.el
;;
;; To be simple, clean, and neat:
;;  NEVER use read-char() / read-event().
;;  NEVER push anything back to unread-command-events.
;;  NEVER keep string in global variables -- to assure user's freedom.
;;  NEVER rely on counts of chars/strings if they are not ASCII.
;;    -- the result depends on emacs/mules version
;;

(defun lkk-j-input (arg)
  "Simple Japanese input automaton -- converts romaji sequence to kana."
  (interactive "*p")

  (let ((roma) (kana)
	(key last-command-char)
	(start (point-marker)))

    (if (or (and lkk-j-beginning
	     (null (lkk-same-line-p lkk-j-beginning (point))))
	    ;; skk-kakutei-early == t equivalent operation.
	    ;; You can't customize this default yet (or forever ?).
	    ;; Lousy ?
	    lkk-J-mode)
	(lkk-sit-back))

    ;; Debugger
    (if lkk-debug-hook (run-hooks 'lkk-debug-hook))

    (or lkk-j-beginning (setq lkk-j-beginning start
			  lkk-j-numarg arg))

    (insert key)			; Show what user entered anyway.
    (setq roma (buffer-substring-no-properties lkk-j-beginning (point)))

    (cond

     ;; Exceptional ASCII direct input mode after `/' is hit.
     (lkk-/-converting ;; already inserted one, so check if gt 1.
      (while (> arg 1) (insert key) (setq arg (1- arg)))
      (setq lkk-j-beginning nil lkk-j-numarg 1))

     ;; Vowel -- one char sequence. Region start -- (point) contains the `key'.
     ((and (equal lkk-j-beginning start)
	   (setq kana (lkk-get-kana0 roma lkk-r2k-alist-V)))
      (lkk-relpace-buffer-substring lkk-j-beginning (point) kana lkk-j-numarg)
      (setq lkk-j-beginning nil lkk-j-numarg 1))

     ;; Else consonant hereunder -- already kana sequence is started.

     ;; Try simple romaji sequence matching in lkk-r2k-alist(-vector).
     ((and (lkk-r2k-terminator-p (substring roma -1))
	   (setq kana (lkk-get-kana roma)))
      (lkk-relpace-buffer-substring lkk-j-beginning (point) kana lkk-j-numarg)
      (setq lkk-j-beginning nil lkk-j-numarg 1))

     ;; Else invalid or still-partial sequence hereunder.
     ;; lkk-r2k-alist is not yet matched hereunder.
     ;; "" and "" type of exception is checked first.

     ((and (>= (- (point) lkk-j-beginning) 2)
	   (string-equal (substring roma 0 1) (substring roma 1 2)))
      (lkk-relpace-buffer-substring lkk-j-beginning (point)
				    (if lkk-q-mode "" "") nil)
      (setq lkk-j-beginning (point-marker) lkk-j-numarg 1) ; ignore arg
      (insert (substring roma 1)))

     ;; "nn", "mm" etc. should have benn matched already.
     ((and (>= (- (point) lkk-j-beginning) 2)
	   (or (string-equal "n" (substring roma 0 1))
	       (string-equal "m" (substring roma 0 1)))
	   (not (lkk-valid-roma-p roma)))
      (lkk-relpace-buffer-substring lkk-j-beginning (point)
				    (if lkk-q-mode "" "" ) nil)
      (setq lkk-j-beginning (point-marker))
      ;;(setq lkk-j-numarg 1)		; ignore arg
      (insert (substring roma 1)))

     ;; Reset "sd" "cccch" sort of invalid sequence as in skk.
     ;; This test must be done at last.
     ((not (lkk-valid-roma-p roma))
      (delete-region lkk-j-beginning (point)) ; Erase including `key' inserted.
      ;; Fake as if we have just started sequence.
      (setq lkk-j-beginning (point-marker) lkk-j-numarg 1)
      (insert key))

     )))

;; Run load-hook
(run-hooks 'lkk-load-hook)
