;;; dump-lkk.el -- dump direct-loadable lkk-dict

;; Wed Mar  8 11:17:22 2000 s.n.
;; Implemented lkk-dict-method == lkk-dict-alist part ONLY.

(require 'lkk)
(provide 'dump-lkk)

(defun dump-lkk (arg)
  "Dump current lkk dictionary image to `lkk-dict-dump-file2' in Level2 format.
If ARG is non-nil, the byte-compiled
dictionary image, and required programs are dumped onto `lkk-core-file'"
  (interactive "P")
  (cond ((eq lkk-dict-method 'lkk-dict-alist)
	 (dump-lkk-dict-alist arg))
	((eq lkk-dict-method 'lkk-dict-buffer)
	 (dump-lkk-dict-buffer arg))
	(t (message "dump-lkk does not support method `%s' yet."
		    lkk-dict-method))))

(defvar lkk-dict-dump-file2 "~/lkk-dict.dump.el")
(defvar lkk-core-file "~/lkk-dict.core.el")

(defun dump-lkk-dict-alist (arg)
  "Create level2 format lkk-dict"
  (cond (arg
	 (and (dump-lkk-level2-dict-file)
	      (dump-lkk-level2-core-file)))
	 (t
	  (dump-lkk-level2-dict-file))))

(defun dump-lkk-level2-dict-file (&optional buf)
  "Dump Level2 ditionary core image onto `lkk-dict-dump-file2'.
Core image is derived from variables `lkk-big-dict-okuri' and `lkk-big-dict',
or just dictionary buffer BUF specified by optional arg.
The contents of BUF is destructively modified. do not specify `lkk-dict-buffer'
for BUF."
  (let ((file1 (expand-file-name lkk-dict-dump-file2))
	(inhibit-quit t)
	(l0buf (if buf buf (lkk-dict-alist-dump-dict 'justimage))) ; level 0
	(l2buf))
    (message "Getting dict file image from core...")
    (setq l2buf (dump-lkk-get-level2-dict-image l0buf))
    (cond
     (l2buf
      (save-excursion
	(set-buffer l2buf)
	(message "Writing %s..." file1)
	(write-region (point-min) (point-max) (expand-file-name file1))
	(message "Writing %s...Done" file1))
      (kill-buffer l2buf)
      t)
     (t nil))
    ))
 
(defun dump-lkk-gunzip (file)
  "If gzip'ed file corresponding to FILE gunzip it."
  (let ((file (expand-file-name
	       (concat file
		       (if (string-match ".gz" file)
			   ""
			 ".gz"))))
	(inhibit-quit t))
    (cond ((and (file-exists-p file)
		(file-writable-p file))
	   (message "gunzipping %s" file)
	   (call-process shell-file-name nil nil nil "-c"
			 (concat "gunzip " file)))
	  (t (error "Can't gunzip %s" file)))))

(defun dump-lkk-gzip (file)
  "Gzip FILE."
  (let ((file (expand-file-name file))
	(inhibit-quit t))
    (cond ((and (file-exists-p file)
		(file-writable-p file))
	   (message "gzipping %s" file)
	   (call-process shell-file-name nil nil nil "-c"
			 (concat "gzip " file)))
	  (t (error "Can't gzip %s" file)))))
    
(defun dump-lkk-get-el-or-gunzip (file)
  "Test if FILE has `.el' suffix and readable. Returns string FILE onsuccess.
If FILE has `.elc' suffix and corresponding `.el' file is readable,
the `el' filename is returned.  Otherwise further search for `el.gz',
and if it exists, gunzip it, and return a list containing just
corresponding `el' filename."
  (let ((file (expand-file-name file)) gzipf)
    (cond
     ((and (string-match "^.+\\.el$" file)
	   (file-readable-p file))
      file)				; The .el is there.
     ((and (string-match "^.+\\.elc$" file) ; corresponding .el exists ?
	   (file-readable-p (substring file 0 -1)))
      (substring file 0 -1))
     ((and (string-match "^.+\\.elc$" file) ; seek for gzip'ed one.
	   (setq file (substring file 0 -1)
		 gzipf (concat file ".gz")) ; destroy original `file' value
	   (file-exists-p gzipf)
	   (null (file-symlink-p gzipf))
	   (file-readable-p gzipf)
	   (file-writable-p gzipf)
	   (condition-case nil
	       (progn
	       (dump-lkk-gunzip file)
	       (list file))
	     (error (progn (ding) (message "Can't gunzip %s" file) nil)))
	   ))
     (t nil))))

(defun dump-lkk-level2-core-file ()
  (let* ((file1 (expand-file-name lkk-dict-dump-file2))
	 (file2 (expand-file-name lkk-core-file))
	 (lkkf1 (locate-library "lkk-lkk-dichead"))
	 (lkkf2 (locate-library "lkk"))
	 (inhibit-quit t)
	 lkkf1zip lkkf2zip buf)
;;; Won't work on insert-file-contents
;;;    (or (featurep 'jam-zcat)
;;;	(featurep 'jka-compr)
;;;	(if (y-or-n-p			; lkk installer installs gzip'ed *.el
;;;	     "Your emacs lacks `jka-compr' feature.  Load it now ")
;;;	    (require 'jka-comr)))
    
    (cond ((listp (setq lkkf1 (dump-lkk-get-el-or-gunzip lkkf1)))
	   (setq lkkf1 (car lkkf1)
		 lkkf1zip t)))
    
    (cond ((listp (setq lkkf2 (dump-lkk-get-el-or-gunzip lkkf2)))
	   (setq lkkf2 (car lkkf2)
		 lkkf2zip t)))

    ;; FIXME Can't re-gzip if error occurs.

    ;; Write to lkk-core-file (*.el)
    (message "Creating %s..." file2)
    (save-excursion
      (if (get-buffer (file-name-nondirectory file1))
	  (kill-buffer (file-name-nondirectory file1)))
      (set-buffer (setq buf (find-file-noselect file1))) ; level2 dict read
      (widen) (goto-char (point-min))
      (forward-line 1)
      (insert "(setq lkk-dict-method 'lkk-dict-alist)\n") ; ensure method

      (goto-char (point-max))
      (save-excursion			; to read lkkf2 top of lkkf1
	(insert-file-contents lkkf2))
      (insert-file-contents lkkf1)

      (set-visited-file-name file2)
      (save-buffer))

    (kill-buffer buf)
    (if lkkf1zip (dump-lkk-gzip lkkf1))
    (if lkkf2zip (dump-lkk-gzip lkkf2))

    (message "Compiling %s..." (concat file2 "c"))
    (save-excursion
      (byte-compile-file file2))))

;; " *lkk-dict*" dict image can't be used
(defun dump-lkk-dict-buffer (arg)
  ;; not yet
  ())

(defun dump-lkk-get-level2-dict-image (buf)
  "Destructively convert lkk dictionary buffer BUF to level2 format.
The `level2' format is dictionary direct-loadable with load or load-file.
Do not feed buffer object pointed by lkk-dict-buffer."
  (let ((inhibit-quit t))
    (cond
     ((and buf (get-buffer buf))
      (let ((hdrstr ";; %s VERSION %s LEVEL2\n")
	    (oahdr ";; okuri-ari\n")
	    (onhdr ";; okuri-nasi\n")
	    (oasexp0 "(setq lkk-big-dict-okuri '(\n")
	    (oasexp1 "))\n")
	    (onsexp0 "(setq lkk-big-dict '(\n")
	    (onsexp1 "))\n")
	    (case-fold-search nil)
	    (inhibit-quit t))
	(save-excursion
	  (set-buffer buf)
	  (goto-char (point-min))

	  (search-forward oahdr)	; okuri-ari
	  (forward-line 0)
	  (insert oasexp0)

	  (search-forward onhdr)
	  (forward-line -1)
	  (insert oasexp1)

	  (forward-line 1)		; okuri-nasi
	  (insert onsexp0)

	  (goto-char (point-max))
	  (insert onsexp1)
	  ))
      buf)
     (t nil))))
