;; Fri Feb 18 01:19:44 2000
;; Hash r2k-alist into 128 (1st char) and strlen.

(eval-and-compile
  (require 'lkk))
(provide 'lkk-compound2)

;; Currently may be unused.
(defun lkk-k-yomi-prepend (alist y0 k0 k1)
  "Read ALIST, and prepend Y0 to yomi, K0 to hiragana, and K1 to katakana
If y0 is nil top char of yomi is doubled \(`komatta' sort of conversion\).
Returns resulted alist.

To see what this function is all about, try:
\(lkk-k-yomi-prepend lkk-r2k-alist-C \"n\" \"$B$s(B\" \"$B%s(B\"\)"
  (let ((table ()))
    (mapcar
     (lambda (cell)
       (let* ((yomi (car cell))
	      (top (substring yomi 0 1))
	      (hira (car (cdr cell)))
	      (kata (cdr (cdr cell)))
	      (newH (concat k0 hira))
	      (newK (concat k1 kata)))
	 (cond
	  (y0 ;; "n" or "m" should specified as y0.
	   (or (and (string= k0 "$B$s(B") (string= k1 "$B%s(B"))
	       (error "buggy lkk-k-yomi-prepend call: K0 or K1 is wrong."))
	   ;; Do not generate "nn"/"mm" head, but "nm/mn" head is built.
	   ;; k a m n o m s a m a ==> $B$+$s$N$s$5$^(B allowed.
	   (or (and (string= top "n") (string= top y0))
	       (and (string= top "m") (string= top y0))
	       (setq table
		     (cons 
		      (cons (concat y0 yomi) 
			    (cons newH newK))
		      table))))
	  ;; Double consonant, Equivalent to (not y0). "tta" etc. only.
	  (t 
	   (or (and (string= k0 "$B$C(B") (string= k1 "$B%C(B"))
	       (error "buggy lkk-k-yomi-prepend call: K0 or K1 is wrong."))
	   ;; NOP: "nnu" => "$B$C$L(B", "mmu" => "$B$C$`(B" etc. are not correct.
	   (or (string= top "n")
	       (string= top "m")
	       (setq table
		     (cons
		      (cons (concat top yomi)
			    (cons newH newK))
		      table)))))))
     alist)
    ;; Remove "nyu" (sort of) generated by prepending "n" to "yu",
    ;; which is already included in `primitive-table'.
    ;; `table' is newly created here (not including primitive-table),
    ;; so removing them is okay.
    ;; TODO: might include equiv proc into mapcar above (might not be readable?).
    ;;(nreverse table)
    (let ((primitive-table lkk-r2k-alist-C)
	  (tem))
      (mapcar (lambda (c)
		(or (assoc (car c) primitive-table)
		    (setq tem (cons c tem)))
		)  table)
      (nreverse tem)) ))

(defvar lkk-r2k-alist-compound nil
  "Compound r2k list containing `tta' etc.")
(defvar lkk-r2k-alist-vector-compound nil
  "Vector of compound r2k list containing `tta' etc.")

(defun lkk-build-r2k-alist-vector-compound (alis)
  "Hash ALIS into 128x8 vector with car's top char ASCII value, and strlen."
  (let (ch
	slen
	(vec (apply 'vector (mapcar (lambda (a) (make-vector 16 nil))
				    (make-list 128 nil)))))
    (mapcar (lambda (c)
	      (setq ch (string-to-char (car c))	; s-to-ch only return 1st ch.
		    slen (length (car c)))
	      (aset (aref vec ch) slen 
		    (cons c (aref (aref vec ch) slen)))
	      ) alis)
    vec))

;;(setq lkk-r2k-alist-vector-compound
;;      (lkk-build-r2k-alist-vector-compound lkk-r2k-alist-compound))

(defun lkk-setup-r2k-internal-compound ()
  "Setup lkk's internal alist/list/vectors for Romaji conversion."
  ;; setq order is very significant.
  (setq lkk-r2k-alist-compound
	(append
	 (lkk-k-yomi-prepend lkk-r2k-alist-C nil "$B$C(B" "$B%C(B")
	 (lkk-k-yomi-prepend lkk-r2k-alist-C "n" "$B$s(B" "$B%s(B")
	 (lkk-k-yomi-prepend lkk-r2k-alist-C "m" "$B$s(B" "$B%s(B")))
  (or lkk-r2k-alist
      (setq lkk-r2k-alist
	    (append 
	     lkk-r2k-alist-C
	     lkk-r2k-alist-V)))
  ;; rebuild consonant list
  (setq lkk-valid-consonant-list
	(lkk-build-valid-consonant-list
	 (append lkk-r2k-alist lkk-r2k-alist-compound)))
  (setq lkk-valid-consonant-vector
	(lkk-build-valid-consonant-vector lkk-valid-consonant-list))

;;;  (setq lkk-r2k-alist-vector-compound
;;;	(lkk-build-r2k-alist-vector lkk-r2k-alist-compound))
  (setq lkk-r2k-alist-vector-compound
	(lkk-build-r2k-alist-vector-compound
	 (append lkk-r2k-alist-compound lkk-r2k-alist)))
  )

(lkk-setup-r2k-internal)
(lkk-setup-r2k-internal-compound)

(defsubst lkk-get-kana-compound (str)
  "Return kana corresponding to romaji in STR."
  (let ((table
	 (aref (aref lkk-r2k-alist-vector-compound (string-to-char str))
	       (length str))))
    (if table (lkk-get-kana0 str table) nil) ))

;;(lkk-bogo-perf '(lkk-get-kana-compound "nsyu") 1)  -> (0 . 127)
;;(lkk-bogo-perf '(lkk-get-kana0 "nsyu" lkk-r2k-alist-compound) 1) -> (0 . 330)
;;
;; Test codes
;;(aref (aref lkk-r2k-alist-vector-compound ?n) 4)
;; (mapcar
;;  (lambda (a)
;;    (mapcar
;;     (lambda (a)
;;       (length a)
;;        ) a)
;;    ) lkk-r2k-alist-vector-compound)
;; 
;; (mapcar
;;  (lambda (a)
;;    (mapcar
;;     (lambda (a)
;;       (list (length a)
;; 	    (mapcar (lambda (c)
;; 		      (car c)
;; 		      ) a))
;;        ) a)
;;    ) lkk-r2k-alist-vector-compound)


(defun lkk-j-input (arg)
  "Simple Japanese input automaton -- converts romaji sequence to kana."
  (interactive "*p")

  (let ((roma) (kana)
	(key last-command-char)
	(start (point-marker)))

    (or lkk-j-beginning (setq lkk-j-beginning start
			  lkk-j-numarg arg))

    (insert key)			; Show what user entered anyway.
    (setq roma (buffer-substring-no-properties lkk-j-beginning (point)))

    (cond
     ;; Vowel -- one char sequence. Region start -- (point) contains the `key'.
     ((and (equal lkk-j-beginning start)
	   (setq kana (lkk-get-kana0 roma lkk-r2k-alist-V)))
      (lkk-relpace-buffer-substring lkk-j-beginning (point) kana lkk-j-numarg)
      (setq lkk-j-beginning nil lkk-j-numarg 0))

     ;; Else consonant hereunder -- already kana sequence is started.

     ;; Try simple romaji sequence matching in lkk-r2k-alist.
     ;;((setq kana (lkk-get-kana roma))
     ((setq kana (lkk-get-kana-compound roma))
      (lkk-relpace-buffer-substring lkk-j-beginning (point) kana lkk-j-numarg)
      (setq lkk-j-beginning nil lkk-j-numarg 0))

     ;; Else invalid or still-partial sequence hereunder.
     ;; lkk-r2k-alist is not yet matched hereunder.
     ;; "$B$C(B" and "$B$s(B" type of excetion is checked first.

;;;     ((and (>= (- (point) lkk-j-beginning) 2)
;;;	   (string-equal (substring roma 0 1) (substring roma 1 2)))
;;;      (lkk-relpace-buffer-substring lkk-j-beginning (point)
;;;				    (if lkk-q-mode "$B%C(B" "$B$C(B") nil)
;;;      (setq lkk-j-beginning (point-marker) lkk-j-numarg 1) ; ignore arg
;;;      (insert (substring roma 1)))
;;;
;;;     ((and (>= (- (point) lkk-j-beginning) 2)
;;;	   (string-equal "n" (substring roma 0 1)))
;;;      (lkk-relpace-buffer-substring lkk-j-beginning (point) "$B$s(B" nil)
;;;      (setq lkk-j-beginning (point-marker) lkk-j-numarg 1) ; ignore arg
;;;      (insert (substring roma 1)))

     ;; "sd" "cccch" sort of invalid sequence.  Reset sequence as in skk.
     ((not (lkk-valid-roma-p roma))
      (delete-region lkk-j-beginning (point)) ; Erase including `key' inserted.
      ;; Fake as if we have just started sequence.
      (setq lkk-j-beginning (point-marker) lkk-j-numarg 1)
      (insert key))

     )))
