#!/usr/local/bin/python

# skk.py:
# SKK JISYO (dictionary) handlers.
# Defines JISYO management (currently read-in only), 
# and editor-independent part of kana-kanji-conversion
# Authour: S.Namba <sn@asahi-net.email.ne.jp>

import sys,os,string

myjisyo='/home/sn/.skk-jisyo'     # Dictionary filename
JISYO={}                # My JISYO. Global

# skk jisyo's structure:
# yomi /cand0/cand1/.../candN/
# The '/' separated kanji candidate part is loaded into a list
# associated with the yomi.
# UTF conversion by (nkf+)tcs
def getjisyo(jisyo):
    #file=open(jisyo, 'r')
    #file=os.popen('nkf -e %s|tcs -sf ujis' % (jisyo))
    file=os.popen('tcs -sf ujis %s' % (jisyo))
    for line in file.readlines():
        line = line[:-1]    # Chomp \n
        try:
            yomi, kanji = string.split(line, ' ')
            JISYO[yomi] = string.split(kanji, '/')[1:-1]      # offset 0, and -1 is ''
            # JISYO[yomi] = kanji      # Old version (No '/' split)
        except:                        # Failed at ;; (comment lines)
            continue

# Variables used while conversion mode is activated.
# Some of these can not correctly be monitored through interactive debugger.
# Why ?
curyomi = ''    # Current yomi -- pronounciation
curcand = []    # Current candidate list
candpos = -1    # Candidate position index starting from 0

# yomipos
def start_henkan(yomi):
    global curyomi, curcand, candpos
    curyomi = yomi
    curcand = JISYO[yomi]
    candpos = 0
    try:    return curcand[candpos]
    except: raise 'NotConverted'

def henkan_forward():
    global candpos
    candpos = candpos + 1
    try:    return curcand[candpos]
    except: raise 'NoMoreCandidate'

def henkan_backward():
    global candpos
    candpos = candpos - 1
    # If candpos < 0 then we should raise FallBackToKana,
    # but python allows negative index, so...
    if candpos < 0: candpos = -1276384   # Guess big enough to raise
    try:    return curcand[candpos]
    except: raise 'FallBackToKana'

def kakutei():
    global curyomi, curcand, candpos
    curyomi = ''
    curcand = []
    candpos = -1

# Just simple converter returns very first candidate only.
def convert(yomi):
    try:    return JISYO[yomi][0]
    except: raise 'NotConverted'

# Test output
def test2(yomi):
    print 'Getting jisyo into memory...',
    getjisyo(myjisyo)
    print 'Done.'
    try:
        print '0: curcand', curcand, 'candpos', candpos
        print start_henkan(yomi)
        print '1: curcand', curcand, 'candpos', candpos
        print henkan_backward()
        print '1a: curcand', curcand, 'candpos', candpos
        print henkan_forward()
        print '2: curcand', curcand, 'candpos', candpos
        print henkan_backward()
        print '3: curcand', curcand, 'candpos', candpos
        print henkan_forward()
        print '4: curcand', curcand, 'candpos', candpos
        kakutei()
        print '5: curcand', curcand, 'candpos', candpos
        
    except 'NotConverted':
        print 'Not converted.'
        
# Test output
def test():
    from sys import *
    
    ## print 'Getting jisyo into memory...',
    getjisyo(myjisyo)
    ## print 'Done.'
    
    if len(sys.argv) < 2:
        for line in stdin.readlines():
        	line = line[:-1]
        	stdout.write("%s\n" % (convert(line)))
    else:
        for i in sys.argv[1:]:
            try:
                print convert(i),
            except 'NotConverted':
               print 'Not converted.'

if __name__ == '__main__':
    test()
    #test2('cpu')


# Benchmark (Just this script, no compile?):
# on Pentium 166MHz+64Mb memory/linux.
# Just skkgetjisyo('/home/sn/.skk-jisyo')::::
# sn@miles:~ $ l .skk-jisyo         (283kb: my own personal dictionary)
#-rw-r--r--   1 sn       root       283176 Jul 23 13:25 .skk-jisyo

#$ time ~/python/wskk-read-dict.py
#1.67user 0.15system 0:02.13elapsed 85%CPU (0avgtext+0avgdata 0maxresident)k
#0inputs+0outputs (0major+0minor)pagefaults 0swaps

# 'わんきょく' is last entry in my dictionay.
# skkgetjisyo('/home/sn/.skk-jisyo'), and thne above test code.
#sn@miles:~/python $ time ~/python/wskk-read-dict.py わんきょく
#湾曲
#1.69user 0.12system 0:02.18elapsed 83%CPU (0avgtext+0avgdata 0maxresident)k
#0inputs+0outputs (0major+0minor)pagefaults 0swaps
