;;
;; lkk-bobo functions. Thu Feb 17 16:58:20 2000
;;

(require 'lkk)

(defun lkk-bogo-profile (sexp count)
  "Eval SEXP COUNT times and returns laptime in \(secdelta . usecdelta\) \
format."
  (let (st en secdelta tem)
    (setq st (current-time))
    (while (>= (setq count (1- count)) 0) (eval sexp))
    (setq en (current-time))
    (cons (setq secdelta (- (nth 1 en) (nth 1 st)))
	  ;; usecdelta part
	  (if (= secdelta 0)
	      (- (nth 2 en) (nth 2 st))
	    (setq tem
		  (+ (nth 2 en) (- 1000000 (nth 2 st))))
	    (if (< tem 1000000)
		tem
	      ;; otherwise, usec part overflowed.
	      (setq secdelta (1+ secdelta))
	      (- tem 1000000))))))

(fset 'lkk-bogo 'lkk-bogo-profile)

(defun lkk-bogo-perf (sexp count)
  "Calls lkk-bogo-profile COUNT times passing SEXP and COUNT, \
and averaged timing list '\(sec . usec\) returned.
The same COUNT is handed to lkk-bogo-profile.  Thus:

\(lkk-bogo-perf '\(bogofunc\) 10\) means calling lkk-bogo-profile
10 times, and each lkk-bogo-profile executes bogofunc x10 times, 
and 10 outputs of lkk-bogo-profile is averaged."
  (garbage-collect)
  (let ((sectotal 0) (usectotal 0) result)
    (mapcar '(lambda (dummy)
	       (setq result (lkk-bogo-profile sexp count)
		     sectotal (+ sectotal (car result))
		     usectotal (+ usectotal (cdr result))))
	    (make-list count 'x))
    (cons (/ sectotal count) (/ usectotal count))))

;; Some warnings are generated on `make install', sorry.
(defun lkk-skk-bench (str &optional arg)
  (require 'skk)
  (or arg (setq arg 1))
  (or skk-mode
      (progn (skk-mode) (skk-mode)))	; turn one once, or str is not inserted.
  (while (> arg 0)
    (insert (format
	     "    skk(%s): %s lkk: %s"
	     skk-version
;;;	       (lkk-bogo-profile
	     (lkk-bogo-perf
	      '(mapcar (lambda (c)
			 (let ((last-command-char c)) (skk-insert 1)))
		       (lkk-string-to-char-list str))
	      1)				 	       
;;;	       (lkk-bogo-profile
	     (lkk-bogo-perf
	      '(mapcar (lambda (c)
			 (let ((last-command-char c)) (lkk-j-input 1)))
		       (lkk-string-to-char-list str))
	      1)))
    (beginning-of-line 1)
    (open-line 1)
    (setq arg (1- arg))))

(defun lkk-bench (str &optional arg)
  (or arg (setq arg 1))
  (while (> arg 0)
    (insert (format
	     "    lkk: %s"
	     (lkk-bogo-perf
	      '(mapcar (lambda (c)
			 (let ((last-command-char c)) (lkk-j-input 1)))
		       (lkk-string-to-char-list str))
	      1)				 	       
	     ))
    (beginning-of-line 1)
    (open-line 1)
    (setq arg (1- arg))))

(defun lkk-macro-j-input (str &optional arg)
  "No profiling. Just input via lkk-j-input."
  (or arg (setq arg 1))
  (while (> arg 0)
    (mapcar (lambda (c)
	      (let ((last-command-char c)) (lkk-j-input 1)))
	    (lkk-string-to-char-list str))
;;;    (beginning-of-line 1)
;;;    (open-line 1)
    (setq arg (1- arg))))

(provide 'lkk-bogo)
