;;; lkk-debug.el --- Bunch of debugger routines.

;;; s.n. Mon Feb 21 13:58:09 2000

;;
;; For authour/developer only.  Requires `pp' package in standard 19.xx & 20.xx.
;;
;; Load this via lkk-load-hook.
;; (setq lkk-load-hook (lambda ()
;;		      (require 'lkk-bogo)
;;		      (require 'lkk-debug)
;;		      ))
;;
;; To popup debugger frame
;; (setq special-display-regexps (cons " \\*\\.*lkk" special-display-regexps))
;; (add-hook 'lkk-debug-hook 'lkk-watch-local-var)
;;

(provide 'lkk-debug)
(require 'lkk)

(require 'pp)

;; with-output-to-temp-buffer can't handle buffer object but just buffer name.
(defsubst lkk-pp-eval-exp (lis &optional buf)
  (with-output-to-temp-buffer (or buf " *lkk-dump*") (pp lis)))

;;(defvar lkk-debug-hook nil "lkk debugger.") ; now is in lkk.el

(defsubst lkk-get-local-var-list ()
  "Return list for pp-eval-expression, or lkk-watch-local-var."
  (let (out)
    (mapcar
     (lambda (c)
       (if (string-match "^lkk" (format "%s" (car c)))
	   (cond
;;; Does not exists anymore. 
;;;	    ;; This list is too long. Not appended.
;;;	    ((string= 'lkk-last-assoc-struct (symbol-name (car c)))
;;;	     (setq out
;;;		   (append
;;;		    out
;;;		    (list
;;;		     (list
;;;		      'lkk-last-assoc-struct+++SUMMARY
;;;		      (list
;;;		       (nth 0 lkk-last-assoc-struct)
;;;		       (lkk-take5 (nth 1 lkk-last-assoc-struct) 5 nil)
;;;		       (lkk-take5 (nth 2 lkk-last-assoc-struct) 5 nil))))))
;;;	     )
	    (t (setq out (append out (list c)))))
	 ))
     (buffer-local-variables)) out))

(defsubst lkk-local-var-list-with-header ()
  (list (cons 'this-command this-command)
	(cons 'point (point))
	(lkk-get-local-var-list)))

;;
;; lkk-watch-local-var -- a function to be registered to lkk-debug-hook
;; See also lkk-debugger command below.
;;

(defun lkk-watch-local-var ()
  "Use pp-eval to pop debugger buffer."
  (lkk-pp-eval-exp
   (lkk-local-var-list-with-header)))

;; ARG unused yet.
(defun lkk-watch (&optional arg)
  "Interactive equivalence to lkk-watch-local-var function.
Output goes to another buffer \" *lkk-watch*\". Normally
lkk-debug-hook is run just after entry to a command/function, and
lkk-watch-local-var dumps the status at the timing. This
lkk-watch() can be called manually by M-x to display the status
after the command/function returned."
  (interactive "P")
  (let* ((bufnm " *lkk-watch*"))	; don't have to kill/erase the buffer.
    (lkk-pp-eval-exp
     (lkk-local-var-list-with-header) bufnm)))

;; (defun lkk-watch-old (arg)
;;  "Old version of lkk-watch."
;;  (interactive "P")
;;  (pp-eval-expression (quote (lkk-local-var-list-with-header))))

;; MISC.
(defun lkk-take5 (list &optional arg reverse)
  "Take top 5 of LIST as list and return it.
Opt 2nd ARG as num of the top. Opt 3rd arg REVERSE specifies reversing
befor the `take'."
  (let ((list (if reverse (reverse list) list))
	(arg (or arg 5))
	(i 0) (l))
    (while (< i arg)
      (if (nth i list)
	  (setq l (append l (list (nth i list)))))
      (setq i (1+ i))
      ) l))

;;
;; `interactive' commands
;;

(defun lkk-debugger (arg)
  "Turn on lkk debugger with explicit non-nil ARG, or turn it off without ARG."
  (interactive "P")
  (cond ((and arg (> (prefix-numeric-value arg) 0))
	 (add-hook 'lkk-debug-hook 'lkk-watch-local-var))
	(t (remove-hook 'lkk-debug-hook 'lkk-watch-local-var)
	   (setq arg nil)))
  (message "lkk debug turned %s." (if arg "on" "off")) )

(defun lkk-dump-local-var ()
  (interactive)
  (lkk-pp-eval-exp
   (let (lkk-local)
     (mapcar (lambda (c)
	       (if (local-variable-if-set-p c)
		   (setq lkk-local (cons (cons c (eval c)) lkk-local))))
	     (apropos-internal "lkk" 'boundp))
     lkk-local)))

(defun lkk-dump-global-var ()
  (interactive)
  (lkk-pp-eval-exp
   (let (lkk-global)
     (mapcar (lambda (c)
	       (if (null (local-variable-if-set-p c))
		   (setq lkk-global (cons (cons c (eval c)) lkk-global))))
	     (apropos-internal "lkk" 'boundp))
     lkk-global)))

(defun lkk-init (&optional arg)
  "Resets lkk status according to current lkk-dict-method value.
Can be called if something goes wrong.
Optional ARG forces dictionary dump, and clears its buffer, or alist."
  (interactive "P")

  (lkk-setup-r2k-internal)

  (cond 
   ((or arg (y-or-n-p "Dump & Clear dictionary as wall ? "))
    (lkk-dump-dict)
    (lkk-select-right-dict-method 'force))
   (t (lkk-select-right-dict-method nil)))

  (lkk-mode nil)
  (lkk-mode t))
