;;; lkk-isearch.el --- A bare lkk support for isearch.

;; Fri Mar  3 02:03:13 2000 (c) S.Namba <sn@asahi-net.email.ne.jp>
;; http://www.asahi-net.or.jp/~pi9s-nnb/

;;
;; Just provide a lkk-mode-already-turned-on minibuffer.
;; Good enough (at least for me).
;;
;; This is very robust, simple, neat, and clean.
;; Even works if (featurep 'skk-isearch), provided that skk-mode is just
;; turned off. This may be a nice backup for me if something wrong in
;; skk-iseach -- which sometimes raises "Args out of range..." and all
;; Kanji search can't be done.
;;

;;
;; Most of codes here are based on codes in isearch-fep-mode() (called
;; by isearch-fep-string()), and isearch-fep-read-string of
;; isearch-ext.el in mule-2.3@19.34.
;;

;; GPL2 should be applied to the emacs-derived functions and variables.
;;
;; As for other code I wrote from scratch (not so many here, though),
;; I just don't put any copying/copyright notice other than this:
;; Copyright (C) 2000 Seiichi Namba <sn@asahi-net.email.ne.jp>
;; There is of course no warranty.
;; Thu Apr 13 18:22:26 2000
;; s.n.

(provide 'lkk-isearch)

(require 'lkk)

(defvar lkk-isearch-switch "\C-j" "lkk turn on key in isearch.")


;; Based on isearch-fep-mode in isearch-ext.el
;;

(defun isearch-lkk-mode ()
  "Open an lkk-turned-on minibuffer while i-search.
Initial mode is set to the same mode as that of origil current-buffer."
  (interactive)
  (let (str (isearch-fep-prompt "LKK: "))
      (message "%s%s" isearch-fep-prompt (isearch-message))
      (setq str (isearch-lkk-read-string isearch-fep-prompt))
      (isearch-process-search-string str str)))

;; Based on isearch-fep-read-string in isearch-ext.el
;;
(defun isearch-lkk-read-string (prompt)
  (let* ((minibuffer-local-map (cons 'keymap minibuffer-local-map))
	 (lkk-initial-mode)
	 (minibuffer-setup-hook)
	 (minibuffer-exit-hook)
	 (lkk-in-read-string t)) ; maybe lkk-sit-back sees it.
    (cond
     (lkk-mode
      (setq
       ;; Determine the minibuffer's lkk submode according to current buf.
       lkk-initial-mode (catch 'hit
			  (mapcar
			   (lambda (m)
			     (if (and (eval m)
				      ;; J-mode is not suitable
				      (not (eq m 'lkk-J-mode)))
				 (throw 'hit m)
			       ))
			   lkk-valid-minor-minor-modes)
			  nil))
      ;; Turn on lkk for the incomming minibuffer only.
      ;;lkk-mode sees the `initial' mode.
      (setq minibuffer-setup-hook (lambda () (lkk-mode nil)))
      ))
    ;; Some program overwrites "\C-m"'s default binding.
    (define-key minibuffer-local-map "\C-m" 'lkk-exit-minibuffer)
    (define-key minibuffer-local-map "\C-r"
      ;; Wrong, no such function in E19.34/20.5
      ;;   'exit-minibuffer-and-isearch-backward

      ;; Not theirs, but mine:
      'lkk-isearch-reverse-exit-minibuffer
      )
    (define-key minibuffer-local-map "\C-s"
      ;; This is okay, but just for code integrity.
      ;;'exit-minibuffer
      'lkk-isearch-forward-exit-minibuffer
      )
    (condition-case condition
	;; isearch.el of GNU Emacs 19.28 used overriding-local-map.
	;; isearch.el of GNU Emacs 19.29, 19.30 (which is latest at
	;; this point) uses overriding-terminal-local-map.  Disable
	;; isearch keymap with binding it to nil.  We bind the map
	;; here so that isearch-abort (actually, by isearch-done)
	;; can set it to nil.
	;;
	;; Version detection added for lkk-isearch.el s.n.
	(cond ((string< "19.28" emacs-version)
	       (let ((overriding-terminal-local-map nil))
		 (read-from-minibuffer
		  (concat prompt (isearch-message)))))
	      (t			; perhaps vanilla mule-2.3@19.28
	       (let ((overriding-local-map nil))
		 (read-from-minibuffer
		  (concat prompt (isearch-message))))
	       ))
      (quit
       ;; Also, this isearch-abort should be called in original
       ;; buffer, because the variable which always has buffer-local
       ;; binding (at least isearch-mode is such variable) should be
       ;; bound in right environment.
       (isearch-abort)))))

;;
;; isearch-reverse-exit-minibuffer from emacs-19.34 and 20.5
;; can't be used for our purpose. It is conceived to be used
;; from isearch-edit-string and it modifies isearch-new-forward.
;;   'isearch-forward-exit-minibuffer
;;

(defun lkk-isearch-forward-exit-minibuffer ()
  (interactive)
  ;; (setq isearch-new-forward t) ;; original
  (setq isearch-forward t)
  (lkk-exit-minibuffer))

(defun lkk-isearch-reverse-exit-minibuffer ()
  (interactive)
  ;; (setq isearch-new-forward nil) ;; original
  (setq isearch-forward nil)
  (lkk-exit-minibuffer))

(if (featurep 'mule)
    (add-hook
     'isearch-mode-hook 
     (if isearch-mode-map		; Needed ?
	 (lambda ()
	   (define-key isearch-mode-map "\C-j" 'isearch-lkk-mode)))))

;;; Old version almost the copy of isearch-fep-mode().
;;; (defun isearch-lkk-mode-old ()
;;;   (interactive)
;;;   (let ( ;; (command this-command)
;;; 	(isearch-fep-mode t)
;;; 	table str)
;;;     (while isearch-fep-mode
;;;       ;;(setq table (assq command isearch-fep-table))
;;;       ;;(setq isearch-fep-prompt (funcall (car (cdr table))))
;;;       (setq isearch-fep-prompt "LKK: ")
;;; 
;;;       (message "%s%s" isearch-fep-prompt (isearch-message))
;;;       (if;;(eq command 'isearch-fep-string)    ;; \C-k
;;; 	  t
;;; 	  (progn
;;; 	    ()
;;; 	    ;;(setq str (funcall (nth 2 table)))
;;; 	    (setq str (funcall 'isearch-lkk-read-string))
;;; 
;;; 	    (setq isearch-fep-mode nil)
;;; 	    (isearch-process-search-string str str))
;;; The latter part of the `if' in isearch-fep-mode() is all deleted here.
;;; See isearch-ext.el for original EGG, Canna etc. support.
;;; 	))))
