;;; lkk-jchunk.el --- move/kill commands for japanese

;; Tue Apr 18 14:42:06 2000

;; (require 'lkk) ;; This file does not require it to load, or exec commands
;; with just M-x.

(provide 'lkk-jchunk)

;;
;; For the codes here I wrote from scratch,
;; I just don't put any copying/copyright notice other than this:
;; Copyright (C) 2000 Seiichi Namba <sn@asahi-net.email.ne.jp>
;; There is of course no warranty.
;; Thu Apr 13 18:22:26 2000
;; s.n.
;;

;; Now perform kill with non-nil arg.
(defun lkk-kill-backward-j-chunk (arg &optional justmove)
  "Kill backward a chunk of japanese with a lousy syntax matching. 
Maybe this is incorrect in emacs `syntax'.
Numeric ARG specifies the times of repeat.
Optional 2nd arg JUSTMOVE just moves to the chunk border.
Fakes as if `this-command' is 'kill-region."
  (interactive "*p")
  (let* ((hregex "[$B$!(B-$B$s!<(B]")
	 (kregex "[$B%!(B-$B%v!<(B]")
	 (Nregex "[$B#0(B-$B#9(B]")
	 (pregex "[$B!$!%(B,.$B!"!#!!(B \t\n]")
	 (Aregex "[$B#A(B-$B#Z#a(B-$B#z(B]")
	 (aregex "[-A-Za-z_0-9]")
	 (Kregex "[^$B$"(B-$B$s%!(B-$B%v!<!$!%(B,.$B!"!#!!(B \t\n]") ; dustbin
	 (negative (< arg 0))
	 (fun0 (if negative 'following-char 'preceding-char))
	 (fun1 (if negative 'forward-char 'backward-char))
	 (fun2 (if negative 'eobp 'bobp))
	 (fun3 (if justmove (lambda (p0 p1) (goto-char p1)) 'kill-region))
	 (getregex (lambda (ch) (cond
				 ((string-match hregex ch) hregex)
				 ((string-match kregex ch) kregex)
				 ((string-match Nregex ch) Nregex)
				 ((string-match pregex ch) pregex)
				 ((string-match Aregex ch) Aregex)
				 ((string-match aregex ch) aregex)
				 ;; This is wildest, so do it at last
				 ((string-match Kregex ch) Kregex)
				 (t nil))))
	 (regex (funcall getregex (char-to-string (funcall fun0))))
	 (repeating nil))
    (cond ((null justmove) (setq this-command 'kill-region)))
    (cond
     (regex
      (if negative (setq arg (- arg)))
      (while (> arg 0)
	(if repeating
	    (append-next-kill))
	(funcall fun3
	 ;; Giving the points the reverse order on bwd kill makes
	 ;; correct the order of kill-ring-append. Is this intended
	 ;; emacs feature ?
	 (point)
	 (save-excursion
	   (while
	       (and (null (funcall fun2))
		    (string-match regex (char-to-string (funcall fun0))))
	     (funcall fun1 1))
	   (point)))
	(setq regex (funcall getregex (char-to-string (funcall fun0)))
	      repeating t
	      arg (1- arg)))))
    ))

(defun lkk-kill-forward-j-chunk (arg &optional justmove)
  "Calls `lkk-kill-backward-j-chunk' with negative arg.
See `lkk-kill-backward-j-chunk'."
  (interactive "*p")
  (lkk-kill-backward-j-chunk (- arg) justmove))

(defun lkk-move-backward-j-chunk (arg)
  "Calls `lkk-kill-backward-j-chunk' with negative arg.
See `lkk-kill-backward-j-chunk'."
  (interactive "p")
  (lkk-kill-backward-j-chunk arg 'justmove))

(defun lkk-move-forward-j-chunk (arg)
  "Calls `lkk-kill-backward-j-chunk' with negative arg.
See `lkk-kill-backward-j-chunk'."
  (interactive "p")
  (lkk-kill-backward-j-chunk (- arg) 'justmove))

(defun lkk-transpose-j-chunk ()
  "Lousy, no error check here."
  (interactive)
  (let ((kill-ring))
    (lkk-kill-backward-j-chunk 1)
    (lkk-move-forward-j-chunk 1)
      (yank)))

(defun lkk-mark-j-chunk (arg)
  "Set mark ARG j-chunk's from point.
The place mark goes is the same place M-C-f would
move to with the same argument.
See also `lkk-kill-backward-j-chunk'."
  (interactive "p")
  (set-mark
   (save-excursion
     (lkk-move-forward-j-chunk arg)
     (point))))

(defun lkk-kill-ring-save-j-chunk (arg)
  "Set mark ARG j-chunk's from point, then copy region as kill.
See also `lkk-kill-backward-j-chunk'."
  (interactive "p")
  (lkk-mark-j-chunk arg)
  (kill-ring-save (region-beginning) (region-end)))

;; Map to keys if lkk is featured.

(cond ((and (featurep 'lkk) lkk-j-mode-map)
       ;; Some misc edit command
       (define-key lkk-j-mode-map "\M-H" 'lkk-kill-backward-j-chunk)
       (define-key lkk-j-mode-map "\M-D" 'lkk-kill-forward-j-chunk)
       (define-key lkk-j-mode-map "\M-B" 'lkk-move-backward-j-chunk)
       (define-key lkk-j-mode-map "\M-F" 'lkk-move-forward-j-chunk)
       (define-key lkk-j-mode-map "\M-T" 'lkk-transpose-j-chunk)
       (define-key lkk-j-mode-map "\M-W" 'lkk-mark-j-chunk)
       (define-key lkk-j-mode-map "\M-\C-w" 'lkk-kill-ring-save-j-chunk)
;;;    (define-key lkk-j-mode-map "\C-x\177" 'lkk-kill-backward-j-chunk)
       ))

(cond ((and (featurep 'lkk) lkk-L-mode-map)
       (define-key lkk-L-mode-map "\M-H" 'lkk-kill-backward-j-chunk)
       (define-key lkk-L-mode-map "\M-D" 'lkk-kill-forward-j-chunk)
       (define-key lkk-L-mode-map "\M-B" 'lkk-move-backward-j-chunk)
       (define-key lkk-L-mode-map "\M-F" 'lkk-move-forward-j-chunk)
       (define-key lkk-L-mode-map "\M-T" 'lkk-transpose-j-chunk)
       (define-key lkk-L-mode-map "\M-W" 'lkk-mark-j-chunk)
       (define-key lkk-L-mode-map "\M-\C-w" 'lkk-kill-ring-save-j-chunk)
;;;    (define-key lkk-L-mode-map "\C-x\177" 'lkk-kill-backward-j-chunk)
       ))
