;;; lkk-lkk-dichead.el --- Research on a dictionary as a big big alist.
;;; Thu Feb 24 22:38:54 2000 s.n.

;; The Dictionary Interface Layer should have only one entry point:
;; The function lkk-dict-search():
;;
;;   (defun lkk-dict-search (yomi &optional loose)
;;
;; which reads the dictionary if needed.
;; Currently lkk-dict-search is fset of lkk-dict-alist-search, or
;; lkk-dict-buffer-search.

;; On running lkk, lkk.el will explicitly load this program if needed.
;; (require 'lkk)

(provide 'lkk-lkk-dichead)

;; Must do nothing unless suitable method is defined.
(cond ((eq lkk-dict-method 'lkk-dict-buffer)
       (fset 'lkk-dict-search 'lkk-dict-buffer-search)
       (fset 'lkk-dict-update 'lkk-dict-buffer-update))
      ((eq lkk-dict-method 'lkk-dict-alist)
       (fset 'lkk-dict-search 'lkk-dict-alist-search)
       (fset 'lkk-dict-update 'lkk-dict-alist-update))
      (t				; else NOP
       ))

;;;
;;; Part 1 `lkk-dict-alist' method
;;;

(defvar lkk-big-dict-okuri nil "`okuri ari' lkk dictionary big alist.")
(defvar lkk-big-dict nil "`okuri nasi' lkk dictionary big alist.")
(defvar lkk-big-dict-top-record nil
  "Used to test the big alist is dirty \(unsaved\) or not.")

;;
;; lkk-get-dict-alist() --- Big dictionary loader.
;;
;; Builds two dictionary alist `lkk-big-dict-okuri' and  `lkk-big-dict' 
;; out of file lkk-dict.
;;

;; Some benchmark on reading whole part as one dictionary.
;; (setq lkk-big-dic-head (lkk-get-dict-alist "./lkk-dict.new"))
;; (length lkk-big-dic-head)
;; 15631                         ;;  15631 lines (entries)
;; 
;; (lkk-bogo-perf
;;  '(let* ((tem (reverse lkk-big-dic-head)))
;;     (list (nth 3 tem)
;; 	  (nth 2 tem)
;; 	  (nth 1 tem)
;; 	  (nth 0 tem))) 1)
;; (0 . 5201)  ;; Took 5ms to take 4 ent at last on Celeron 300A + 128Mb
;; (0 . 5193)
;; (0 . 5210)
;; (list (length lkk-big-dict-okuri) (length lkk-big-dict))
;;  ==> (1974 13658)
;;
;;(lkk-bogo-profile '(lkk-get-dict-alist) 1)
;;  ==> (1 . 1259860) ;; perhaps this means 1.259 sec.

;; 
;; Mon Feb 28 16:12:11 2000
;; with-temp-buffer() is lacked in emacs-19.xx.
;;
;; Perhaps (require 'apel) enables these, but I won't do so because lkk is
;; a sort of `system program' and I think depending on such a library
;; for casual programmers here in this serious program is SUICIDAL.
;;

(defun lkk-get-dict-alist (&optional dicfile)
  "Setup lkk-big-dict-okuri and lkk-big-dict from lkk-dict FILE."
  (let ((dicfile (or dicfile lkk-dict)) (case-fold-search nil)
	(tmpbuf (generate-new-buffer " *lkk-temp*"))
	(inhibit-quit t))
    (setq lkk-okuri-ari-dict-top nil	; unused in this method.
	  lkk-okuri-nasi-dict-top nil)

    (save-excursion
      (set-buffer tmpbuf) (widen) (erase-buffer)
      (condition-case nil
	  (progn
	    (message "Reading lkk dictionary %s ..." dicfile)
	    (lkk-insert-file-contents dicfile))
	(error
	 (error "Can't read lkk dictionary %s." dicfile)))
      (message "Reading lkk dictionary %s ...done" dicfile) (sit-for 0)

      (goto-char (point-min))

      (let ((case-fold-search t))
	(lkk-dict-get-dict-version)
	;; Just use okuri-nasi header as delimiter of the two part.
	(save-excursion (insert "(")
			(or
			 (search-forward "\n;; okuri-nasi" (point-max) t)
			 (error "%s may not be lkk dictionary." dicfile))
			(forward-line 1)
			(insert ")(")
			(goto-char (point-max))
			(insert ")")))

      (condition-case nil
	  (setq lkk-big-dict-okuri (read (current-buffer))
		lkk-big-dict       (read (current-buffer))
		lkk-big-dict-top-record (car lkk-big-dict)) ; keep ptr of current top.
	(error (error "Can't read the dictionary. Destroyed %s ?" lkk-dict)))
      )
    (kill-buffer tmpbuf)
    t ))

(defun lkk-list-dump (list)
  "Pretty print LIST on current buffer with prin1."
  (mapcar (lambda (c)
	    (prin1 c (current-buffer))
	    (insert ?\n)
	    ) list ))

;; (fset 'lkk-dump-big-dict 'lkk-dict-alist-dump-dict)
(defun lkk-dict-alist-dump-dict (&optional justimage)
  "Dump lkk-big-dict-okuri and lkk-big-dict to lkk-dict-dump-file
Temporary user command until real dictionary upadte system is completed."
  (cond
   ((or (null (and lkk-big-dict-okuri lkk-big-dict))
	;; Is it a good test to decide save or not ?
	(and (not justimage) lkk-big-dict-top-record
	     (eq lkk-big-dict-top-record (car lkk-big-dict))))
    nil)
   (t
    (let* ((buf (generate-new-buffer " *lkk-temp*"))
	   (hdrstr ";; %s VERSION %s\n")
	   (oahdr ";; okuri-ari\n")
	   (onhdr ";; okuri-nasi\n")
	   (inhibit-quit t))
      ;; with-temp-file and other gadgets are lacked in E19.xx.
      (save-excursion
	(set-buffer buf) (widen) (erase-buffer)
      
	;; FIXME -- Must use the same coding system on reading.
	(lkk-set-file-coding-system (if (< (string-to-int emacs-version) 20)
					'*euc-japan*unix 'euc-japan-unix))

	(insert (format hdrstr lkk-dict-header lkk-version))
	(or justimage (message "Writing okuri-ari..."))
	(insert oahdr)
	(lkk-list-dump lkk-big-dict-okuri)
	(or justimage (message "Writing okuri-nasi..."))
	(insert onhdr)
	(lkk-list-dump lkk-big-dict)
	(or justimage (message "Writing okuri-nasi...Done"))

	(cond
	 (justimage (current-buffer))
	 (t (condition-case nil
		(prog2
		    (message "lkk dictionary dumping...")
		    (write-region (point-min) (point)
				  (expand-file-name lkk-dict-dump-file))
		  ;; (set-buffer-modified-p nil) ; unnecessary.
		  (message "lkk dictionary dumped as %s" lkk-dict-dump-file)
		  )
	      (error (error "Dump onto %s faild." lkk-dict-dump-file)))
	    (kill-buffer buf)
	    ;; Record new top ptr.
	    (setq lkk-big-dict-top-record (car lkk-big-dict))
	    )))
      ))))

(lkk-deflocalvar lkk-dict-alist-the-latest-hit
  "lkk Internal use only.  Contains the latest hit record of dictionary search.
Set by function lkk-dict-alist-search.
Will be used to feed delq function, to old record.")

;;
;; lkk-dict-alist-search() -- Dict search function for lkk-dict-alist method.
;;

;; Would-be alias of lkk-dict-search
(defun lkk-dict-alist-search (yomi &optional loose)
  "Accepts YOMI and LOOSE search flag, and returns a *copy* of dictionary record
as list or nil if failed.

The real hit record is bound to lkk-dict-alist-the-latest-hit for
coming delq function call on dictionary update, and thus you don't
have to worry about destroying the `delq' data in upper-layer programming."
    (or (and lkk-big-dict-okuri lkk-big-dict)
	(lkk-get-dict-alist))
    (let ((fun (if loose 'lkk-dict-alist-loose-search 'assoc)))
      (setq lkk-dict-alist-the-latest-hit
	    (funcall fun yomi (if lkk-okuri lkk-big-dict-okuri lkk-big-dict)))
      (copy-list lkk-dict-alist-the-latest-hit)))

;;
;; lkk-dict-alist-loose-search() -- mostly for completion/elettronico
;; Perhaps `while' is faster...
;;      (mapcar (lambda (en)
;;		(if (string-match (format "^%s" yomi) (car en))
;;		    (throw 'hit en))
;;		) dic )
;;(list
;; (lkk-bogo-profile
;;  '(lkk-dict-alist-loose-search "󤭤" lkk-big-dict) 1)
;; (lkk-bogo-profile ; mapcar version above
;;  '(lkk-dict-alist-loose-search-mapcar "󤭤" lkk-big-dict) 1) )	
;;((6 . 1649227) (7 . 1165556)) ; 486DX 50MHz 14952 lines (whole) dict.
;;((7 . 665705) (8 . 374815))
;; 
;; However, even with `while', scan by string-match() is quite slow.
;;

(defun lkk-dict-alist-loose-search (yomi dict)
  "Scan big alist DICT with string-match function."
;;;  (lkk-dict-alist-assoc yomi dict 'loose)
  (let (entry)
    (catch 'hit
      (while dict
	(if ;; (string-match (format "^%s" yomi) (car (setq entry (car dict))))
	    (lkk-string-match yomi (car (setq entry (car dict)))
			      'loose)
	    (throw 'hit entry))
	(setq dict (cdr dict)) ))))

;;
;; lkk-dict-alist-update() -- Dictionary UPDATE routine
;;
;; Just erase the-latest-hit record pointed by variable
;; lkk-dict-alist-the-latest-hit in the dictionary with delq(), and
;; put a new record updated by upper layer on the top of it by cons().
;;

;; Would-be alias of lkk-dict-update()
(defun lkk-dict-alist-update (newrecord &optional justadd)
  "Update dictionary on lkk-sit-back. Accept maybe-modified NEWRECORD.
Optional 2nd arg JUSTADD specifies just adding a brand new record."
  (cond
   ;; Just add brand new record: should be first
   (justadd
    (let* ((inhibit-quit t)
	   (dictsym (if lkk-okuri
			'lkk-big-dict-okuri
		      'lkk-big-dict))
	   (dict (eval dictsym))
	   ;;(newcdr (delq lkk-dict-alist-the-latest-hit dict))
	   )
      (cond
       ((> (length newrecord) 1)
	(set dictsym (cons newrecord dict))))
      ;; lkk-dict-update() may be repeatedly called without
      ;; new search, especially on lkk-purge-it(), so we have to
      ;; re-calibrate lkk-dict-alist-the-latest-hit.
      (setq lkk-dict-alist-the-latest-hit (car (eval dictsym)))
      )
    )
   ;; means conversion failed
   ((null (or lkk-current-record lkk-kanji-index))
    t)
   ;; yomi is displayed. Success gracefully.
   ((and (integerp lkk-kanji-index)
	 (= lkk-kanji-index 0))
    t)

   ;; Update entity,
   (lkk-dict-alist-the-latest-hit
    (let* ((inhibit-quit t)
	   (dictsym (if lkk-okuri
			'lkk-big-dict-okuri
		      'lkk-big-dict))
	   (dict (eval dictsym))
	   (newcdr (delq lkk-dict-alist-the-latest-hit dict)))
      (cond
       ((> (length newrecord) 1)
	(set dictsym (cons newrecord newcdr)))
       ;; If record has just seed, just delete and no add-record
       (t
	(set dictsym newcdr)))
      ;; lkk-dict-update() may be repeatedly called without
      ;; new search, especially on lkk-purge-it(), so we have to
      ;; re-calibrate lkk-dict-alist-the-latest-hit.
      (setq lkk-dict-alist-the-latest-hit (car (eval dictsym)))
      ))
   (t (message "How did'ya get there ?  Maybe buggy call for %s"
	       (symbol-function 'lkk-dict-update)))
   ))

;;;
;;; Part 2 `lkk-dict-buffer' method
;;;

(defsubst lkk-dict-buffer-get-current-record ()
  "Reads sexp after point. Almost equivalent to \(read \(current-buffer\)\)."
  (let* (line entry)
    ;; I don't rely on that I am on BOL (I can see the line 5 lines
    ;; above now, but I am not sure if I could even 5 days later.
    (forward-line 0)

;;; If read() is preferred, must narrow to this whole line -- otherwise
;;; read() may not raise error until it hits the bottom of buffer.
;;;    (save-excursion			; May look good if sw-buffer to dictbuf.
;;;      (condition-case nil
;;;	  (setq entry (read (current-buffer))) ; Quite easy, isn't it ?
;;;	(error (error "Can't read this entry: %s"
;;;		      (lkk-get-current-line)))) )

    ;; Or just get this whole line then read-from-string() like this ?
    (setq line (lkk-get-current-line)
	  entry (condition-case nil
		    (car (read-from-string line))
		  (error (error "Can't read this entry: %s" line
				))) )
    ;; but how does this error occur ?
    (if (listp entry) entry (error "lkk dictionary entry should be list: %s"
				   (lkk-get-current-line)))))

(lkk-deflocalvar lkk-dict-buffer-the-latest-hit-point
  "lkk Internal use only.  Contains the latest hit record of dictionary search.
Set by function lkk-dict-alist-search.
Will be used to feed delq function, to old record.")

;; Would-be alias of lkk-dict-search
(defun lkk-dict-buffer-search (yomi &optional loose)
  "Search YOMI and returns dictionary entry lists or nil if failed.
Accepts YOMI and LOOSE search flag."
  (let ((dtop (if lkk-okuri 'lkk-okuri-ari-dict-top 'lkk-okuri-nasi-dict-top))
	(target (concat "(\"" yomi (if loose "" "\" ")))
	recordpt hitrecord)
    (save-excursion
      (set-buffer (lkk-dict-get-dict))	; in lkk.el
      (setq dtop (eval dtop))
      ;; Currently, okuri is not supported in this skk-dichead.
      (goto-char dtop)
      ;; yomi is enclosed in paren and double quoted.
      (cond ((or (looking-at target)
		 (search-forward (concat "\n" target) (point-max) t))
	     (forward-line 0)
	     (setq recordpt (point-marker)
		   hitrecord (lkk-dict-buffer-get-current-record)))
	    (t nil)))
    (setq lkk-dict-buffer-the-latest-hit-point recordpt)
    hitrecord))

;;
;; lkk-dict-buffer-update() -- Dictionary UPDATE routine
;;
;; Just erase the-latest-hit record pointed by variable
;; lkk-dict-buffer-the-latest-hit-point in the dictionary with 
;; lkk-dict-buffer-delete-record() and
;; put a new record updated by upper layer on the top of it by 
;; lkk-dict-buffer-add-record().
;;

;; Would-be alias of lkk-dict-update()
(defun lkk-dict-buffer-update (newrecord &optional justadd)
  "Update dictionary on lkk-sit-back. Accept maybe-modified NEWRECORD.
Optional 2nd arg JUSTADD specifies just adding a brand new record."
  (cond
   ;; Just add brand new record: should be first
   ;; Add brand new record.
   (justadd
    (let ((inhibit-quit t)
	  (dtop (if lkk-okuri
		    lkk-okuri-ari-dict-top
		  lkk-okuri-nasi-dict-top)))
      (lkk-dict-buffer-update0 dtop nil	; This nil specifies just add on top.
			       newrecord)
      (setq lkk-dict-buffer-the-latest-hit-point dtop)
    ))

   ((null (or lkk-current-record lkk-kanji-index))
    t)					; means conversion failed
   ((and (integerp lkk-kanji-index)
	 (= lkk-kanji-index 0))
    t)					; yomi is displayed. Success gracefully.
   (lkk-dict-buffer-the-latest-hit-point
    (let ((inhibit-quit t)
	  (recordpt lkk-dict-buffer-the-latest-hit-point)
	  (dtop (if lkk-okuri
		    lkk-okuri-ari-dict-top
		  lkk-okuri-nasi-dict-top)))
      (lkk-dict-buffer-update0 dtop recordpt newrecord)
      ;; lkk-dict-update() may be repeatedly called without
      ;; new search, especially on lkk-purge-it(), so we have to
      ;; re-calibrate lkk-dict-buffer-the-latest-hit-point.
      (setq lkk-dict-buffer-the-latest-hit-point dtop)
      ))
   (t (message "How did'ya get there ?  Maybe buggy call for %s"
	       (symbol-function 'lkk-dict-update)))
   ))

(defun lkk-dict-buffer-update0 (dtop recordpt newrecord)
  "Subroutine called by lkk-dict-buffer-update in lkk-dict-buffer method."
  (let ((recordstr (prin1-to-string newrecord)))
    (save-excursion
      (set-buffer (lkk-dict-get-dict))	; Reload MUST not occur here (really ?)
      (if recordpt (lkk-dict-buffer-delete-record recordpt)) ; Don't delete, just add.
      ;; If record has just seed, just delete and no add-record
      (if (> (length newrecord) 1)
	  (lkk-dict-buffer-add-record dtop recordstr)))))

(defun lkk-dict-buffer-delete-record (pt)
  "Move to record marker PT, and delete the record."
  (save-excursion (goto-char pt)
		  (delete-region (save-excursion (forward-line 0) (point))
				 (save-excursion (forward-line 1) (point)))
		  ))

(defun lkk-dict-buffer-add-record (pt record)
  "Move to record marker PT, and add one RECORD."
  (save-excursion (goto-char pt)
		  (forward-line 0)
		  (insert record ?\n)
		  ))
