;;; lkk-skkserv.el --- skkserv interface for lkk

(provide 'lkk-skkserv)

;;
;; Most of routines here are borrowed from Classic skk5.30
;;
;; GPL2 should be applied to the skk-derived functions and variables.
;;
;; As for other code I wrote from scratch,
;; I just don't put any copying/copyright notice other than this:
;; Copyright (C) 2000 Seiichi Namba <sn@asahi-net.email.ne.jp>
;; There is of course no warranty.
;; Thu Apr 13 18:22:26 2000
;; s.n.
;;

;;
;; Routine Collections from Classic skk
;;
;; Mostly the same as in original skk5.30, but small fixes are
;; over there to interface with lkk, or make it work in modern emacsen.
;; Some of variable and immediate values are also modified to duck collision
;; on skk. Eg., server buffer name is " *.lkk-skkserv*".
;; s.n. Wed Mar  8 00:15:03 2000
;;

;; skk server related variables

(defvar j-local-length 0)
(make-variable-buffer-local 'j-local-length)

(defvar j-server-called nil)
(make-variable-buffer-local 'j-server-called)

;; The following variable was introduced at the suggestion of Masakazu
;; Takahashi, 1991.10.1.

(defvar j-remote-shell-program
  (or (getenv "REMOTESHELL")
      (cond
       ((eq system-type 'berkeley-unix) "/usr/ucb/rsh")
       ((eq system-type 'usg-unix-v)
	(if (file-exists-p "/usr/ucb/remsh") "/usr/ucb/remsh" "/bin/rsh"))
       ((eq system-type 'hpux) "/usr/bin/remsh")
       ((eq system-type 'EWS-UX/V) "/usr/ucb/remsh")
       ((eq system-type 'pcux) "/usr/bin/rcmd")
       (t "rsh"))))

;;@; Names changed, and made defvar. Seem to be shared in current skk still.
;; s.n.
(defvar lkk-skk-server-host (or (getenv "SKKSERVER") "localhost")
  "*Name of the host running SKK Jisyo Server")

(defvar lkk-skk-serv (or (getenv "SKKSERV")
			 "/usr/local/libexec/skkserv")
  "*Name of the SKK Jisyo Server program")

(defconst lkk-skk-portnum nil
  "*If the value is non-nil, it must be the PORTNUM defined in skkserv.h")

(defvar lkk-skk-report-server-response nil
  "*When SKK sends a request to SKK SERVER, SKK waits for the process output
by repeatedly executing the accept-process-output function.  If this constant
is t, then SKK will report the number of times the function is called.")

;; skk server related functions

;;
;; j-open-server() -- Ring server via NW stream.
;;
;; Modifications on the original Classic:
;;   * Added coding system fiddle for modern emacsen.
;; s.n. Wed Mar  8 00:15:03 2000
;;

(defun j-open-server ()
  "Establish connection with skk server."
  (or (get-process "lkk-skkservd")
      (let ((succ nil))
	(get-buffer-create " *.lkk-skkserv*")
	(if (j-open-server-subr)
	    (setq succ t)
	  (progn
	    (let ((count 7))
	      (while (> count 0)
		(message "SKK SERVER is not active, I will activate it%s"
			 (make-string count ?.))
		(if (or (string= lkk-skk-server-host "localhost")
			(string= lkk-skk-server-host (system-name)))
		    ;; server host is local machine
		    (call-process lkk-skk-serv nil 0 nil)
		  (call-process
		   j-remote-shell-program nil 0 nil
		   lkk-skk-server-host lkk-skk-serv))
		(sit-for 3)
		(if (j-open-server-subr)
		    (setq succ t
			  count 0)
		  (setq count (1- count)))))
	    (if succ
		(progn
		  (message "SKK SERVER is active now.")
		  (sit-for 1)))))
	(if succ
	    ;; Added corresponding coding system fiddle for emacs-19/20/18
	    ;; s.n. Wed Mar  8 02:04:46 2000
	    (let ((ver (string-to-int emacs-version))
		  (proc (get-process "lkk-skkservd")))
	      (cond ((< 19 ver)
		     ;; This won't accept cons ARG. 
		     ;; cf. default-proc, also in 19.xx.
		     (set-process-coding-system proc 'euc-japan 'euc-japan))
		    ((= 19 ver)
		     (set-process-coding-system
		      proc '*euc-japan*unix '*euc-japan*unix))
		    ((<= 18 ver)
		     (save-excursion
		       ;; If your Nemacs is an old one,
		       ;; it may not have the function set-process-kanji-code.
		       (if (fboundp 'set-process-kanji-code)
			   (set-process-kanji-code
			    (get-process "lkk-skkservd") 0))
		       (set-buffer " *.lkk-skkserv*")
		       (setq kanji-fileio-code 0) ; for old Nemacs
		       ))
		    (t t		; NOP
		       )))
	  (message "Could not activate SKK SERVER.")
	  (sit-for 1)
	  ;; reset skk-server-host so as not to use server in this session
	  (setq lkk-skk-server-host nil)
	  (ding)))))

(defun j-open-server-subr ()
  (condition-case nil
      (open-network-stream
       "lkk-skkservd" " *.lkk-skkserv*" lkk-skk-server-host (or lkk-skk-portnum "skkserv"))
    (error
     nil)))

;;
;; j-search-server() -- Maneuvers server search.
;;
;; Modifications on the original Classic:
;;   * Modified to accept `yomi' string as arg KEY
;;   * Just returns hit skk record as string.
;; s.n. Wed Mar  8 00:15:03 2000
;;

;; (defun j-search-server (yomi)
(defun j-search-server (yomi)
  "Search for KEY through skkserver and create the vector of candidates."
  (let (;;(key j-henkan-key)
	;;(okuri-ari j-okurigana)
	;;(okurigana j-henkan-okurigana)
	;;(v (vector))
	(key yomi)
	)
    (save-excursion
      (set-buffer " *.lkk-skkserv*")
      (erase-buffer)
      (or (process-status "lkk-skkservd") (j-open-server))
      (if (process-status "lkk-skkservd")
	  (let ((cont t) (count 0))
	    (process-send-string "lkk-skkservd" (concat "1" key " "))
	    (while (and cont (memq (process-status "lkk-skkservd") '(open run)))
	      (accept-process-output)
	      (setq count (1+ count))
	      (if (> (buffer-size) 0)
		  (progn
		    (if (= ?1 (char-after 1))
			;; found key successfully, so check if a whole line
			;; is received.
			(if (= ?\n (char-after (1- (point-max))))
			    (setq cont nil))
		      ;; not found or error, so exit
		      (setq cont nil)))))
	    (goto-char (point-min))
	    (if lkk-skk-report-server-response
		(message "waited for server response %d times." count))
	    (if (looking-at "1")
		;; s.n. Tue Mar  7 23:44:16 2000
		;; Return slash-delimited kanji record only.
		(substring (lkk-get-current-line) 1)
;;;		(progn
;;;		  (forward-char 2)
;;;		  (let ((l (j-compute-henkan-vectors)))
;;;		    (j-vector-add (nth 0 l) (nth 2 l))))
	      ;; not found or error
	      ;;[]
	      nil))
	;;[]
	nil))))

(defun j-vector-add (u v)
  "Concatenate U and elements of V not in U."
    (let ((w u) (lv (length v)) (i 0))
      (while (< i lv)
	(let ((e (aref v i)))
	  (or (j-vector-mem e u)
	      (setq w (vconcat w (vector e))))
	  (setq i (1+ i))))
      w))

(defun j-vector-mem (e v)
  "Test if the string E is a member of the vector V."
  (let ((l (length v)) (i 0) (cont t))
    (while (and cont (< i l))
      (if (string= e (aref v i)) (setq cont nil) (setq i (1+ i))))
    (not cont)))

;; Renamed. s.n.
;;(defun skk-server-version ()
(defun j-server-version ()
  (interactive)
  (if lkk-skk-server-host
      (let (v)
	(set-buffer " *.lkk-skkserv*")
	(erase-buffer)
	(if (process-status "lkk-skkservd")
	    (progn
	      (process-send-string "lkk-skkservd" "2")
	      (accept-process-output "lkk-skkservd")
	      (setq v (buffer-substring (point-min) (1- (point-max))))
	      (erase-buffer)
	      (process-send-string "lkk-skkservd" "3")
	      (accept-process-output "lkk-skkservd")
	      (message "SKK SERVER %s running on HOST %s." v
		       (buffer-substring
			(point-min)
			(progn
			  (goto-char (point-min))
			  (search-forward ":")
			  (- (point) 1)))))
	  (message "SKK SERVER is not active.")))
    (message "You are not using SKK SERVER.")))

;;
;; Interfaces for lkk
;;
;; Wed Mar  8 00:14:41 2000 s.n.
;;

(require 'lkk-skk-dichead)

(defun lkk-j-search-server-get-lkk-record (yomi)
  "Search YOMI via Classic skkserv interface `j-search-server'.
Returns lkk record."
  (or (get-buffer " *.lkk-skkserv*")
      (j-open-server))
  (let ((klist (j-search-server yomi)) record)
    (cond
     (klist
      ;; lkk-skk-dichead.el has this:
      (setq record (lkk-skk-dict-to-list (concat yomi " " klist)))
      (cond
       ;; cdr part should be listified if okuri-converting.
       (lkk-okuri
	;; (lkk-okuri-converting-p) ; can't be used when reconverting.
	;; (cons (car record) (list (cdr record)))
	(list (car record) (cdr record))
	)
       (t record)) )
     (t nil))))

;;
;; lkk-j-search-server() -- The server affair
;;
;; This is still experimental and lousy in:
;;        1) Inserts kanji in this lower level of the engine layer.
;;        2) Nothing cared for dict update afterwards, and assumes that
;;           the hit server record would be thrown away.
;;
;; If adding data from server to my private dictionary is preferred
;; there might be some different implementations:
;;   (1) Always add data to private dict from server.
;;   (2) Do not add record to private dict from server.
;;     (2') Have user option controlling addig from server or not.
;;

(defun lkk-j-search-server (seed &optional nodisplay)
  "Roll over skkserv with SEED and modifies `lkk-current-display-record'.
Returns `lkk-current-display-record' pointer, or nil if failed.
The raw record derived from the server is set to `lkk-current-server-record'.
If `lkk-conv-beginning' is set, go for display unless optional 2nd arg
NODISPLAY is non-nil."
  (let ((record (lkk-j-search-server-get-lkk-record seed)))
    ;; If all server words turned out to be bogus, nil is is thrown to
    ;; 'bogus and this form returns nil.
    (catch 'nil-on-bogus
      (cond
       ((and record (>= (length record) 1))
	(setq lkk-server-affair 'server-hit ; or 'server-won ?
	      lkk-current-server-record record
	      ;; lkk-kanji-index 1
	      )
	(cond
	 ;; Called from lkk-convert(), there's no lkk-current-entry
	 ((null lkk-convert-hit-bottom)
	  (setq lkk-current-display-record record
		lkk-kanji-index 1)
	  )

	 ;; Otherwize called from lkk-reconvert().
	 ;; We have lkk-current-display-record and lkk-current-record here.
	 (lkk-convert-hit-bottom

	  ;; Append the whole server record to display record (for now).
	  ;; Remember that lkk-current-display-record will never used for
	  ;; update.

	  ;; Must omit duplicated kanji in the server record.
	  (cond
	   ;; okuri-ari reconverting, should generate new display record like:
	   ;; ("$B$*$/$k(B" "$BAw$k(B" "$BCY$k(B" "$BB#$k(B" "$Bn/$k(B")
	   (lkk-current-record-okuri
	    (let ((seed (car record))
		  (klist (car (cdr record)))
		  newdisprecord)
	      (mapcar (lambda (c)
			(setq klist (delete c klist)))
		      (nth 1 lkk-current-record-okuri))
	      (setq record (lkk-convert-okuri-crunch (list seed klist))
		    )
	      ))
	   ;; else okuri-nasi reconverting: (nth 1 record) is not a list.
	   (t 
	    (mapcar (lambda (c)
		      (setq record (delete c record)))
		    (cdr lkk-current-display-record)))) ; kanji part.

	  ;; All words from server may be bogus and not left here:
	  (cond ((= 1 (length record)) (throw 'nil-on-bogus nil)))

	  (setq ;;; lkk-current-record nil
	   lkk-kanji-index (1+ lkk-kanji-index)
	   lkk-current-display-record (nconc
				       ;; append
				       lkk-current-display-record
				       (cdr record)))
	  ))

	;; If NODISPLAY is not specified, do display as well:
	(cond
	 (nodisplay)
	 (lkk-conv-beginning
	  (lkk-replace-buffer-substring (+ lkk-j-bytes lkk-conv-beginning)
					(point)
					(nth lkk-kanji-index
					     lkk-current-display-record))
	  (cond ((null lkk-J-mode)
		 (lkk-setup-minor-mode 'lkk-J-mode)
		 (lkk-update-insignia lkk-insignia1)))
	  ))

	;; Here we can claim this:
	(message "Server hit: \"%s\"" seed)

	lkk-current-display-record)	; What is returned on success

       (t (setq lkk-server-affair 'server-lost)
	  nil)))))

;; (lkk-j-search-server "$B$+$s$8(B")
;; (lkk-j-search-server "$B$O$I$&$[$&(B")
