;;; `Lousy' --- Kana Kanji front end for emacs/mule

;; Normal people do not use lkk. Use skk instead.

;;
;; GPL2 should be applied to functions and variables derived from skk or emacs.
;;
;; As for other code I wrote from scratch,
;; I just don't put any copying/copyright notice other than this:
;; Copyright (C) 2000 Seiichi Namba <sn@asahi-net.email.ne.jp>
;; There is of course no warranty.
;; Thu Apr 13 18:22:26 2000
;; s.n.
;;

;;
;; Authour: Seiichi Namba <sn@asahi-net.email.ne.jp>
;; http://www.asahi-net.or.jp/~pi9s-nnb/
;;

;; Include this sort of codes in ~/.emacs to use lkk: 
;; (define-key global-map "\C-xJ" 'lkk-mode)
;; (autoload 'lkk-mode "lkk" nil t)
;;
;; Setup for windows.el/revive.el, to protect dict / work buffer for
;; wipe/restore:
;;   (setq revive:ignore-buffer-pattern "^ \\*.skk\\|^ \\*.*lkk")

;; Version 0
;; Wed Feb  2 03:47:53 2000 skkdict->kanji converter primitives 
;; Fri Feb  4 21:26:22 2000 Romaji-Kana input started.
;; Sat Feb  5 10:10:07 2000 Refined kana input.
;; Wed Feb  9 20:51:13 2000 Roma-kana input optimized with brief hashing
;; Sat Feb 12 16:08:01 2000 lkk-l-mode, lkk-L-mode, as minor mode.
;; Sun Feb 20 15:57:29 2000 lkk-J-mode (conversion) routines started.
;; Wed Feb 23 12:50:29 2000 lkk-elettronico invented.
;; Thu Feb 24 04:11:10 2000 Took lkk-dict as standard dict format.
;; Mon Feb 28 02:43:18 2000 'lkk-dict-alist method now updates lkk-big-dict.
;; Wed Mar  1 03:55:18 2000 Rough codes for okuri-ari conversion written.
;; Thu Mar  2 03:44:18 2000 Dictionary update code in okuri-ari conversion
;; Fri Mar  3 09:21:18 2000 Rough Auto-fill support
;; Sat Mar  4 12:31:18 2000 Numeric conversion is almost done.
;; Sun Mar  5 01:20:18 2000 Multiple keyclick support
;; Mon Mar  6 13:36:21 2000 lkk-isearch.el -- very short. Neat and clean.
;; Thu Mar  9 11:42:12 2000 Infinite multiple keyclick accepted.
;; Sun Mar 12 14:33:24 2000 Edition 1, skkserv codes, timer-dump
;; Fri Mar 17 18:51:04 2000 Lightspeeded lkk-j-input (30% or more).
;; Thu Mar 23 14:23:21 2000 New okurigana generation capable.
;; Fri Mar 24 15:34:27 2000 Purge data (by X).
;; Sun Mar 26 03:00:41 2000 Recursive register seemed easier.

(provide 'lkk)

(require 'lkk-mclick)
;; (require 'lkk-skklib)  ; not default
;; (require 'lkk-skkserv) ; not default

(defvar lkk-dict-type nil "Auto-detected but unused yet.")
(defvar lkk-dict-version nil "Auto-detected but unused yet.")

;; Actually this variable specifies dictionary TYPE, and interfacing METHOD.
;; Ugly, but might be easier for users and programmers.
(defvar lkk-dict-method
  ;; 'lkk-skk-dict-buffer
  'lkk-dict-buffer
  ;; 'lkk-dict-alist
  "lkk dictionary accessing method.  Specify one out of these:
'lkk-skk-dict-buffer
'lkk-dict-buffer
'lkk-dict-alist
Currently all of method are not yet completely implemented.")

(defconst lkk-version "0")
(defconst lkk-edition "8")
(defconst lkk-opus lkk-version "lkk-version alias")
(defconst lkk-Nr lkk-edition "lkk-edition alias")

(defun lkk-version ()
  (interactive)
  (let ((string (format "LKK Opus %s Nr %s" lkk-opus lkk-edition)))
    (if (interactive-p)
	(message string)
      string)))

;; The only official document, or help system for lkk.
;; Just one line. Cool, and I will never write the crap -- texi for this softoo.
;;
;; In emacs-20.xx, [mouse-2] on `lkk' in *Apropos* buffer shows the
;; genuine entity of the whole document.

(defun lkk-apropos (arg)
  "One liner lkk help system. emacs-20.xx may be nice for you."
  (interactive "P")
  (apropos "^lkk" arg))

(cond ((fboundp 'apropos-command)	; lacked in 19.28
       (defun lkk-apropos-command (arg)
	 "One liner lkk help system for lkk interactive commands."
	 (interactive "P")
	 (let ((fun 'apropos-command))	; to shut up BCcompiler.
	   (funcall fun "^lkk" arg)))))

(cond ((fboundp 'apropos-variable)	; lacked in 19.28/19.34
       (defun lkk-apropos-variable ()
	 "One liner lkk help system for lkk variables"
	 (interactive)
	 (let ((fun 'apropos-variable))
	   (funcall fun "^lkk" t)))))

;; (setq enable-recursive-minibuffers t) ; Not yet right here.

;;
;; Porting Stuff -- No version # test if possible -- don't know why, though.
;;

(or (fboundp 'defsubst)	; for emacs-18 ? (but porting is not progressed).
    (fset 'defsubst 'defun))
(or (fboundp 'buffer-substring-no-properties)
    (fset 'buffer-substring-no-properties 'buffer-substring))
(if (fboundp 'set-file-coding-system)	; E19.xx
    (fset 'lkk-set-file-coding-system 'set-file-coding-system)
  (if (fboundp 'set-buffer-file-coding-system) ; E20.xx maybe
      (fset 'lkk-set-file-coding-system 'set-buffer-file-coding-system)))
(or (fboundp 'local-variable-if-set-p) ; E19.28 lacks ...-if-set-p
    (fset 'local-variable-if-set-p 'local-variable-p))
(or (fboundp 'copy-list) ; E19.28, just for mouth-shut BCompiler.
    (require 'cl))

;;;
;;; Macro's and Defsubst's
;;;

;;
;; Buffer local variable definition macro
;;
;; Skelton derived from viper-util.el as in skk-10, modified to make
;; things simple -- It has no `default-value' arg.
;;
;; The second arg `documentation' is not &optional in lkk to ensure
;; raising "wrong # of arguments" error for wrong calling sequence,
;; and encourage the programmer to write adequate docstring.
;;
;; However viper-util's deflocalvar is not so bad if programmer definitely
;; knows it is equivalent to
;; (progn (setq-default var val)
;;        (defvar var val "xxx")
;;        (make-variable-buffer-local))
;; See also skk-set-marker comments in skk-foreword.el (skk10.49).
;;
(defmacro lkk-deflocalvar (var documentation) ;;;; default-value)
  (` (progn
       (make-variable-buffer-local '(, var)) ; same as (setq-default var nil)
       (defvar (, var) nil
	       (, (format "%s\n\(buffer local\)" documentation)))
;;;       (if (, default-value)
;;;	   (setq-default (, var) (, default-value)))
     )))

(lkk-deflocalvar lkk-no-undo-boundary-count
 "Buffers' no-undo counter.  The idea came from egg.el and skk.el.")
(setq-default lkk-no-undo-boundary-count 1)

;; The core part from mule-2.3's cancel-undo-boundary()
(defsubst lkk-cancel-undo-boundary-subr ()
  "Cancel undo boundary if `lkk-no-undo-boundary-count' is le to 20.
Otherwise `lkk-no-undo-boundary-count' is reset to 1."
  (cond ((< lkk-no-undo-boundary-count 20)
	 (if (and (consp buffer-undo-list)
		  (null (car buffer-undo-list)) )
	     (setq buffer-undo-list (cdr buffer-undo-list)) )
	 (setq lkk-no-undo-boundary-count (1+ lkk-no-undo-boundary-count)))
	(t (setq lkk-no-undo-boundary-count 1))))

;;; Not so many, so by hand. 
;;; (defmacro lkk-without-undo-update (sex)
;;;   `(let ((buffer-undo-list t)) ,sex))

(defsubst lkk-replace-buffer-substring (rb re str &optional arg noundo)
  "Replace region RB to RE with string STR.
Optional 4th ARG : times of STR insert.
Optional 5th ARG : disables `buffer-undo-list' recording."
  ;; This is symmetric with lkk-j-input()'s `key' insert,
  ;; but other functions call this. How should I do with them ?

;;; Thu Dec 14 19:40:01 2000 s.n.
;;; TESTING disabling the `noundo' control.
  ;; (if noundo
  ;;    (let ((buffer-undo-list t)) (delete-region rb re) )
  ;;   (delete-region rb re))
  (delete-region rb re)

  (or arg (setq arg 1))

  (while (and str (> arg 0))
    (insert str)
    (setq arg (1- arg))))

(defun lkk-string-match (str1 str2 &optional loose)
  "String match strict, or loose if optional 3rd arg LOOSE is non-nil."
  (if loose (string-match (format "^%s" str1) str2) (string= str1 str2)))

(defun lkk-same-line-p (pt0 pt1)
  ;; Non-nil if pt0 =< pt1 and no newline between them.
  ;; The idea was in skk-5.30.
  (and (<= pt0 pt1)
       (save-excursion
	 (goto-char pt1)
	 (beginning-of-line 1)
	 (<= (point) pt0))))

(defun lkk-run-hooks (hooksym)
  "Private version of run-hooks. Returns the last evaluated hooks' value
Supports only a flat list of functions."
  (and (boundp hooksym)
       (let (result (hooks (eval hooksym)))
	 (mapcar (lambda (h)
		   (setq result (funcall h)))
		 hooks)
	 result)))

(defun lkk-clear-marker (sym)
  "Bind object pointed by SYM to nil.
If the object is a marker, make it void, then set to nil."
  (let ((entity (eval sym)))
    (cond
     (entity
      (and (markerp entity)
	   (marker-position entity)
	   (set-marker entity nil))
      (set sym nil)))))

;; RESETCMD semantics is not so clear.  May have to rethink.
;;  "Check `lkk-okuri-converting', and lkk-insignia2 status.  If
;;optional arg RESETCMD is 'reset-vars-on-failure, all the okuri
;;variables are reset unless insignia2 found at `lkk-okuri-converting'.
;;If RESETCMD is 'init-all-on-success, unconditionally all the okuri
;;variables are cleared and the insignia is erased.
;;
;;The command 'reset-vars-on-failure is for lkk-j-input to reset vars
;;(insignia2 dissapeared already), and determines the operation on
;;failure of insignia2 search, while 'init-all-on-success is for
;;lkk-sit-back to reset stale insignia and vars, and determines the
;;operation on sucess of insignia2 search."
(defun lkk-okuri-converting-p (&optional resetcmd)
  "This function should be rewritten see src clicking above `lkk'."
  (and 
   ;;(and lkk-okuri-beginning (marker-position lkk-okuri-beginning))
   lkk-okuri-beginning

   (save-excursion
     (goto-char lkk-okuri-beginning)
     ;; should be regex, but is any char of insignia2 okay ?
     (cond
      ((looking-at (concat "\\" lkk-insignia2))
       (cond ((eq resetcmd 'init-all-on-success)
	      (delete-char 1)
	      (lkk-reset-okuri))
	     (t t)))
      ;; returns nil anyway
      ((eq resetcmd 'reset-vars-on-failure)
       (lkk-reset-okuri))))))

;; Resetting marker to void is tried for lkk-okuri-beginning
(defsubst lkk-reset-okuri ()
  "Initialize all the okurigana related engine variables."
  ;; returns nil
  (lkk-clear-marker 'lkk-okuri-beginning)
  (setq lkk-okuri-root nil
	lkk-okuri-h nil
	lkk-okuri nil
	lkk-current-record-okuri nil))

;; Auto-fill mode support -- Good enough.
;; I don't know why skk does complicated stuff like that.

(defsubst lkk-do-auto-fill ()
  "Auto-fill handler called from `lkk-sit-back', and `lkk-j-input'.
`lkk-j-input' does not call this if `lkk-do-auto-fill-in-j-input' is nil."
  (cond
   ((and auto-fill-function		; Now in auto-fill-mode
	 (> (current-column) fill-column))
    ;; (fill-paragraph nil)           ; This is too much.
    (funcall auto-fill-function)	; What's wrong with this ?
    ;;     (save-excursion
    ;;       (forward-line 0) (newline) (end-of-line)
    ;;       (fill-paragraph nil)
    ;;       (forward-line -1) (delete-backward-char 1)
    ;;       )
    )))

;; auto-fill-mode() coverup
(defun lkk-auto-fill-mode (arg)
  "Pass ARG to `auto-fill-mode'.
If ARG is greater than 15, `lkk-do-auto-fill-in-j-input' is set to t.
If ARG is smaller than -15, `lkk-do-auto-fill-in-j-input' is set to nil, but
`auto-fill-mode' is not called in this case.  Thus, numeric arguments of
absolute value greater than 15 can be used to control only
`lkk-do-auto-fill-in-j-input'.

The absolute value 16 of numeric argument can be input via
\\[universal-argument] \\[universal-argument] \\[lkk-auto-fill-mode] \
for positive value, or for negative value,
C-- \
\\[universal-argument] \\[universal-argument] \\[lkk-auto-fill-mode] in \
emacs-20, or \
\\[negative-argument] 1 6 \\[lkk-auto-fill-mode] in emacs-19."
  (interactive "P")
  (cond
   ;; C-u C-u
   ((and arg (> (prefix-numeric-value arg) 15))
    (setq lkk-do-auto-fill-in-j-input t)
    (auto-fill-mode arg)
    )
   ;; C-- C-u C-u
   ((and arg (< (prefix-numeric-value arg) -15))
    (setq lkk-do-auto-fill-in-j-input nil)
    ;; Not calling the function is rational and practical.
    ;; (auto-fill-mode arg)
    )
   (t (auto-fill-mode arg))))

;;;
;;; Variables, Primitive Tables (user customizable maybe, or maybe not)
;;;

;; Dictionary related definitions

(defvar lkk-big-dict nil "`okuri nasi' lkk dictionary big alist.")
(defvar lkk-big-dict-okuri nil "`okuri ari' lkk dictionary big alist.")
(defconst lkk-dict-buffer " *lkk-dict*" "lkk dictionary buffer name.")

(defvar lkk-dict "~/.lkk-dict" "lkk jisyo filename")

(defconst lkk-dict-header "LKK dictionary"
  "lkk dictionary header keyword.  Perhaps lisp comment char precedes, and
version number may follow. Used for auto-detection of lkk-dict-type.
but the auto-detection is not yet implemented.")

;; Cursor colors. No facility to detect background. Specify directly.

(defvar lkk-j-mode-cursor-color "dodgerblue")
(defvar lkk-J-mode-cursor-color lkk-j-mode-cursor-color)
(defvar lkk-L-mode-cursor-color "gold")
(defvar lkk-l-mode-cursor-color "slateblue2")
(defvar lkk-q-mode-cursor-color "palegreen")

;; Auto-fill controller

(defvar lkk-do-auto-fill-in-j-input nil
  "Call `auto-fill-function' in `auto-fill-mode' on kana input, or not.
`auto-fill-function' is always called on `lkk-sit-back' while `auto-fill-mode'.")

;; Primitive tables -- defvar'ed, but modification on some tables do not 
;; take effect unless (lkk-setup-r2k-internal) is called.

(defvar lkk-r2k-alist nil
  "Romaji -- Kana conversion table.  Setup on-the-fly. Internal use only.")
(defvar lkk-r2k-alist-C
  '(("cha" . ("" . "")) ("chi" . ("" . ""))
    ("chu" . ("" . "")) ("che" . ("" . ""))
    ("cho" . ("" . ""))

    ("sha" . ("" . "")) ("shi" . ("" . ""))
    ("shu" . ("" . "")) ("she" . ("" . ""))
    ("sho" . ("" . "")) ("sha" . ("" . ""))
    ("shi" . ("" . ""))     ("shu" . ("" . ""))
    ("she" . ("" . "")) ("sho" . ("" . ""))

    ("kya" . ("" . "")) ("kyu" . ("" . ""))
    ("kyo" . ("" . ""))

    ("sya" . ("" . "")) ("syu" . ("" . ""))
    ("syo" . ("" . "")) ("tya" . ("" . ""))

    ("tyu" . ("" . "")) ("tyo" . ("" . ""))

    ("nya" . ("ˤ" . "˥")) ("nyu" . ("ˤ" . "˥"))

    ("nyo" . ("ˤ" . "˥")) ("hya" . ("Ҥ" . "ҥ"))

    ("hyu" . ("Ҥ" . "ҥ")) ("hyo" . ("Ҥ" . "ҥ"))

    ("mya" . ("ߤ" . "ߥ")) ("myu" . ("ߤ" . "ߥ"))
    ("myo" . ("ߤ" . "ߥ")) ("rya" . ("" . ""))

    ("ryu" . ("" . "")) ("ryo" . ("" . ""))

    ("gya" . ("" . "")) ("gyu" . ("" . ""))
    ("gyo" . ("" . ""))

    ("zya" . ("" . "")) ("zyu" . ("" . ""))
    ("zyo" . ("" . ""))

    ("dya" . ("¤" . "¥")) ("dyu" . ("¤" . "¥"))
    ("dyo" . ("¤" . "¥"))

    ("bya" . ("Ӥ" . "ӥ")) ("byu" . ("Ӥ" . "ӥ"))
    ("byo" . ("Ӥ" . "ӥ"))

    ("pya" . ("Ԥ" . "ԥ")) ("pyu" . ("Ԥ" . "ԥ"))
    ("pyo" . ("Ԥ" . "ԥ"))

    ("jya" . ("" . "")) ("jyu" . ("" . ""))
    ("jyo" . ("" . ""))

    ("tsu" . ("" . ""))

    ("mm" . ("" . ""))
    ("nn" . ("" . ""))
    ("m'" . ("" . ""))
    ("n'" . ("" . ""))

    ("fa" . ("դ" . "ե")) ("fi" . ("դ" . "ե"))
    ("fu" . ("" . ""))     ("fe" . ("դ" . "ե"))
    ("fo" . ("դ" . "ե"))

    ("ka" . ("" . "")) ("ki" . ("" . ""))
    ("ku" . ("" . "")) ("ke" . ("" . ""))
    ("ko" . ("" . ""))
    ("sa" . ("" . "")) ("si" . ("" . ""))
    ("su" . ("" . "")) ("se" . ("" . ""))
    ("so" . ("" . ""))
    ("ta" . ("" . "")) ("ti" . ("" . ""))
    ("tu" . ("" . "")) ("te" . ("" . ""))
    ("to" . ("" . ""))
    ("na" . ("" . "")) ("ni" . ("" . ""))
    ("nu" . ("" . "")) ("ne" . ("" . ""))
    ("no" . ("" . ""))
    ("ha" . ("" . "")) ("hi" . ("" . ""))
    ("hu" . ("" . "")) ("he" . ("" . ""))
    ("ho" . ("" . ""))
    ("ma" . ("" . "")) ("mi" . ("" . ""))
    ("mu" . ("" . "")) ("me" . ("" . ""))
    ("mo" . ("" . ""))
    ("ya" . ("" . "")) ("yu" . ("" . ""))
    ("yo" . ("" . ""))
    ("ra" . ("" . "")) ("ri" . ("" . ""))
    ("ru" . ("" . "")) ("re" . ("" . ""))
    ("ro" . ("" . ""))
    ("wa" . ("" . "")) ("wi" . ("" . ""))
    ("we" . ("" . "")) ("wo" . ("" . ""))

    ("ga" . ("" . "")) ("gi" . ("" . ""))
    ("gu" . ("" . "")) ("ge" . ("" . ""))
    ("go" . ("" . ""))
    ("za" . ("" . "")) ("zi" . ("" . ""))
    ("zu" . ("" . "")) ("ze" . ("" . ""))
    ("zo" . ("" . ""))
    ("da" . ("" . "")) ("di" . ("" . ""))
    ("du" . ("" . "")) ("de" . ("" . ""))
    ("do" . ("" . ""))
    ("ba" . ("" . "")) ("bi" . ("" . ""))
    ("bu" . ("" . "")) ("be" . ("" . ""))
    ("bo" . ("" . ""))
    ("pa" . ("" . "")) ("pi" . ("" . ""))
    ("pu" . ("" . "")) ("pe" . ("" . ""))
    ("po" . ("" . ""))

    ("ja" . ("" . "")) ("ji" . ("" . ""))
    ("ju" . ("" . "")) ("je" . ("" . ""))
    ("jo" . ("" . ""))

    ("va" . ("" . "")) ("vi" . ("" . ""))
    ("vu" . ("" . "")) ("ve" . ("" . ""))
    ("vo" . ("" . ""))

    ("xa" . ("" . ""))
    ("xka" . ("" . ""))
    ("xwa" . ("" . ""))
    ("xya" . ("" . ""))
    ("xi" . ("" . ""))
    ("xwi" . ("" . ""))
    ("xu" . ("" . ""))
    ("xtu" . ("" . ""))
    ("xtsu" . ("" . ""))
    ("xtu" . ("" . ""))
    ("xyu" . ("" . ""))
    ("xe" . ("" . ""))
    ("xke" . ("" . ""))
    ("xwe" . ("" . ""))
    ("xo" . ("" . ""))
    ("xyo" . ("" . ""))
    )
  "Romaji -- Kana basic table 1 -- all yomi beginning from consonant.
If you are modified this alist, eval:
 \(lkk-setup-r2k-internal\)
to reflect the changes.")

(defconst lkk-r2k-alist-V
  '(("a" . ("" . ""))
    ("i" . ("" . ""))
    ("u" . ("" . ""))
    ("e" . ("" . ""))
    ("o" . ("" . "")))
  "Romaji -- Kana basic table 0 -- yomi includes vowels only.")

;; Perhaps unnecessary
;; (defconst lkk-vowel-alist
;;  ;;(mapcar (lambda (c) (car c)) lkk-r2k-alist-V)
;;  '("a" "i" "u" "e" "o"))

;;(lkk-dump-chars ?  (1+ (- 127 ? )) ?\x20 16)
(defconst lkk-ascii-vector
  [nil  nil  nil  nil  nil  nil  nil  nil
   nil  nil  nil  nil  nil  nil  nil  nil 
   nil  nil  nil  nil  nil  nil  nil  nil 
   nil  nil  nil  nil  nil  nil  nil  nil 
   " "  "!"  "\"" "#"  "$"  "%"  "&"  "'"
   "("  ")"  "*"  "+"  ","  "-"  "."  "/"
   "0"  "1"  "2"  "3"  "4"  "5"  "6"  "7" 
   "8"  "9"  ":"  ";"  "<"  "="  ">"  "?" 
   "@"  "A"  "B"  "C"  "D"  "E"  "F"  "G" 
   "H"  "I"  "J"  "K"  "L"  "M"  "N"  "O" 
   "P"  "Q"  "R"  "S"  "T"  "U"  "V"  "W" 
   "X"  "Y"  "Z"  "["  "\\"  "]"  "^"  "_" 
   "\`"  "a"  "b"  "c"  "d"  "e"  "f"  "g" 
   "h"  "i"  "j"  "k"  "l"  "m"  "n"  "o" 
   "p"  "q"  "r"  "s"  "t"  "u"  "v"  "w" 
   "x"  "y"  "z"  "{"  "|"  "}"  "~"  nil]
  "Ascii characters -- mainly for `l' mode.")

(defconst lkk-alpha-vector
  [nil  nil  nil  nil  nil  nil  nil  nil
   nil  nil  nil  nil  nil  nil  nil  nil 
   nil  nil  nil  nil  nil  nil  nil  nil 
   nil  nil  nil  nil  nil  nil  nil  nil 
   nil  nil  nil  nil  nil  nil  nil  nil
   nil  nil  nil  nil  nil  nil  nil  nil 
   nil  nil  nil  nil  nil  nil  nil  nil 
   nil  nil  nil  nil  nil  nil  nil  nil 
   nil  "A"  "B"  "C"  "D"  "E"  "F"  "G" 
   "H"  "I"  "J"  "K"  "L"  "M"  "N"  "O" 
   "P"  "Q"  "R"  "S"  "T"  "U"  "V"  "W" 
   "X"  "Y"  "Z"  nil  nil  nil  nil  nil 
   nil  "a"  "b"  "c"  "d"  "e"  "f"  "g" 
   "h"  "i"  "j"  "k"  "l"  "m"  "n"  "o" 
   "p"  "q"  "r"  "s"  "t"  "u"  "v"  "w" 
   "x"  "y"  "z"  nil  nil  nil  nil  nil ]
  "Alphabetic characters.")

(defconst lkk-zenkaku-vector
  [nil  nil  nil  nil  nil  nil  nil  nil
   nil  nil  nil  nil  nil  nil  nil  nil 
   nil  nil  nil  nil  nil  nil  nil  nil 
   nil  nil  nil  nil  nil  nil  nil  nil 
   ""  "" "" "" "" "" "" ""
   "" "" "" "" "" "" "" ""
   "" "" "" "" "" "" "" ""
   "" "" "" "" "" "" "" ""
   "" "" "" "" "" "" "" ""
   "" "" "" "" "" "" "" ""
   "" "" "" "" "" "" "" ""
   "" "" "" "" "" "" "" ""
   "" "" "" "" "" "" "" ""
   "" "" "" "" "" "" "" ""
   "" "" "" "" "" "" "" ""
   "" "" "" "" "" "" "" nil]
  "Zenkaku characters -- mainly for `L' mode.")

;;;
;;; Table build and Utilities referring the tables
;;;

;;
;; Romaji -- Kana Table build, and
;; Romaji -- Kana conversion handling subroutines
;;

(defvar lkk-r2k-terminator-list nil
  "List of possible string of romaji yomi terminator. Internal use only.")
(defvar lkk-valid-consonant-list nil
  "List of possible consonant part of romaji yomi. Internal use only.")
(defvar lkk-valid-consonant-vector nil
  "Vector of valid consonant list hashed by 1st char. Internal use only.")
(defvar lkk-r2k-alist-vector nil
  "Vector of romaji-kana alist hashed by 1st char of yomi. Internal use only.")

(defun lkk-build-valid-consonant-list (alis)
  "Return all consnant part of romaji yomi as uniqued list."
  (let ((uniqued nil))
    (mapcar (lambda (c) ; Uniqueing squeezes the list down to 1/3 to 1/4.
	      (if (and (null (member c uniqued))
		       (> (length c) 0))
		  (setq uniqued (cons c uniqued))))
	    (mapcar (lambda (c) (substring c 0 -1)) ; all consonant part
		    (mapcar (lambda (c) (car c)) ; all yomi
			    alis)))
    uniqued))

(defun lkk-build-valid-consonant-vector (xlis)
  "Return vector of valid consonant lists built out of XLIST.
The vector is hashed by 1st char."
  (let ((vec (make-vector 128 nil))	; Always start with fresh table.
	ch tem)
    (mapcar
     (lambda (consonant)
       (setq ch (string-to-char (substring consonant 0 1))
	     tem (aref vec ch))
       (aset vec ch (append tem (list consonant))) )
     xlis)
    vec))

(defun lkk-build-r2k-alist-vector (xlis)
  "Return a hashed vector of alists built from XLIST (alist maybe).
The returned vector contains alists hashed by 1st char."
  (let ((vec (make-vector 128 nil))	; Always start with fresh table.
	ch tem)
    (mapcar
     (lambda (cell)
       (setq ch (string-to-char (substring (car cell) 0 1))
	     tem (aref vec ch))
       (aset vec ch (append tem (list cell))) )
     xlis)
    vec))

(defun lkk-build-r2k-terminator-list (&optional table)
  (let (tem lastchar (table (or table lkk-r2k-alist-C)))
    (mapcar (lambda (c)
	      (setq lastchar (substring (car c) -1))
	      (or (member lastchar tem)
		  (setq tem (cons lastchar tem))))
	    table)
    (nreverse tem)))

(defun lkk-setup-r2k-internal ()
  "Setup lkk's internal alist/list/vectors for Romaji conversion."
  ;; setq order is very significant.
  (setq lkk-r2k-alist
	(append 
	 lkk-r2k-alist-C
	 lkk-r2k-alist-V))
  (setq lkk-r2k-terminator-list
	(lkk-build-r2k-terminator-list lkk-r2k-alist))
  (setq lkk-valid-consonant-list
	(lkk-build-valid-consonant-list lkk-r2k-alist))
  (setq lkk-valid-consonant-vector
	(lkk-build-valid-consonant-vector lkk-valid-consonant-list))
  (setq lkk-r2k-alist-vector
	(lkk-build-r2k-alist-vector lkk-r2k-alist)))

;; Setup roma -- kana table on loading this program.
(lkk-setup-r2k-internal)

(defsubst lkk-r2k-terminator-p (str)
  "Return non-nil if STR is member of lkk-r2k-terminator-list."
  (member str lkk-r2k-terminator-list))

(defsubst lkk-valid-roma-p0 (str &optional lis)
  "Predicate if STR is head part of valid Romaji yomi.
Optional 2nd arg LIST is used as list of valid strings, or
lkk-valid-consonant-list directly used if optional 2nd arg is nil.
This function is not so fast. Used by faster one, so don't purge this function."
  (let ((xlis (or lis lkk-valid-consonant-list)) tem)
    (catch 'hit
      (while xlis
	(and ;;(string-match (concat "^" str) (car xlis))
	 (setq tem (string-match str (car xlis)))
	 (= tem 0)
	 (throw 'hit t)) 
	(setq xlis (cdr xlis)))
      nil)))

(defun lkk-valid-roma-p (str)
  "Predicate if STR is head part of valid Romaji yomi."
  (let* ((table (aref lkk-valid-consonant-vector
		      (string-to-char (substring str 0 1)))))
    (if table (lkk-valid-roma-p0 str table) nil) ))

(defsubst lkk-get-kana0 (str &optional alis)
  "Return kana corresponding to romaji in STR.  Slow primitive subroutine."
  (let ((cell (assoc str (or alis lkk-r2k-alist))))
    (if cell 
	(if lkk-q-mode (cdr (cdr cell))
	  (car (cdr cell)))
      nil)))

(defun lkk-get-kana (str)
  "Return kana corresponding to romaji in STR."
  (let ((table (aref lkk-r2k-alist-vector
		      (string-to-char (substring str 0 1)))))
    (if table (lkk-get-kana0 str table) nil) ))

;;
;; ascii case, h2k, k2h conversion.  Very lousy.
;;

(defsubst lkk-isalpha (char)
  (and (<= ?  char)
       (<= char ?~)
;;;    (aref lkk-alpha-vector char)
       ))

(defsubst lkk-isnum (char)
  (and (<= ?0 char) (<= char ?9)))

(defsubst lkk-isdabbrev (char)
  (or (lkk-isalpha char)
      (lkk-isnum char)))

(defsubst lkk-tolower (x)
  (cond ((stringp x) 
	 (char-to-string (+ (string-to-char x) 32))) ; (- ?a ?A) 32
	(t (+ x 32))			; else x must be char
	 ))

(defsubst lkk-toupper (x)
  (cond ((stringp x) 
	 (char-to-string (- (string-to-char x) 32))) ; (- ?a ?A) 32
	(t (- x 32))			; else x must be char
	))

;; This length hack returns 3 and 1 respectively on mule-2.3(emacs-19.xx)
;; and emacs-20.5.  I don't know what happens emacs-20.[1234].
;; Perhaps 18.xx returns 2.
;; current-column() seems like compatible in emacs-19.xx and
;; 20.xx, but is it possible to use only `column' ?  -- Nop (it's no
;; interface with `marker').

(defconst lkk-j-bytes
  ;;(char-bytes (string-to-char ""))
  ;;(let (p) (with-temp-buffer
  ;;          (erase-buffer) (insert lkk-insignia0  "")
  ;;          (goto-char 1)  (forward-char 1) (- (point) 1)))
  (length "")
  )

;; sref() is alias of aref in emacs-20.xx, while
;; sref() is special and different from aref() in mule-2.3@emacs-19.xx
;; char-bytes() still exists in emacs-20.xx, and always returns 1.
;; See mule-util.el:
;; string-to-char-list() is in mule-util.el of 19.xx (poem.el has it too).
;; string-to-list() is in mule-util.el of 20.xx.

;; These fset's are just to suppress BCcompile warnings in e20.xx.
(fset 'lkk-char-bytes 'char-bytes)
(fset 'lkk-sref (if (eq (symbol-function 'sref) 'aref) 'aref 'sref))
(defun lkk-string-to-char-list (str)
  ;; Code is stolen from mule-util.el in mule-2.3@emacs-19.34.
  (let ((len (length str))
	(idx 0)
	c l)
    (while (< idx len)
      (setq c (lkk-sref str idx))		; Must be sref in emacs-19.xx
      (setq idx (+ idx (lkk-char-bytes c)))
      (setq l (cons c l)))
    (nreverse l)))

;; Unused. Works only on emacs-20.
;;(defsubst lkk-string-to-char-list (str)
;;  "Hopefully equivalent to string-to-char-list in mule-2, and emacs-20.x."
;;  (mapcar (lambda (c) c) str))
;;(defsubst lkk-string-to-char (str)
;;  "Hopefully equivalent to string-to-char in mule-2, and emacs-20.x."
;;  (car (lkk-string-to-char-list str)))

(defvar lkk-h2k-alist nil "Alist for hiragana to katakana conversion.")
(defvar lkk-k2h-alist nil "Alist for katakana to hiragana conversion.")

;; Lousy ! "" can't be supported.
(defun lkk-build-h2k-k2h-alist ()
  (setq lkk-h2k-alist nil lkk-k2h-alist nil) ; reset
  (let ((hc ?) (kc ?) (hs) (ks)
	h2k-alist k2h-alist)
    (while (>= ? hc)
      (setq hs (char-to-string hc)
	    ks (char-to-string kc)
	    h2k-alist (cons (cons hs ks) h2k-alist)
	    k2h-alist (cons (cons ks hs) k2h-alist))
      (setq hc (1+ hc)
	    kc (1+ kc)))
    (setq lkk-h2k-alist (nreverse h2k-alist)
	  lkk-k2h-alist (nreverse k2h-alist))))

(lkk-build-h2k-k2h-alist)

;; Lousy ! "" cann't be supported.
(defun lkk-h2k (hstr)
  "Convert HIRAGANA string to katakana."
  (mapconcat
   (lambda (c) 
     (let ((tem (assoc (char-to-string c) lkk-h2k-alist)))
       (if tem (cdr tem) (char-to-string c))))
   (lkk-string-to-char-list hstr) ""))

(defun lkk-k2h (kstr)
  "Convert KATAKANA string to hiragana."
  (mapconcat
   (lambda (c)
     (let ((tem (assoc (char-to-string c) lkk-k2h-alist)))
       (if tem (cdr tem) (char-to-string c))))
   (lkk-string-to-char-list kstr) ""))

(defun lkk-han2zen (arg)
  (if (integerp arg)
      (aref lkk-zenkaku-vector arg)
    (apply 'concat (mapcar (lambda (c)
			     (lkk-han2zen c)) ; Only 1 level of recursion.
			   (lkk-string-to-char-list arg)))))

(defvar lkk-zen2han-alist nil
  "Alist for zenkaku alpha to hankaku conversion.")

(defun lkk-build-zen2han-alist ()
  (let (tem (c ? ))
    (while (<= c ?~)
      (setq tem (cons (cons (aref lkk-zenkaku-vector c)
			    (char-to-string c)) tem)
	    c (1+ c)))
      (setq lkk-zen2han-alist (nreverse tem))))
(lkk-build-zen2han-alist)

(defun lkk-zen2han (arg)
  (if (integerp arg)
      (cdr (assoc (char-to-string arg) lkk-zen2han-alist))
    (apply 'concat (mapcar (lambda (c)
			     (lkk-zen2han c)) ; Only 1 level of recursion.
			   (lkk-string-to-char-list arg)))))

;;
;; Misc edit commands
;;

(defun lkk-h2k-region (rb re)
  "Translate hiragana region RB RE to katakana."
  (interactive "*r")
  (let* ((str (buffer-substring-no-properties rb re))
	 (result (lkk-h2k str)))
    (goto-char rb)
    (delete-region rb re)
    (insert result)))
(fset 'lkk-katakana-region 'lkk-h2k-region)

(defun lkk-k2h-region (rb re)
  "Translate  katakana region RB RE to hiragana."
  (interactive "*r")
  (let* ((str (buffer-substring-no-properties rb re))
	 (result (lkk-k2h str)))
    (goto-char rb)
    (delete-region rb re)
    (insert result)))
(fset 'lkk-hiragana-region 'lkk-k2h-region)

;; Cheap. May raise error if non-hankaku in region
(defun lkk-han2zen-region0 (rb re)
  "`lkk-han2zen-region' subroutine."
  (interactive "*r")
  (let* ((zstr (lkk-han2zen (buffer-substring-no-properties rb re))))
    (if zstr (lkk-replace-buffer-substring rb re zstr))))

(defun lkk-han2zen-region (rb re)
  "Translate zenkaku region RB RE to hankaku."
  (interactive "*r")
  (let ((end (make-marker)) (pt rb) (regex "[ -~]"))
    (set-marker end re)			; Marker should be used for zenkaku insert.
    (goto-char rb)
    (while (and (< (point) end) (not (eobp)))
      (cond ((looking-at regex)
	     (setq pt (point))
	     (while (and (< (point) end) (not (eobp)) (looking-at regex))
	       (forward-char 1))
	     (lkk-han2zen-region0 pt (point)))
	    (t
	     (forward-char 1))))
    (set-marker end nil)))

;; Cheap. May raise error if non-zenkaku in region
(defun lkk-zen2han-region (rb re)
  "Translate hankaku region RB RE to zenkaku."
  (interactive "*r")
  (let* ((hstr (lkk-zen2han (buffer-substring-no-properties rb re))))
    (if hstr (lkk-replace-buffer-substring rb re hstr))))

;;;
;;; lkk-mclick.el support
;;;

;; Assumes [M-return] has no default bindig.
(defvar lkk-fireup-old-command nil 

;;; No, this ain't work nice if user wants to use another bind.
;;; Immediate value is far better.
;;;  (let ((com (key-binding [?\M- ])))
;;;    ;; If user bound lkk-fireup already, can't use it.
;;;    (if (eq com 'lkk-fireup) 'just-one-space com))

  "The command executed if \\[lkk-fireup] \(lkk-fireup\) can't get \
double-XXXX pseudo key event.

The default value is nil, assuming that `lkk-fireup' would be bound to
[M-return] \(and a multiple-click sequence such as \[double-M-return\]
toggles lkk-mode\), which seems to have no default global binding in
emacsen.

If you want to bind `lkk-fireup' to any other key, and did not set a
correct value to `lkk-fireup-old-command', it may cause disaster.
In case you want to bind `lkk-fireup' to [M-SPC], the correct value may
be 'just-one-space.

Normally this variable's value should be a symbol of command bound to
a key sequence derived by removing `double' prefix from the double-X-Y
pseudo sequence you want to use for toggling lkk.  Eg.,

If you want to bind \[double-C-SPC\] pseudo event to `lkk-fireup', the
right value of lkk-fireup-old-command might be `set-mark-command'. Of
course you must bind the pseudo key \[double-C-SPC\] to `lkk-toggle-lkk-mode'
in `lkk-multiple-click-map' as well:

  \(define-key map [double-C-SPC] 'lkk-toggle-lkk-mode\)

Do not setq the value of this variable by `key-binding' etc.. Always
use immediate value, or you lose in most cases I guess.

See also `lkk-multiple-click-map', `lkk-multiple-click-crank'.")

(defun lkk-toggle-lkk-mode ()
  "Equivalent to `lkk-mode', but always return t."
  (interactive)
  ;; Without this cover-up, you'll lose on hitting \\[lkk-fireup] \(lkk-fireup\),
  ;; because lkk-multiple-click-crank() is always checked.
  (lkk-mode)
  t)

;; \\<lkk-multiple-click-map> hack doesn't work both in E19.34/E20.5.
;; Maybe because it is virtual map, or just not bound to legal map.
(defun lkk-fireup (arg)
  "Toggles lkk-mode with using very very dirtly trick.
Binding this command to \"\\e \" in global-map enables toggling `lkk-mode' with
\[double-M-SPC\]. See also `lkk-multiple-click-map', `lkk-multiple-click-crank'."
  (interactive "P")
  (cond
   ((and (null lkk-inhibit-multiple-keyclick)
	 (lkk-multiple-click-crank))
    ())
   (t ;; (lkk-maneuver-old-keybind) ; This can't be used. lkk-fireup is bound in global-map
    (if (fboundp lkk-fireup-old-command)
	(call-interactively lkk-fireup-old-command))
      )))

;;;
;;; LKK Automaton status variables (includes minor-mode variable definitions)
;;;

;; For debug use. Guess it's right to make it buffer-local.
(lkk-deflocalvar lkk-debug-hook "lkk debugger.")

(defconst lkk-valid-minor-minor-modes '(lkk-j-mode
					lkk-J-mode
					lkk-l-mode
					lkk-L-mode
					lkk-q-mode)
  "List of valid submode variable symbols.")

(lkk-deflocalvar lkk-j-mode "`j' mode. Kana mode.")
(lkk-deflocalvar lkk-J-mode "`J' mode. Turned on while kanji conversion.")
(lkk-deflocalvar lkk-l-mode "`l' mode. Direct ascii mode.")
(lkk-deflocalvar lkk-L-mode "`L' mode. Direct zenkaku mode.")
(lkk-deflocalvar lkk-q-mode "`q' mode. Katakana mode.")
(lkk-deflocalvar lkk-/-mode "`/' mode. Set in / convert.")
(lkk-deflocalvar lkk-mode "lkk basic -- lowest layer mode.")

(defvar lkk-insignia0 "" "Insignia shown while conversion.") 
(defvar lkk-insignia1 "" "Insignia shown on kakutei status.")
(defvar lkk-insignia2 "." "Insignia shown on okurigana hit.")
;;(defvar lkk-insignia0 "" "Insignia shown while conversion.") 
;;(defvar lkk-insignia1 "" "Insignia shown on kakutei status.")
;;(defvar lkk-insignia2 "@" "Insignia shown on okurigana hit.")

;; r2k conversion

(lkk-deflocalvar lkk-j-beginning
		 "lkk internal. Point/marker pointing kana start.")
(lkk-deflocalvar lkk-j-numarg
		 "lkk internal. Arg passed to Kana sequence.")

;; k2k conversion

(lkk-deflocalvar lkk-conv-beginning
 "lkk internal. Conversion beginning. Nil while not-converting.")
(lkk-deflocalvar lkk-okuri
 "lkk internal. Okurigana consonant -- ASCII. Nil while non-okuri conv.")
(lkk-deflocalvar lkk-okuri-h
 "lkk internal. Okurigana -- Hiragana. Nil while non-okuri conv.")
(lkk-deflocalvar lkk-okuri-root
 "lkk internal. Root kanji part -- Hiragana yomi. Nil while non-okuri conv.")

(lkk-deflocalvar lkk-okuri-beginning
 "lkk internal. Enclose okurigana. Nil while non-okuri conv.")

(lkk-deflocalvar lkk-current-record-okuri
 "lkk internal. The whole okuri-ari record of the latest hit.
Set by lkk-convert-okuri.

Typical value:

\(\"Ǥr\" \(\"\"\) \(\"\" \"\"\) \(\"\" \"\"\)\)")

;; Deprecated. See lkk-/-mode. Sun Mar 26 17:18:40 2000 s.n.
;;; (lkk-deflocalvar lkk-/-converting
;;;		 "lkk internal. t while / conversion.")

(lkk-deflocalvar lkk-full-auto		; Really experimental.
		 "Non-nil while full auto yomi complete.")

;; Unused. Equivalent to lkk-J-mode.
;;(lkk-deflocalvar lkk-converting "Non-nil while kanji conversion.")
(lkk-deflocalvar lkk-current-record
		 "Accommodates entry under conversion.")
(lkk-deflocalvar lkk-kanji-index
		 "Indexes kanji under conversion. Starts from 0.")
(lkk-deflocalvar lkk-current-display-record
 "Display version of lkk-current-record. Currently used only in numeric conv.")

;; Variables relevant to skkserv support (see lkk-skkserv.el):
(lkk-deflocalvar lkk-current-server-record
  "A buffer to propagate a dictionary record from skkserv.")

(lkk-deflocalvar lkk-convert-hit-bottom
  "Flag set when lkk-reconvert hit the bottom.
Meaning a record having current yomi is found, but user has gone far at least
to server search.")

(lkk-deflocalvar lkk-server-affair
 "Becomes non-nil value if you had an affair with the server while conversion.
The values are maybe one of 'server-hit, 'server-lost. \
Nil if you had no affair.")

;; Variable used to propagate a record explicitly registered by user
;; Mostly for `recovert' phase.
(lkk-deflocalvar lkk-current-user-record
  "A buffer to propagate a dictionary record explicitly registered by user.")

;;;
;;; Minor mode definitions and utils
;;;

;; Several utils related to mode fiddles.

(defconst lkk-exit-key
  (or 
   (car (where-is-internal 'lkk-mode '(keymap))) ; search only global-map
   "\C-xJ")
  "User defined key to exit lkk-mode.
Normally that key is bound in ~/.emacs...
Used to exit directly from lkk-L/l-mode.")

(defvar lkk-default-cursor-color nil
  "Initial cursor color used when lkk is turned on.")

(lkk-deflocalvar lkk-old-cursor-color "Old cursor color backed up.")

(defvar lkk-inhibit-set-cursor-color nil
  "Non-nil turns off cursor color hack. Mostly for slow machines.")

(defsubst lkk-set-cursor-color0 (color)
  (or lkk-inhibit-set-cursor-color
      (set-cursor-color color)))
  
(defsubst lkk-get-cursor-color ()
  "Get current cursor color -- mainly for getting color before lkk is enabeled."
  (if lkk-inhibit-set-cursor-color
      nil
    (cdr (assoc 'cursor-color (frame-parameters (selected-frame))))))

(defun lkk-set-cursor-color (color)
  (condition-case nil
      (lkk-set-cursor-color0 color)
    (error
     (lkk-set-cursor-color0 lkk-default-cursor-color)
     ;; color exhausted, or just invalid color string ?
     (message "Cursor color `%s' can't be set." color)
     )))

(defun lkk-init-cursor-color ()
  (setq lkk-default-cursor-color (lkk-get-cursor-color))
  (add-hook 'kill-buffer-hook
	    (lambda () (lkk-set-cursor-color0 lkk-default-cursor-color))))

(defun lkk-set-right-cursor-color ()
  (if lkk-mode
      (mapcar
       (lambda (m) 
	 (if (eval m)	      
	     (lkk-set-cursor-color
	      (eval (intern-soft (format "%s-%s" m "cursor-color"))))))
       lkk-valid-minor-minor-modes)
    (lkk-set-cursor-color lkk-default-cursor-color)))

;; I don't know what I am doing... Just stole the idea from skk-10.49.

(or (fboundp 'set-cursor-color)
    (defun set-cursor-color (&rest) nil)) ; NOP

(defadvice hilit-recenter (after lkk-ad activate)
  "Set lkk's right cursor color."
  (and lkk-mode (lkk-set-right-cursor-color)))

(defadvice hilit-yank (after lkk-ad activate)
  "Set lkk's right cursor color."
  (and lkk-mode (lkk-set-right-cursor-color)))

(defadvice hilit-yank-pop (after lkk-ad activate)
  "Set lkk's right cursor color."
  (and lkk-mode (lkk-set-right-cursor-color)))

(defadvice yank (after lkk-ad activate)
  "Set lkk's right cursor color."
  (and lkk-mode (lkk-set-right-cursor-color)))

(defadvice yank-pop (after lkk-ad activate)
  "Set lkk's right cursor color."
  (and lkk-mode (lkk-set-right-cursor-color)))

(defun lkk-setup-minor-mode (modesym &optional cursorcolor no-sit-back)
  "Set lkk's minor minor-mode to MODESYM.
Passing 'lkk-default as MODESYM turns off all minor minor-mode's,
and sets cursor color to lkk-default-cursor-color.
If optional 2nd arg CURSORCOLOR is specified, use it."
;;; 'lkk-default fiddle may be tricky, and I may be confused (in the future).
;;; 'lkk-default matches none of them.
  (mapcar (lambda (mode)
	    (if (eq modesym 'lkk-J-mode)
		(if (eq modesym mode) (set mode t)) ; just turn on `modesym'.
	      (set mode (eq modesym mode)) ; This turns off other mode.
	      ))
	  lkk-valid-minor-minor-modes)

  (or (eq modesym 'lkk-J-mode) ; Don't sit-back if conversion on fire,
      no-sit-back	       ; or no-sit-back is explicitly spacified.
      (lkk-sit-back))

  (cond
   ((member modesym lkk-valid-minor-minor-modes)
    (or lkk-default-cursor-color (lkk-init-cursor-color)) ; init done here.
    (setq lkk-old-cursor-color (lkk-get-cursor-color))
    (run-hooks (intern-soft (format "%s-%s" modesym "hook")))))

  ;; Not in the `cond' above, because caller use 'lkk-default to exit mode.
  (lkk-set-cursor-color (or cursorcolor
			    (eval (intern-soft
			       (format "%s-%s" modesym "cursor-color")))))
  (force-mode-line-update))

;; This must be called only by lkk-mode().
(defsubst lkk-all-minor-mode-disable ()
  "Turn off lkk's minor minor-mode passing 'lkk-default to lkk-setup-minor-mode."
  (lkk-elettronico-exit)
  (lkk-setup-minor-mode 'lkk-default lkk-default-cursor-color))

;;
;; lkk-j-mode as ordinary minor mode, 
;;

(defvar lkk-j-mode-map nil  "lkk-j-mode-map")

(defun lkk-setup-j-mode-map (&optional remap)
  (or remap
      (setq lkk-j-mode-map (make-sparse-keymap)))
  (let (keylist tem)
    ;; Fill with zenkaku-vector. Dirty.
    (setq tem 0)
    (while (< tem 128)
	(if (aref lkk-zenkaku-vector tem)
	    (define-key lkk-j-mode-map (char-to-string tem)
	      'lkk-L-self-insert))
	(setq tem (1+ tem)))

    ;; get romaji key
    (mapcar
     (lambda (cell)
       (setq tem (substring (car cell) 0 1))
       ;; Squeeze dup's with or(). length squeezed from 136 -> 34.
       (or (member tem keylist)
	   (setq keylist (cons tem keylist))))
     (append lkk-r2k-alist-C lkk-r2k-alist-V))

    ;; Bind romaji key
    (mapcar
     (lambda (kstr)
       (define-key lkk-j-mode-map kstr 'lkk-j-input)
       (define-key lkk-j-mode-map (lkk-toupper kstr) 'lkk-start-seeding))
     keylist)

    ;; Bind numeric -- mostyly for lkk-/-convert
    (mapcar
     (lambda (c)
       (define-key lkk-j-mode-map c 'lkk-num-input)
       ) (list "0" "1" "2" "3" "4" "5" "6" "7" "8" "9"))

;;;    (define-key lkk-j-mode-map "'" 'lkk-j-input) ; one exception
    (define-key lkk-j-mode-map "'" 'lkk-insert-j-squote-char)

    (define-key lkk-j-mode-map "," 'lkk-insert-j-comma)
    (define-key lkk-j-mode-map "." 'lkk-insert-j-period)
    (define-key lkk-j-mode-map "-" 'lkk-insert-j-minus-char)
    (define-key lkk-j-mode-map "[" 'lkk-insert-j-paren0)
    (define-key lkk-j-mode-map "]" 'lkk-insert-j-paren1)

    (define-key lkk-j-mode-map "Q" 'lkk-set-conv-beginning)
    (define-key lkk-j-mode-map "q" 'lkk-toggle-kana)

    (define-key lkk-j-mode-map "L" 'lkk-L-mode)
    (define-key lkk-j-mode-map "l" 'lkk-l-mode)

    (define-key lkk-j-mode-map " " 'lkk-convert)
    (define-key lkk-j-mode-map "/" 'lkk-start-/-convert)
    (define-key lkk-j-mode-map "\177"  'lkk-backward-delete-char)

    ;; Not yet complete. These two are almost the same.
    (define-key lkk-j-mode-map "\t" 'lkk-completion-convert)

    ;; Testing these two for lkk-multiple-click-crank()
    (define-key lkk-j-mode-map [S-return] 'lkk-l-return) ; l- function, but it's okay.
    (define-key lkk-j-mode-map [return] 'lkk-l-return) ; l- function,

    (define-key lkk-j-mode-map "\e " 'lkk-completion-convert)
    (define-key lkk-j-mode-map "\C-j" 'lkk-sit-back)

    ;; Insert lkk-last-sit-back content, Tue Oct  3 13:38:58 2000
    (define-key lkk-j-mode-map [C-M-return] 'lkk-insert-last-sat-back)

    (define-key lkk-j-mode-map "\C-xj" 'lkk-auto-fill-mode) ; skk emulation
    (define-key lkk-j-mode-map "\M-s" 'lkk-dump-dict)
    ))

(defvar lkk-j-mode-hook nil
  "Normal hook run when starting to lkk-j-mode.")

(defun lkk-j-mode (&optional arg)
  "\

`j' mode in LousyKanaKanji interface.

Keybindings:

\\{lkk-j-mode-map}"
  (interactive "P")
  (cond
   ((and arg
	 (if (> (prefix-numeric-value arg) 0)
	     lkk-j-mode (not lkk-j-mode)))
    ())					; Do nothing if already OK.
   (lkk-j-mode (lkk-j-mode-disable))
   (t (lkk-j-mode-enable))))

(defun lkk-j-mode-enable ()
  "Turn on lkk mode."
;;;  Not an ordinary minor mode.
;;;  (if (<= (string-to-int emacs-version) 19)
;;;      (add-hook 'change-major-mode-hook 'lkk-j-mode-disable nil)
;;;    (make-local-hook 'change-major-mode-hook)
;;;    (add-hook 'change-major-mode-hook 'lkk-j-mode-disable nil t))
  (lkk-setup-minor-mode 'lkk-j-mode))

(defun lkk-j-mode-disable ()
  "Turn off lkk mode."
;;;  Not an ordinary minor mode.
;;;  (if (> (string-to-int emacs-version) 19)
;;;      (remove-hook 'change-major-mode-hook 'lkk-j-mode-disable t)
;;;    (remove-hook 'change-major-mode-hook 'lkk-j-mode-disable))
  (interactive)
  ;;(lkk-sit-back);; lkk-all-minor-mode-disable does this (I believe).
  (lkk-all-minor-mode-disable))

(fset 'lkk-mode-all-exit 'lkk-j-mode-disable)

;;
;; lkk-J-mode -- kanji conversion mode as ordinary minor mode, 
;;

(defvar lkk-J-mode-map nil  "lkk-J-mode-map")

(defun lkk-setup-J-mode-map ()
  (setq lkk-J-mode-map (make-sparse-keymap))
  (let (keylist tem)

    (define-key lkk-J-mode-map " " 'lkk-reconvert)
;;; -> Lousy! should change this part
;;;    (define-key lkk-J-mode-map "a" 'lkk-select-cand)
;;;    (define-key lkk-J-mode-map "s" 'lkk-select-cand)
;;;    (define-key lkk-J-mode-map "d" 'lkk-select-cand)
;;;    (define-key lkk-J-mode-map "f" 'lkk-select-cand)
;;;    (define-key lkk-J-mode-map "j" 'lkk-select-cand)
;;;    (define-key lkk-J-mode-map "k" 'lkk-select-cand)
;;;    (define-key lkk-J-mode-map "l" 'lkk-select-cand)
;;;    (define-key lkk-J-mode-map ";" 'lkk-select-cand)
;;; -> Lousy! should change this part 

    (define-key lkk-J-mode-map "x" 'lkk-prev-cand)
    (define-key lkk-J-mode-map "X" 'lkk-purge-it)

    (define-key lkk-J-mode-map "\C-g" 'lkk-unwind-up)
    (define-key lkk-J-mode-map "\C-j" 'lkk-sit-back)
    ))

(defun lkk-update-insignia (insignia)
  "Change insignia to INSIGNIA if displayed on lkk-conv-beginning marker."
  (let ((targetsig (cond ((string= insignia lkk-insignia0) lkk-insignia1)
			 ((string= insignia lkk-insignia1) lkk-insignia0)
			 (t nil))))
    (cond
     ((null targetsig) (message "Buggy lkk-update-insignia call."))
     (lkk-conv-beginning
      (save-excursion
	(goto-char lkk-conv-beginning)
	(cond ((looking-at targetsig)
	       (delete-char 1)
	       (insert insignia))
	      (t (message "You may have deleted %s." targetsig)
		 ;;(lkk-sit-back)
		 )))))))

(defsubst lkk-J-mode-exit ()
  (setq lkk-J-mode nil)
  (force-mode-line-update))

;;
;; `q' -- aka katakana -- mode.
;; Uses also lkk-j-mode-map.
;;

(defvar lkk-q-mode-hook nil
  "Normal hook run when starting to lkk-q-mode.")

(defun lkk-q-mode (&optional arg)
  "\

`q' mode in LousyKanaKanji interface.

Keybindings:

\\{lkk-j-mode-map}"
  (interactive "P")
  (lkk-setup-minor-mode 'lkk-q-mode))

(defun lkk-q-mode-exit ()
  "Exit from `L' mode to old lkk mode."
  (interactive)
  (lkk-sit-back)
  (lkk-j-mode-enable))

;;
;; `l' mode (lkk-l-mode) as (minor-)minor-mode.
;; Almost the same as lkk-mode below, except mode name string.
;;

(defvar lkk-l-mode-map nil)

(defun lkk-setup-l-mode-map (&optional remap)
  (or remap
      (setq lkk-l-mode-map (make-sparse-keymap)))
  (let ((i 0))
    (while (< i 128)
      (if (aref lkk-ascii-vector i)
	  (define-key lkk-l-mode-map (char-to-string i) 'lkk-l-self-insert) )
      (setq i (1+ i))))
  (define-key lkk-l-mode-map "\C-i" 'lkk-l-TAB)
  (define-key lkk-l-mode-map "\C-j" 'lkk-l-mode-exit)
  ;;(define-key lkk-l-mode-map lkk-exit-key 'lkk-mode-all-exit) ; wrong.
  (define-key lkk-l-mode-map lkk-exit-key 'lkk-mode) ; This disables all minor.
  (define-key lkk-l-mode-map "\C-xj" 'lkk-auto-fill-mode) ; skk emulation

  ;; Testing these two,
  (define-key lkk-l-mode-map [S-return] 'lkk-l-return)
  (define-key lkk-l-mode-map [return] 'lkk-l-return)
  )

(defvar lkk-elettronico-l-hack-method
  'dabbrev
  ;; 'dictionary
  "Full auto completion method used in lkk-l-mode. Defaults 'dabbrev-expand.
Currently 'dabbrev or 'dictionary is possible.
In 'dabbrev method, your favorite dabbrev function can be specified
with variable lkk-dabbrev-func.")

(defvar lkk-dabbrev-func
  'dabbrev-expand
;;  'dabbrev-completion
;;  'hippie-expand
  "Symbol of function called if lkk-elettronico-l-hack-method is 'dabbrev.
Can be bound any of your favorite dabbrev commands such as

\\[dabbrev-expand] \(dabbrev-expand\)
\\[dabbrev-completion] \(dabbrev-completion\)

The \\[hippie-expand] \(hippie-expand\) may not be work correctly.")
  
(defun lkk-elettronico-l-hack ()
  "Do sort-of-overkill full-auto word completion in lkk-l-mode.
Called at the end of lkk-l-self-insert."
;; Light waight version of lkk-elettronico-j-hack().
;; There is no such point like lkk-conv-beginning in lkk-l-mode, so
;; take the preceding `word' in `dictionary' method.
  (cond
   ((sit-for 0 lkk-elettronico-delay) ; Non-NODISPLAY arg -- redisplay!
    (let ((threshold			; Share lkk-elettronico-/-conv-salt
	   (+ lkk-full-auto-hchar-threshold lkk-elettronico-/-conv-salt))
	  (wordstart) (wordlen 0))
      (save-excursion
	;; No (= (point) (point-min)) check: backward-word fails gracefully.
	(backward-word 1)
	(if (looking-at "[a-zA-Z_]+")	; perhaps not so slow here.
	    (setq wordstart (point))))
      (cond
       ((and wordstart (setq wordlen (- (point) wordstart))  
	    (> wordlen threshold))
	(cond 
	 ((eq lkk-elettronico-l-hack-method 'dabbrev)
	  (while (and (sit-for 0 lkk-elettronico-delay)
		      (condition-case nil
			  ;; How should this be called ?
			  ;; (funcall 'hippie-expand 1)
			  (funcall lkk-dabbrev-func 1)
			(error (if (sit-for 0 lkk-elettronico-delay)
				   (message "dabbrev faild"))))) t))
	 ((eq lkk-elettronico-l-hack-method 'dictionary)

	  ;; See also lkk-convert-region() for j/J-mode hack
	  (let* ((dtop 'lkk-okuri-nasi-dict-top)
		 (seed (buffer-substring-no-properties wordstart (point)))
		 (entry 
		  (let ((lkk-okuri nil))
		    (lkk-dict-search seed 'loose))))
	    (cond 
	     (entry
	      (lkk-replace-buffer-substring wordstart (point) (nth 0 entry)))
	     (t nil)))
	  ;;(lkk-convert-region wordstart (point) 1 'loose 'seedonly)
	  )
	 )))))))

;;
;; lkk-l-self-insert() -- alphanumeric input almost transparently
;;
;; Now lkk-l-self-insert() is called from lkk-/-mode, so, in
;; lkk-J-mode, we have to allow `S' etc. keys (having no mapping in
;; lkk-J-mode) to work as if lkk-/-mode is turned off because
;; lkk-/-mode is not switched off until lkk-sit-back().
;; Was it my fault to implement /-converting hack as an ordinary minor-mode ?
;;

(defun lkk-l-self-insert (arg)
  "`l' mode insert.  Direct input without conversion."
  (interactive "*p")
  ;; [double-SPC] hack
  (cond
   ((and (null  lkk-inhibit-multiple-keyclick)
	 (eq last-command-char ? )
	 (lkk-multiple-click-crank))
    ())
   ;; Or normal insert.
   (t
    ;; lkk-/-mode support. Sun Mar 26 17:43:16 2000
    (cond
     (lkk-J-mode
      ;; lkk-maneuver-old-keybind() sort of hack
      (let* ((skeys (lkk-get-command-key (this-command-keys)))
	     (cmd (let (lkk-/-mode)
		    ;; See lkk-maneuver-old-keybind() for this [return] hack.
		    (cond ((and (vectorp skeys) (equal skeys [return]))
			   (setq skeys "\r")))
		    (key-binding skeys)
		    )))
	(if (and cmd (not (eq cmd this-command)))
	    (call-interactively cmd) ))
      )
     (t (self-insert-command arg)))

    (or lkk-inhibit-elettronico-in-l-mode
	(null (lkk-isdabbrev last-command-char))
	(null lkk-full-auto)
	(lkk-elettronico-l-hack)))))

(defun lkk-l-TAB (arg)
  "Normally bound to TAB in `lkk-l-mode-map'.
Multiple keyclick supported. See also `lkk-inhibit-multiple-keyclick'."
  (interactive "*p")
  ;; [double-TAB] hack
  (cond
   ((and (null lkk-inhibit-multiple-keyclick)
	 (lkk-multiple-click-crank))
    ())
   (t (lkk-maneuver-old-keybind))))
(fset 'lkk-l-return 'lkk-l-TAB)

(defvar lkk-l-mode-hook nil
  "Normal hook run when starting to lkk-l-mode.")

(defun lkk-l-mode (arg)
  "\

`l' mode in LousyKanaKanji interface.

Keybindings:

\\{lkk-l-mode-map}"
  (interactive "p")			; p is okay. lkk-j-input() sees it.
;;;  (if lkk-/-converting
;;;      (lkk-j-input arg)
  (lkk-setup-minor-mode 'lkk-l-mode)
;;;    )
  )

(defun lkk-l-mode-exit ()
  "Exit from `l' mode to old lkk mode."
  (interactive)
  (lkk-sit-back)
  (lkk-j-mode-enable))

;;
;; `L' mode (lkk-L-mode) as (minor-)minor-mode.
;;

(defvar lkk-L-mode-map nil "Keymap for L mode.")

(defun lkk-setup-L-mode-map ()
  (setq lkk-L-mode-map (make-sparse-keymap))
  (let ((i 0))
    (while (< i 128)
      (if (aref lkk-zenkaku-vector i)
	  (define-key lkk-L-mode-map (char-to-string i) 'lkk-L-self-insert) )
      (setq i (1+ i)))
    (define-key lkk-L-mode-map "\C-j" 'lkk-L-mode-exit)
    ;;(define-key lkk-L-mode-map lkk-exit-key 'lkk-mode-all-exit) ; wrong.
    (define-key lkk-L-mode-map lkk-exit-key 'lkk-mode) ; This disables all minor.
    (define-key lkk-L-mode-map "\C-xj" 'lkk-auto-fill-mode) ; skk emulation
    ))

(defun lkk-L-self-insert (arg &optional vec)
  "`L' mode insert. Uses lkk-zenkaku-vector."
  (interactive "*p")
;;;  (if lkk-/-converting
;;;      (lkk-j-input arg)
    (let* ((key (aref (or vec lkk-zenkaku-vector) last-command-char)))
      (while (> arg 0)
	(insert key) (setq arg (1- arg))))
;;;    )

;;; Isearch support intended, but found to be annoying.
;;;  (cond
;;;   ((and (boundp 'lkk-in-read-string) ; must be t, so just boundp
;;;	 lkk-exit-minibuffer-on-sit-back ; user option defaulting t
;;;	 (interactive-p)		; comment, to exit by lkk-num-input()
;;;	 (> (buffer-size) 0)		; you entered some in the minibuffer
;;;	 )
;;;    (exit-minibuffer)))
  )

(defun lkk-num-input (arg)
  "Call lkk-L-self-insert unless lkk-/-converting is non-nil.
Otherwise lkk-j-input is called while lkk-/-converting, and
not in lkk-J-mode \(conversion mode\)."
  (interactive "*p")
  (let ((key last-command-char))
    (cond
;;;     ((and lkk-/-converting
;;;	   (null lkk-J-mode))
;;;      ;; lkk-conv-beginning
;;;      (lkk-j-input arg))
     (t 
      ;; Feel like sit-back is not `must' here on the contrary of
      ;; hiragana after kanji conv, I don't know why. On the
      ;; other hand, have other feeling that I should give priority
      ;; to integrity of control semantics, and I took the latter
      ;; thought here.
      (if lkk-J-mode (lkk-sit-back))
      (lkk-L-self-insert arg)))
    (if lkk-full-auto (lkk-elettronico-j-hack))
    ))

(defvar lkk-L-mode-hook nil
  "Normal hook run when starting to lkk-L-mode.")

(defun lkk-L-mode (arg)
  "\

`L' mode in LousyKanaKanji interface.

Keybindings:

\\{lkk-L-mode-map}"
  (interactive "p")			; p is okay. lkk-j-input() sees it.
;;;  (if lkk-/-converting
;;;      (lkk-j-input arg)
    (lkk-setup-minor-mode 'lkk-L-mode)
;;;    )
  )

(defun lkk-L-mode-exit ()
  "Exit from `L' mode to old lkk mode."
  (interactive)
  (lkk-sit-back)
  (lkk-j-mode-enable))

;; These special functions are used in lkk-j-mode, but
;; must be removed in future version.

(defvar lkk-j-space-char "")
(defvar lkk-j-comma-char "")
(defvar lkk-j-period-char "")
(defvar lkk-j-minus-char "")
(defvar lkk-j-paren0-char "")
(defvar lkk-j-paren1-char "")
(defvar lkk-j-squote-char "")

(defun lkk-insert-a-thing (&optional str arg)
  "Insert STR by ARG times. ARG should be number."
  (while (> arg 0) (insert str) (setq arg (1- arg))))

(defun lkk-insert-j-space (arg)
  "Insert `lkk-j-space-char' by ARG times."
  (interactive "*p")
  (lkk-sit-back) (lkk-insert-a-thing lkk-j-space-char arg))

(defun lkk-insert-j-comma (arg)
  "Insert `lkk-comma-char' by ARG times."
  (interactive "*p")
;;;  (if lkk-/-converting
;;;      (lkk-j-input arg)
  (lkk-sit-back) (lkk-insert-a-thing lkk-j-comma-char arg)
;;;  )
  )

(defun lkk-insert-j-period (arg)
  "Insert `lkk-period-char' by ARG times."
  (interactive "*p")
;;;  (if lkk-/-converting
;;;      (lkk-j-input arg)
  (lkk-sit-back) (lkk-insert-a-thing lkk-j-period-char arg)
;;;    )
  )
  
(defun lkk-insert-j-minus-char (arg)
  "Insert `lkk-period-char' by ARG times."
  (interactive "*p")
  ;; (lkk-sit-back) ; No ! We have to enter `otohiki' in kanji yomi.
;;;  (if lkk-/-converting
;;;      (lkk-j-input arg)
  (lkk-insert-a-thing lkk-j-minus-char arg)
;;;    )
  )

(defun lkk-insert-j-paren0 (arg)
  "Insert `lkk-space-char' by ARG times."
  (interactive "*p")
;;;  (if lkk-/-converting
;;;      (lkk-j-input arg)
  (lkk-sit-back) (lkk-insert-a-thing lkk-j-paren0-char arg)
;;;    )
    )

(defun lkk-insert-j-paren1 (arg)
  "Insert `lkk-space-char' by ARG times."
  (interactive "*p")
;;;  (if lkk-/-converting
;;;      (lkk-j-input arg)
    (lkk-sit-back) (lkk-insert-a-thing lkk-j-paren1-char arg)
;;;    )
  )

(defun lkk-insert-j-squote-char (arg)
  "Insert `lkk-j-squote-char' by ARG times.
While `lkk-conv-beginning', `lkk-j-input' is called
and the key input is taken as a part of a romaji seed."
  (interactive "*p")
  (cond (lkk-conv-beginning
	 (lkk-j-input arg))
	(t
	 (lkk-sit-back) (lkk-insert-a-thing lkk-j-squote-char arg)
	 )))

(defun lkk-toggle-kana ()
  "Toggles hiragana / katakana input mode."
  (interactive)
;;;  (if lkk-/-converting
;;;      (lkk-j-input 1)
    (if lkk-q-mode (lkk-q-mode-exit) (lkk-q-mode))
;;;    )
  )

;;
;; lkk-mode: very basic mode. 
;; Almost NOP except setting header part of minor mode string, and invoking
;; entity of lkk-initial-mode.
;; Uses no key map. 
;;

(defvar lkk-initial-mode 'lkk-j-mode
  "Initial mode. Defaults 'lkk-j-mode. Maybe you can speciry lkk-l-mode etc.")

(defvar lkk-mode-hook nil
  "Normal hook run when starting to lkk mode.")

;; This is key-binding() compatible.
;; (defun lkk-lookup-key (key)
;;  "Lookup KEY in "
;;  (let ((map (or (current-local-map) global-map)))
;;    (or (lookup-key map key t)
;;	(lookup-key global-map key t))))

(defun lkk-mode (&optional arg)
  "\

LousyKanaKanji interface basic layer.

lkk-mode: very basic mode. 
Almost NOP except setting header part of minor mode string, and invoking
entity of lkk-initial-mode.
Uses no key map. 
"
  (interactive "P")

  (cond
   ((and arg
	 (if (> (prefix-numeric-value arg) 0)
	     lkk-mode (not lkk-mode)))
    ())					; Do nothing if already OK
   (lkk-mode
    (setq lkk-mode nil)
    (lkk-mode-all-exit)
    (force-mode-line-update))
   (t 
    (setq lkk-mode t)
    (force-mode-line-update)
    ;; This is not handled by lkk-setup-minor-mode().
    (run-hooks 'lkk-mode-hook)

    (lkk-j-mode-enable) ; Call this, then lkk-initial-mode (dirty).
    (cond ((eq lkk-initial-mode 'lkk-l-mode) (lkk-l-mode 1))
	  ((eq lkk-initial-mode 'lkk-L-mode) (lkk-L-mode 1))
	  ((eq lkk-initial-mode 'lkk-q-mode) (lkk-q-mode))
	  ((eq lkk-initial-mode 'lkk-j-mode) (lkk-j-mode-enable))
	  ))))

;;;
;;; Setup minor-mode alists here. Order is very significant.
;;;

(defvar lkk-minor-mode-strings '((lkk-/-mode "/")
				 (lkk-j-mode "j")
				 (lkk-J-mode "*")
				 (lkk-q-mode "q")
				 (lkk-l-mode "l")
				 (lkk-L-mode "L")
				 (lkk-full-auto "")
				 (lkk-mode " "))
  "Alist containing modeline strings for each lkk submode.
You may use preferred strings whatsoever.")

(defsubst lkk-get-minor-mode-string (mode)
  (assoc mode lkk-minor-mode-strings))

(defun lkk-register-minor-mode (modelist)
  "Place MODELIST \(modesym modestr\) at the top of minor-mode-alist."
  (or (assq (car modelist) minor-mode-alist)
      (setq minor-mode-alist
	    (cons modelist minor-mode-alist))))

(defun lkk-register-minor-mode-map (modesym map)
  "Place MODESTR at the top of minor-mode-alist."
  (or (assq modesym minor-mode-map-alist)
      (setq minor-mode-map-alist
	    (cons (cons modesym map) minor-mode-map-alist))))

(progn
  ;; Setup minor-mode strings.
  (mapcar (lambda (c)
	    (lkk-register-minor-mode c))
	  lkk-minor-mode-strings)

  ;; Setup minor-mode-map's

  ;; j-mode
  (or lkk-j-mode-map
      (lkk-setup-j-mode-map))
  (lkk-register-minor-mode-map 'lkk-j-mode lkk-j-mode-map)

  ;; l-mode
  (or lkk-l-mode-map
      (progn (lkk-setup-l-mode-map)	; Also setup lkk-/-mode-map
	     (setq lkk-/-mode-map (copy-keymap lkk-l-mode-map))
	     (define-key lkk-/-mode-map " " 'lkk-convert)))
  (lkk-register-minor-mode-map 'lkk-l-mode lkk-l-mode-map)

  ;; /-mode
  ;;
  ;; Must precede J-mode below (J-mode must win), but this is not good
  ;; enough:  Keys like `X' mapped in lkk-J-mode-map works, but `S' not
  ;; mapped in lkk-J-mode-map won't work as in plain lkk-j-mode:
  ;; This is remedied in lkk-l-self-insert() (See the fun).
  (lkk-register-minor-mode-map 'lkk-/-mode lkk-/-mode-map)

  ;; J-mode (conversion mode)
  ;;
  ;; lkk-J-mode-map *MUST* precede lkk-j-mode-map in minor-mode-map-alist, or
  ;; we can't override lkk-j-mode-map.
  (or lkk-J-mode-map
      (lkk-setup-J-mode-map))
  (lkk-register-minor-mode-map 'lkk-J-mode lkk-J-mode-map)

  ;; q-mode
  (lkk-register-minor-mode-map 'lkk-q-mode lkk-j-mode-map)

  ;; L-mode
  (or lkk-L-mode-map
      (lkk-setup-L-mode-map))
  (lkk-register-minor-mode-map 'lkk-L-mode lkk-L-mode-map)

  ;; lkk-mode --the basic layer
  ;; Just one char is good enough.
  )

;;;
;;; Dictionary handlers
;;; Sat Mar 11 18:23:46 2000
;;;

;; The `buffer' type interface method uses these:

(defvar lkk-okuri-ari-dict-top nil)
(defvar lkk-okuri-nasi-dict-top nil)

;;
;; The `buffer' type dictionary loader used by dictionary search functionsions
;; in lkk-skk-dichead.el and lkk-lkk-dichead.el, and aliased to lkk-dict-search.
;;
;; Even if the dictionary file is skk-formatted, the filename is assumed to
;; be lkk-dict (defaults "~/.lkk-dict") in view of safty.
;;

(defun lkk-dict-get-dict-version ()
  "Subr called by lkk-dict-read-dict. Returns lkk-dict-version number as string.
Assumed to be in dictionary buffer already."
  (let ((case-fold-search t))
    (save-excursion
      (goto-char (point-min))
      (if (search-forward-regexp	; first 2000 chars(bytes?) only.
	   (format
	    "^;;[ \t]*\\(%s\\)[ \t]*VERSION[ \t]*\\(%s\\)"
	    lkk-dict-header lkk-version ) 2000 t)
	  (setq lkk-dict-type lkk-dict-method ; just copy method, maybe lousy.
		lkk-dict-version (buffer-substring-no-properties
				  (match-beginning 2)
				  (match-end 2)))))))

;; lkk-insert-file-contents() -- Gzip'ed file (dictionary) insertion support
;; Tue Mar 14 15:33:45 2000 s.n.

(defun lkk-insert-file-contents (file)
  "Insert FILE.gz content if FILE it ever exists, or just insert FILE.
Gunzip is done via jka-compr or jam-zcat. No coding system fiddle."
  (let* ((file (expand-file-name file))
	 (gname (cond ((or
			;; Maybe the zip gizmoz isn't setup, or I am just lazy
			;; (string-match "\\.bz2" file)
			;; (string-match "\\.Z" file)
			(string-match "\\.gz" file)) file)
		      (t (concat file ".gz"))))
	 (zip-gizmo (or (featurep 'jka-compr) (featurep 'jam-zcat)))
	 (gexists (file-readable-p gname))
	 (targ (cond ((and gexists zip-gizmo) gname)
		     (t (cond ((file-readable-p file) file)
			      (t nil) )))))
  (cond
   ((and gexists (null zip-gizmo))
    (error "No zip package such as jka-compr or jam-zcat"))
   ((null targ) (error "Can't read %s nor %s" file gname))
   (t 
    ;; (message "G:%s gexists:%s FILE:%s GNAME:%s TARG:%s"
    ;;          zip-gizmo gexists file gname targ)
    (let ((inhibit-quit t)) (insert-file-contents targ))
    ))))

(defun lkk-dict-read-dict ()
  "Read in lkk-dict destructively, and return the buffer object of dict buffer.
Low level routine. Don't call directly. Use lkk-skk-dict-get-dict instead."
  (let ((inhibit-quit t)
	(buf (get-buffer-create lkk-dict-buffer))
	;; Must be any other nice way to do it.
	(lkk-dict (if (eq lkk-dict-method 'lkk-skk-dict-buffer)
		      "~/.skk-jisyo" lkk-dict)))
    (save-excursion
      (set-buffer buf) (erase-buffer)
      (condition-case nil
	  (progn
	    (message "Reading lkk dictionary %s..." lkk-dict)
	    (lkk-insert-file-contents (expand-file-name lkk-dict)))
	(error
	 (kill-buffer lkk-dict-buffer)
	 (setq lkk-okuri-ari-dict-top nil ; setq is good enough.
	       lkk-okuri-nasi-dict-top nil)
	 (error "Can't read lkk dictionary %s..." lkk-dict)))
      (message "Reading lkk dictionary %s ...done" lkk-dict) (sit-for 0)
      (widen)
      (goto-char (point-min))

      (let ((case-fold-search t))
	(lkk-dict-get-dict-version)
	(save-excursion (search-forward ";; okuri-ari" (point-max) t)
			(forward-line 1)
			(setq lkk-okuri-ari-dict-top (point-marker)))
	(save-excursion (search-forward "\n;; okuri-nasi" (point-max) t)
			(forward-line 1)
			(setq lkk-okuri-nasi-dict-top (point-marker))))

      (setq case-fold-search nil)	; nil while actually running.

      ) buf))

;; This may have to be in dict method layer ?
(defun lkk-dict-get-dict ()
  "Return dict buffer as buffer obj."
  (or (get-buffer lkk-dict-buffer)
      (lkk-dict-read-dict)))

(defvar lkk-dict-dump-file "~/.lkk-dict.today"
  "Dictionary filename on which the current dictionary image is dumped.
Current version does not overwrite onto dictionary image, and has no
facility to do overwrite-save on kill-emacs. But, almost same operation
can be done by modifying lkk-dict-dump-file and kill-buffer-hook setup.
You know what I mean, so, I don't explain the detail of it here.")

(defun lkk-dict-dump-dict ()
  "Dump dictionary buffer image onto lkk-dict-dump-file
Temporary user command until real dictionary upadte system is completed."
  (let ((inhibit-quit t)
	(buf (get-buffer lkk-dict-buffer)))
    (cond
     ((and buf (buffer-modified-p buf)) ; Reload MUST not occur here
      (save-excursion
	(set-buffer buf)
	;; No coding system fiddle here.
	(condition-case nil
	    (prog2
		(message "lkk dictionary dumping...")
		(write-region (point-min) (point-max)
			      (expand-file-name lkk-dict-dump-file))
	      (message "lkk dictionary dumped as %s" lkk-dict-dump-file)
	      (set-buffer-modified-p nil)
	      )
	  (error (error "Dump onto %s faild." lkk-dict-dump-file)))
	;; (message "No lkk dict buffer to be dumped.") ; No Msg. here
	)))))

(defun lkk-dump-dict ()
  "Dumps lkk dictionary onto lkk-dict-dump-file.
Adequate dump method is selected accoding to lkk-dict-method."
  (interactive)
  (cond
   ((eq lkk-dict-method 'lkk-dict-alist)
    ;; (require 'lkk-lkk-dichead) ; should have been loaded already.
    (cond ((and lkk-big-dict lkk-big-dict-okuri)
	   (lkk-dict-alist-dump-dict))
	  (t (message "No big-dict alist to dump in core."))))
   ;; else 'lkk-skk-dict-buffer or 'lkk-dict-buffer
   (t 
    (cond ((get-buffer lkk-dict-buffer)
	   (lkk-dict-dump-dict))
	  (t (message "No dict buffer to dump."))))))

;; This may be lkk-edit-dict in future version. 
(defun lkk-show-lkk-dict-buffer (arg)
  "Display lkk dictionary buffer if ever exists.
ARG forces to read in `lkk-dict'."
  (interactive "P")
  (cond
   ((eq lkk-dict-method 'lkk-dict-alist)
    (message "No dict buffer available in %s method" 'lkk-dict-alist))
   ((or (get-buffer lkk-dict-buffer)
	(and arg (lkk-dict-get-dict) ))
    ;; This window selection makes sense if " \\*.*lkk" sort of
    ;; regex is not included in `special-display-regexps', and
    ;; harms nothing even if the regex is included.
    (let ((w (selected-window)))
      (switch-to-buffer-other-window lkk-dict-buffer)
      ;;(display-buffer lkk-dict-buffer)
      ;; To goto bob is easy, so,
      (goto-char lkk-okuri-nasi-dict-top)
      (sit-for 0 500)
      (select-window w))
    )
   (t (message "Buffer `%s\' not yet exist" lkk-dict-buffer))))

;;
;; lkk-dump-dict-with-timer() -- dumps dict with timer
;; every X seconds if emacs is idle more than Y seconds.
;; s.n. Sat Mar 11 17:32:09 2000
;;

;; E19.28 seems to need this.
(eval-when-compile
  (require 'timer))

(defvar lkk-dump-dict-automatic nil
  "Non-nil value turns on automatic dictionary dump.
See also `lkk-dump-dict-timer' and `lkk-dump-dict-idle-timer'.")

(defvar lkk-dump-dict-timer 180
  "Interval between automatic `lkk-dump-dict' trial.
Even if this timer expired,  the dump would not executed unless
`lkk-dump-dict-idle-timer' expired by idling of emacs.
Run `lkk-dump-dict-with-timer' if you modified this value.")

(defvar lkk-dump-dict-idle-timer 4
  "Idle timer which should be expired before automatic execution of
`lkk-dump-dict' by expiration of `lkk-dump-dict-timer'.
If this idle timer did not expire, the dump would not be executed.
Run `lkk-dump-dict-with-timer' if you modified this value.")

(defvar lkk-dump-dict-timer-obj nil
  "Our timer object, set by `lkk-dump-dict-with-timer'.")

(defun lkk-dump-dict-with-timer ()
  "Startup special timer for automatic dictionary dump.
See also `lkk-dump-dict-automatic', `lkk-dump-dict-timer'
`lkk-dump-dict-idle-timer', `lkk-cancel-dump-dict-with-timer' etc."
  (interactive)
  (let ((re (if lkk-dump-dict-timer-obj "re" "")))
    (lkk-cancel-dump-dict-with-timer)	; Cancel, if some other is working.
    (if (setq lkk-dump-dict-timer-obj
;;; emacs-19.28 has none.
;;;	      (run-with-timer lkk-dump-dict-timer
	      (run-at-time (format "%d sec" lkk-dump-dict-timer)
			   lkk-dump-dict-timer 'lkk-dump-dict-on-idle))
	(message "lkk dictionary dumper timer job %sstarted." re))))

(defun lkk-cancel-all-dump-dict-with-timer ()
  "Cancel all the `lkk-dump-dict-on-idle' timer jobs."
  (interactive)
  (cancel-function-timers 'lkk-dump-dict-on-idle)
  (setq lkk-dump-dict-timer-obj nil))

(defun lkk-dump-dict-on-idle ()
  "Ther automatic `lkk-dump-dict' handler."
  (if (sit-for lkk-dump-dict-idle-timer)
      (lkk-dump-dict)))

(defun lkk-cancel-dump-dict-with-timer ()
  "Cancel the timer object `lkk-dump-dict-timer-obj'."
  (interactive)
  (cond (lkk-dump-dict-timer-obj
	 (cancel-timer lkk-dump-dict-timer-obj)
	 (setq lkk-dump-dict-timer-obj nil))))

(and lkk-dump-dict-automatic		; Startup timer job if requested.
     (lkk-dump-dict-with-timer))

;;;
;;; Low level, but generic routines for dict string <=> list
;;;

;; Recursive version fails on long line. Made `while' version.
(defun lkk-str-to-list (str &optional delim)
  "Shred STR into word list. Default delimiter is space or tab.
Optional 2nd arg DELIM is used if it's non-nil."
  (let ((list ()) (delim (or delim "[ \t]")) n)
    (while (setq n (string-match delim str)) ; equiv (match-beginning 0)
      (setq list (cons (substring str 0 n) list)
	    str (substring str (match-end 0))))
    (setq list (cons str list))		; add the rest too.
    (nreverse list)))

(defun lkk-get-current-line ()
  "Return current whole line string. Newline is chopped."
  (buffer-substring-no-properties
   (save-excursion (beginning-of-line) (point))
   (save-excursion (end-of-line)       (point))))

;; from dired.el
(defun lkk-replace-in-string (regexp newtext string)
  "Replace REGEXP with NEWTEXT all over STRING. NEWTEXT is taken literally."
  (let ((result "") (start 0))
    (while (string-match regexp string start)
      (setq result
	    (concat
	     result (substring string start (match-beginning 0)) newtext)
       start (match-end 0)))
    (concat result (substring string start))))

;; and its `just once' version
(defun lkk-replace-in-string-one (regexp newtext string)
  "Replace REGEXP with NEWTEXT in STRING just once. NEWTEXT is taken literally."
  (let ((result ""))
    (if (string-match regexp string)
	(concat (substring string 0 (match-beginning 0)) 
		newtext
		(substring string (match-end 0)))
      string)))

;;
;; lkk-dict-alist-delete() -- delete dictionary entry if yomi is matched.
;;
;; Moved from lkk-lkk-dichead.el as generic routine for use in
;; generic assoc-delete function.
;;
;; lkk-dict-alist-delete() can be used in horizontal update of okuri-ari
;; record. This job is not dictionary-method dependent.
;; 

(defun lkk-dict-alist-delete (key alist)
  "Delete entry having KEY in dictionary ALIST, and return the dictionary.
Side effect should not be always assumed because this uses `delq'.
Use set/setq to bind returned value to a variable."
  (let ((record (assoc key alist)))
    (delq record alist)))

;; This old version's skelton was taken from some function in
;; alist.el in apel (was it ?), but found to be very slow (about 2 times 
;; slower than the current version above). Original called equal()
;; to compare element.
;; (defun lkk-dict-alist-delete (yomi dict)
;;   (if (string= yomi (car (car dict)))
;;       (cdr dict)			; dict is not modified in this case.
;;     (let ((front dict)
;; 	  (rear (cdr dict)))
;;       (catch 'hit
;; 	(while rear
;; 	  (cond ((string= yomi (car (car rear)))
;; 		 (setcdr front (cdr rear))
;; 		 (throw 'hit dict)))
;; 	  (setq front rear
;; 		rear (cdr rear)))
;; 	dict))))

(defun lkk-dict-check-record (&optional record kanjionly)
  "Check if lkk-kanji-index is within the RECORD boundary.
Optional arg KANJIONLY forces a check of the kanji boundary only.
If RECORD is nil, lkk-current-record is checked."
  ;;(length '("" "" "礯" "ݤ" "ߤ")) => 5
  (and (if kanjionly (< 0 lkk-kanji-index) (<= 0 lkk-kanji-index))
       (< lkk-kanji-index (length (or record lkk-current-record)))))

;;;
;;; Kanji Conversion -- User Interface
;;; Not yet complete at all.
;;;

;; Recording and restoring last sat-back data.
;; Tue Oct  3 13:58:33 2000

(defun lkk-insert-last-sat-back (arg)
  "Insearts content of `lkk-last-sit-back' ARG times."
  (interactive "p")
  (cond
   ((and lkk-last-sit-back (stringp lkk-last-sit-back))
    (lkk-insert-a-thing lkk-last-sit-back arg)
    (message "`lkk-last-sit-back' is `%s'." lkk-last-sit-back))
   ))

(lkk-deflocalvar lkk-last-sit-back
		 "Contains `lkk-conv-beginning' to point on sit-back.")

(defun lkk-sit-back (&optional arg)
  "Confirm all partial status, and sit back."
  (interactive "*P")
  (let ((inhibit-quit t)
	(okuri-conv			; Too many or's ?
	 (or lkk-okuri lkk-okuri-beginning lkk-okuri-root lkk-okuri-h)))

    ;; Debugger
    (if lkk-debug-hook (run-hooks 'lkk-debug-hook))

    ;; Kakutei -- dict update must be done here.
    (cond
     (lkk-J-mode
      ;; Sometimes lkk-kanji-index is broken here: perhaps bugs are somewhere.
      ;; This is a way to sooth lkk-dict-update:
      ;; (or lkk-kanji-index (setq lkk-kanji-index 0))
      ;; but unwind-protect() is used here instead.
      (unwind-protect
	  (cond
	   ((and 
	     lkk-kanji-index
	     (if okuri-conv
		 ;; In okuri-hack, this is the record used in display:
		 lkk-current-display-record
	       lkk-current-record)
	     ;; Current feature is `throw-away-server-record'.
	     (or
	      ;; In non-okuri, this not() returns t. Was that okay ?
	      (not (and okuri-conv (eq lkk-server-affair 'server-hit)))
	      ;; User explicitly registered something. The record should be
	      ;; set by lkk-recursive-register()
	      ;; or lkk-recursive-register-okuri() in reconvert case.
	      lkk-current-user-record)
	     )
	    (let ((newrelease
		   (if okuri-conv
		       (lkk-dict-new-record-okuri)
		     (lkk-dict-new-record))      
		   ))
	      (if newrelease (lkk-dict-update newrelease)))))
	(lkk-J-mode-exit))))

    ;; Same check above, but do like this for readability.
    (cond
     ;; strict check for okuri here. Not by lkk-okuri-converting-p()
     (okuri-conv
					; Gnuess this must not not enough...
					; What do I forget ?
      ;; Erase stale insignia2, the init vars.
      (lkk-okuri-converting-p 'init-all-on-success)
      (lkk-reset-okuri)
      ))

    ;; Then, reset simple data first.
    (setq lkk-current-record nil
	  lkk-kanji-index nil
	  lkk-current-display-record nil
	  lkk-current-server-record nil
	  lkk-current-user-record nil
	  lkk-server-affair nil
	  lkk-convert-hit-bottom nil
	  )
    (if (boundp 'lkk-dict-buffer-the-latest-hit-point)
	(setq lkk-dict-buffer-the-latest-hit-point nil))

    (cond
     (lkk-conv-beginning
      (save-excursion
	(goto-char lkk-conv-beginning)
	(if (or (looking-at lkk-insignia0) (looking-at lkk-insignia1))
	    (delete-char 1)))

      ;; Record lkk-last-sit-back, Tue Oct  3 13:38:58 2000
      (setq lkk-last-sit-back
	    (buffer-substring-no-properties lkk-conv-beginning (point)))

      ;;(setq lkk-conv-beginning nil)
      (lkk-clear-marker 'lkk-conv-beginning)
      ))

    (cond
     (lkk-j-beginning
      (let ((char-left
	     (buffer-substring-no-properties lkk-j-beginning (point))))
	(cond ((string-match "^[nm]+$" char-left)
	       (lkk-replace-buffer-substring lkk-j-beginning
					     (point)
					     (if lkk-q-mode "" "")
					     nil 'noundo))
	      ((string-match "\\([a-zA-Z]\\)\\1$" char-left)
	       (lkk-replace-buffer-substring lkk-j-beginning
					     (point)
			      (if lkk-q-mode "" "")
			      nil 'noundo))
	      ((lkk-same-line-p lkk-j-beginning (point))
		 ;;(delete-region lkk-j-beginning (point))
	       (lkk-replace-buffer-substring lkk-j-beginning
					     (point)
					     nil nil 'noundo))
	      ))))

    (lkk-clear-marker 'lkk-j-beginning)
    (setq
     ;; lkk-/-converting nil
     ;; lkk-j-beginning nil
     lkk-/-mode nil
     lkk-j-numarg 1)

    (lkk-do-auto-fill)			; do auto-fill if necessary

    ;; lkk-isearch support. Sees special var set by isearch-lkk-read-string().
    (cond
     ((and (boundp 'lkk-in-read-string) ; must be t, so just boundp
	   lkk-exit-minibuffer-on-sit-back ; user option defaulting t
	   (interactive-p)		; Don't exit unless it's explicit C-j
	   (> (buffer-size) 0))		; you entered some in the minibuffer
      (exit-minibuffer)
      ))
    ;; conversion status cleaned up here.

    ;; Debugger
    (if lkk-sit-back-hook (run-hooks 'lkk-sit-back-hook))

    ))

(fset 'lkk-kakutei 'lkk-sit-back)

;; (lkk-deflocalvar lkk-sit-back-hook
(defvar lkk-sit-back-hook nil		; Currently Mostly for debug.
  "Hook run at the end of `lkk-sit-back'.")

(defvar lkk-exit-minibuffer-on-sit-back t
  "Determines whether to exit isearch minibuffer on lkk-sit-back \(kakutei\).
Defaults t.
User option.")

;;
;; lkk-dict-new-record() -- Reproduce new updated record for sit-back.
;;
;; Should not destructvely modify lkk-current-record or it's sublist here.
;;

(defun lkk-dict-new-record ()
  "Reconstruct lkk-current-record for accoding to lkk-kanji-index."
  (if (lkk-dict-check-record lkk-current-record 'kanjionly)
      (let* ((yomi     (car lkk-current-record))
	     (hitkanji (nth lkk-kanji-index lkk-current-record))
	     (kanjilis (cdr lkk-current-record)))
	(cons yomi
	      (cons hitkanji
;;; Now we may destructively modify lkk-current-record
;;; See lkk-dict-alist-search() in lkk-lkk-dichead.el.
;;;		      (delete hitkanji (copy-list kanjilis))
		    (delete hitkanji kanjilis)
		    )))))

(defun lkk-set-conv-beginning (&optional arg no-sit-back)
  "Normally bound to `Q' in lkk-j-mode-map."
  (interactive "*P\nP")			; arg unused yet
  (or no-sit-back
      (if lkk-conv-beginning (lkk-sit-back)))
  (setq lkk-conv-beginning (point-marker))
  ;; lkk-update-insignia unused. Good enough.
  (insert lkk-insignia0))

;;
;; lkk-convert(), and its entity lkk-convert0() -- An entry point of the engine
;;                                                 mainly for okuri-nasi.
;;
;; Now I am convinced that non-auto-okuri-ari conversion should not be
;; implemented in lkk.
;; Thu Mar 30 02:20:44 2000, s.n.
;;

;; Currently this does not handle okuri-ari conversion.
(defun lkk-convert (&optional arg)
  "Henkan command. Normally bound to SPC in lkk-j-mode-map.
Explicit numeric ARG such as M-3, C-4 etc. converts such chars before point."
  (interactive "*P")
  (cond ((and (null lkk-inhibit-multiple-keyclick)
	      (lkk-multiple-click-crank))
	 ())
	(t (lkk-convert0 arg))))

;;
;; lkk-convert()'s entity.
;;

(defun lkk-convert0 (arg)
  "The entity of the `lkk-convert'."
  (let ((inhibit-quit t))
    ;; Debugger
    (if lkk-debug-hook (run-hooks 'lkk-debug-hook))

    ;; Into J-mode
    ;; (setq lkk-/-converting nil) ; not here

    ;; (lkk-setup-minor-mode 'lkk-J-mode) ; not here, but when any
    ;; conv is done.  Now lkk-j-search-server() (lkk-skkserv.el), or
    ;; lkk-recursive-register() sets lkk-J-mode by theiaselves if it
    ;; is needed.

    ;; Romaji is left -- Must clobber them first, and,
    (cond
     (lkk-j-beginning
      (let ((pc (preceding-char))
	    ;; Similar hack in start-seeding(), similar one in lkk-j-input()...
	    (clobberit (lambda (k h)
			 ;; Is this hacking one char good enough ?
			 ;; (setq lkk-j-beginning nil)
			 (lkk-clear-marker 'lkk-j-beginning)
			 (let ((buffer-undo-list t))
			   (delete-backward-char 1))
			 (insert (if lkk-q-mode k h))
			 ;; (lkk-start-seeding-okuri arg) ;; Not, here.
			 )))
	(cond
	 ((null (lkk-same-line-p lkk-j-beginning (point)))
	  ;; (lkk-sit-back) ; To enable moving back to lkk-insignia0 line...
	  ;; Use error() here, not to proceed.  Catch-throw may introduce
	  ;; confusion in this function...
	  (error
	   "Maybe you'are at wrong point. Move back line where %s sits"
	   lkk-insignia0))
	 ((eq (preceding-char)		; stale double consonant
	      (lkk-tolower last-command-char))
	  (funcall clobberit "" ""))
	 ((or (eq (preceding-char) ?n)	; stale `n'
	      (eq (preceding-char) ?m))
	  (funcall clobberit "" "" ))
	 (t 
	  (lkk-replace-buffer-substring lkk-j-beginning (point) nil nil 'noundo)
	  ;; Or it may cause trouble afterwards.
	  ;;(setq lkk-j-beginning nil)
	  (lkk-clear-marker 'lkk-j-beginning)
	  )
	 ))))

    ;; Then conversion cond...
    (cond
     ;; j-space exception.
     ((and (null lkk-conv-beginning) (null arg))
      (lkk-insert-j-space (prefix-numeric-value arg)))

;;; Deprecated.  This worked, but wasn't found to be so nice because 
;;; *4-letter-word*-before-cursor-conversion is often required in
;;; japanese, and it is inconvenient if just C-u is not taken as number
;;; `four'.  Further, lkk-convert-region() is (interactive "*r\np")
;;; command now, and can directly be executed with 
;;; M-x lkk-convert-region RET after certain region is marked.
;;;
;;;   ;; Convert region on plain C-u hit
;;;   ((and arg (listp arg) 		; Just C-u only
;;;	 (condition-case nil (region-end) (error nil))) ; nil if no mark at all.
;;;    (lkk-convert-region (region-beginning) (region-end) 1)
;;;    ;; (lkk-update-insignia lkk-insignia1) ; done in lkk-convert-region()
;;;    )

     ;; Convert previous ARG chars
     ;; lkk-henkan-with-arg equivalent.
     ((and arg (> (setq arg (prefix-numeric-value arg)) 1))
      (let ((rb (save-excursion (backward-char arg) (point)))
	    (re (point)))
	(if (lkk-convert-region rb re 1)
	    (lkk-setup-minor-mode 'lkk-J-mode))
	;; (lkk-update-insignia lkk-insignia1) ; done in lkk-convert-region()
	))

     ;; The same line check.
     ((and lkk-conv-beginning
	   (null (lkk-same-line-p lkk-conv-beginning (point))))
      (lkk-sit-back)
      (message "Point must be in the same line as %s sits. Gave up conversion."
	       lkk-insignia0))

     ;; Do non-auto okuri-ari conversion may be located around here, in an
     ;; `ordinary' input method,  but lkk is not an `ordinary' IM.
     ;; (lkk-inhibit-auto-okuri-conv (lkk-convert-okuri-crank lkk-okuri-end))

     ;; Ordinary `henkan' (non-okuri-ari)
     ;; 
     (t
      (cond
       (lkk-conv-beginning
	(cond
	 ((lkk-convert-region
	   (+ (length lkk-insignia0) lkk-conv-beginning) (point) 1)
	  ;; (lkk-update-insignia lkk-insignia1) ; done in lkk-convert-region()
	  t)
	 ;; Try server (or large dick -- nop, in lkk), then registering.
	 (t 
	  (let ((seed
		 (buffer-substring-no-properties
		  (+ (length lkk-insignia0) lkk-conv-beginning) (point))))
	  (lkk-convert-server-affair seed)
	  (lkk-convert-recursive-register seed))
	  ))))))
    ;; Shouldn't -- user may go back to (nth 0 lkk-current-record) then hit SPC.
    ;; (setq lkk-/-converting nil)
    ))

(defun lkk-convert-server-affair (seed)
  "Server affair handler for okuri-nasi conversion."
  (cond
   ((and (featurep 'lkk-skkserv)
	 ;; Suppress no further server search: Duplicated server
	 ;; affair does not harm records, but it's a waste of time.
	 (null lkk-server-affair))
    (setq lkk-convert-hit-bottom nil)	; in lkk-skkserv.el
    ;; This adjusts kanji-index
    (lkk-j-search-server seed))))

(defun lkk-convert-recursive-register (seed)
  "Recursive-registering handler for okuri-nasi conversion."
  (cond
   ((or (null (featurep 'lkk-skkserv))
	(eq lkk-server-affair 'server-lost))
    (lkk-recursive-register seed 'force)
    )))  

;;
;; lkk-recursive-register() -- Not yet fully implemented. Don't rely on this.
;;

(defun lkk-recursive-register (seed &optional force)
  "Do recursive registering non-okuri record \
using SEED and user input from minibuffer.
Optional 2nd arg FORCE specifies registering a brand new record.
otherwise new data is just added to current record."
  (cond
   ;; User may come here after he quit from minibuf in okuri-ari
   ;; registering, leaving all okuri status legal.
   ((lkk-okuri-converting-p 'reset-vars-on-failure)
    (lkk-recursive-register-okuri (concat lkk-okuri-root lkk-okuri) force))
   (t
    (let* ((enable-recursive-minibuffers t)
	   (numregex "[0-9]+") numseed numdata
	   (newword (lkk-get-cand-minibuf seed)))
      (cond ((string-match numregex seed)
	     (setq numdata (lkk-numeric-convert-seed-crunch seed)
		   numseed (lkk-replace-in-string numregex "#" seed))))
      (cond
       ((> (length newword) 0)
	(cond
	 ;; 'force is specified by lkk-convert0 on the first SPC hit.
	 ((or force
	      ;; Currently, null lkk-current-record may occur, eg., when user
	      ;; forced recursive registration even after server-hit.
	      ;; Remember that current lkk throws the server record away and
	      ;; never use it for dict update. Tue Mar 28 19:40:20 2000 s.n.
	      (null lkk-current-record))
	  ;; Number crunch, newword is completed by user.
	  (setq lkk-current-record (list
				    (lkk-replace-in-string numregex "#" seed)
				    newword)
		lkk-kanji-index 1)
	  (lkk-dict-update lkk-current-record 'justadd)
	  (lkk-replace-buffer-substring (+ (length lkk-insignia0)
					   lkk-conv-beginning)
					(point)
					;; Just num-replace one word.
					(if numseed
					    (lkk-convert-num-replace
					     newword numdata)
					  newword))
	  (lkk-setup-minor-mode 'lkk-J-mode)
	  (lkk-update-insignia lkk-insignia1) )
	 ;; Surely called from lkk-reconvert().
	 ;; In this case, (lkk-dict-update ... 'justadd) is not called.
	 ;; Just the curent record is faked.
	 (t
	  (setq lkk-current-record (nconc lkk-current-record (list newword))
		;; The server record may added to index, so re-calibrate index
		lkk-kanji-index (1- (length lkk-current-record))
		lkk-current-user-record lkk-current-record)

	  (setq lkk-current-display-record
		(if numseed
		    ;; Case numeric conversion,
		    (mapcar
		     (lambda
		       (cand) (lkk-convert-num-replace cand numdata))
		     lkk-current-record)
		  ;; else non-numeric conversion 
		  lkk-current-record))

	  (lkk-replace-buffer-substring (+ (length lkk-insignia0)
					   lkk-conv-beginning)
					(point)
					(nth
					 lkk-kanji-index
					 lkk-current-display-record))

	  (cond ((null lkk-J-mode)
		 (lkk-setup-minor-mode 'lkk-J-mode)
		 (lkk-update-insignia lkk-insignia1)))
	  )))
       )))))

;;
;; lkk-get-cand-minibuf() -- Read kanji word from minibuffer
;; Mostly for recursive-minibuffer word registration.
;;
;; Sat Mar 25 15:29:39 2000 s.n.
;;

(defun lkk-exit-minibuffer ()
  "Call `lkk-sit-back' then `exit-minibuffer'."
  (interactive) (lkk-sit-back) (exit-minibuffer))

(defun lkk-get-cand-minibuf (&optional seed altprompt)
  "Get kanji word from minibuffer. Prompt SEED is optional.
If SEED is non-nil, kanji conversion for the SEED is automatically set up.
Mostly for recursive-minibuffer word registration."
  (condition-case nil
      (let* ((minibuffer-local-map (cons 'keymap minibuffer-local-map))
	     ;; (lkk-in-read-string t) ; No ! Should make sit-back*s*.
	     (lkk-initial-mode 'lkk-j-mode)
	     (minibuffer-setup-hook-old minibuffer-setup-hook)
	     ;; suggesting okuri alpha is annoying.
	     (seedpromt (if (and lkk-okuri
				 (string-match "[a-z]" (substring seed -1)))
			    (substring seed 0 -1)
			  seed))
	     (minibuffer-setup-hook
	      (lambda ()
		(lkk-mode nil)
		(lkk-set-conv-beginning)
		(cond (seed (insert seedpromt)))
		;; One time use only: User may want to
		;; C-x C-f while recursive minibuffer
		;; is enabled.
		(setq minibuffer-setup-hook
		      minibuffer-setup-hook-old))))
	(define-key minibuffer-local-map "\C-m" 'lkk-exit-minibuffer)
	(read-string (cond 
		      (altprompt altprompt)
		      (seed (concat seed ": "))
		      (t "input kanji: "))))
    (quit nil)))

(defun lkk-recursive-register-okuri (seed &optional force)
  "Do recursive registering okuri record \
using SEED and user input from minibuffer.
Optional 2nd arg FORCE specifies registering a brand new okuri record."
  (let* ((enable-recursive-minibuffers t)
	 ;; User should input kroot (kanji part) only in lkk.
	 (newkroot (lkk-get-cand-minibuf
		    seed
		    (format "Input just kanji part for \"%s\": " seed)))
	 newrecord)
    (cond
     ((> (length newkroot) 0)
      (cond
       ;; 'force is specified by lkk-convert-okuri() on the first conversion.
       ((or force
	    ;; See lkk-recursive-register()
	    (null lkk-current-record-okuri))
	(setq newrecord (list seed (list newkroot) (list lkk-okuri-h newkroot)))

	(lkk-dict-update newrecord 'justadd)
	;; Is this needed here ?
	(setq lkk-current-display-record (lkk-convert-okuri-crunch newrecord)
	      lkk-current-record-okuri newrecord
	      lkk-kanji-index 1)

	(lkk-replace-buffer-substring (+ (length lkk-insignia0)
					 lkk-conv-beginning)
				      (point) (concat newkroot lkk-okuri-h))
	(lkk-setup-minor-mode 'lkk-J-mode)
	(lkk-update-insignia lkk-insignia1) )

       ;; Code below is executed when called from lkk-reconvert().
       ;; In this case, (lkk-dict-update ... 'justadd) is not called.
       ;; Just the curent record is faked.
       (t

	;; modify lkk-current-record-okuri right here.
	(let ((klist (nth 1 lkk-current-record-okuri)))
	  (cond ((listp klist)
		 ;; This ain't my coding style, but
		 ;; lkk-current-record-okuri seems always modified as
		 ;; intended (really ?).
		 (nconc klist (list newkroot))
		 )))

	(setq lkk-current-display-record
	      (nconc lkk-current-display-record
		     (list (concat newkroot lkk-okuri-h)))
	      lkk-kanji-index (1- (length lkk-current-display-record))
	      lkk-current-user-record lkk-current-record-okuri)

	(lkk-replace-buffer-substring (+ (length lkk-insignia0)
					 lkk-conv-beginning)
				      (point)
				      (nth
				       lkk-kanji-index
				       lkk-current-display-record))

	(cond ((null lkk-J-mode)
	       (lkk-setup-minor-mode 'lkk-J-mode)
	       (lkk-update-insignia lkk-insignia1)))

	)))
     )))

;;
;; Numeric conversion subroutines
;; Sat Mar  4 00:48:41 2000 s.n.
;;

;;
;; Numeric converter template for whoever wants to implement.
;; These functions is just required to return mapped string.
;;

(defun lkk-numconv2 (num) num) ; noconv
(defun lkk-numconv3 (num) num) ; noconv

;;
;; lkk-convert-num-replace() -- Numeric conversion renderer.
;;
;; candidate: to be displayed -- (nth lkk-current-record lkk-kanji-index)
;; numdata:   num list pre-generated by lkk-numeric-convert-seed-crunch from
;;             seed (in ASCII)
;;

(defun lkk-convert-num-replace (candidate numdata)
  "Convert CANDIDATE with mapping over NUMDATA list.

Mnemonics can be included in candidates are subset of skk:

#0 : ASCII convert \(actually no conversion\) 
#1 : Zenkaku convert
#2 : Not supported in lkk
#3 : Not supported in lkk

The mnemonics in CANDIDATE are replaced with a number string in NUMDATA
one by one from left to right, with the conversion specified by the
mnemonic.

The number of the mnemonics in CANDIDATE, and the numeric string in
NUMDATA shoud be the same."
  (mapcar
   (lambda (num)
     (setq candidate
	   (lkk-replace-in-string-one "#0" num candidate))
     (setq candidate
	   (lkk-replace-in-string-one "#1" (lkk-han2zen num) candidate))
     (if (fboundp 'lkk-numconv2)
	 (setq candidate
	       (lkk-replace-in-string-one "#2" (lkk-numconv2 num) candidate)))
     (if (fboundp 'lkk-numconv3)
	 (setq candidate
	       (lkk-replace-in-string-one "#3" (lkk-numconv3 num) candidate)))
     ) numdata)
  ;; cheap! Perhaps plain "#" is left in the YOMI string in a dict recored.
  (mapcar
   (lambda (num)
     (setq candidate
	   (lkk-replace-in-string-one "#" num candidate))) ; shortest the last.
   numdata)
  candidate)

;; The same number of '#[012]' is
;; (lkk-numeric-convert-seed-crunch "3a922x388e4d")
;; ("3" "922" "388" "4")

(defun lkk-numeric-convert-seed-crunch (seed)
  "Crunch numeric yomi SEED and returns a list of the numeric strings.
The numerals in the list are used afterwards for displaying candidates."
  (let (posdata (start 0))
    (while (string-match "[0-9]+" seed start)
      (setq posdata
	    ;;(cons (substring seed (match-beginning 0) (match-end 0)) posdata)
	    (cons (match-data) posdata)
	    start (match-end 0)))
     (mapcar (lambda (c)
	       (apply 'substring seed c)
	       ) (nreverse posdata))))	; the mapcar returns nil if fail

;;
;; lkk-convert-region() -- conversion-and-display engine.
;;
;; There is a STRUCTURAL INCONSISTENCY and shoud be remedied later:
;;
;;   1) Okuri-ari hack is not done by anymore in lkk-convert-region()
;;   2) There are two work records lkk-current-display-record and
;;      lkk-current-record directly.
;; 
;; These may have to be fixed before go forward further serious
;; confusion would be raised.
;;
;; Wed Mar  1 00:35:21 2000 s.n.
;;

(defvar lkk-inhibit-elettronico-num-completion nil
  "Non-nil inhibits number coversion entry's expantion on j-elettronico.")

;;
;; The advertised skk's `program henkan' parser
;; Skelton was from skk5.30
;;

;; I feel annoyed by the sexp-parse, so default is t.
(defvar lkk-inhibit-sexp-parse t
  "PROGRAMU HENKAN is deprecated in lkk default, but if you want it \
setq this to nil.")

;; (lkk-bogo '(string-match "^ *(.*)$" word) 1)
;; (0 . 3406) ; 486dx 20mhz
;; The code below (two substring()s) =>
;; (0 . 424)

(defsubst lkk-parse-cand (word)
  "The `program conversion' entity."
  (cond
   (lkk-inhibit-sexp-parse word)
   (;; (string-match "^ *(.*)$" word)
    (and (string= (substring word 0 1) "\(")
	 (string= (substring word -1) "\)"))
    (condition-case nil
	;; format for failsafe
	(format "%s" (eval (car (read-from-string word))))
      (error word)))
   (t word)))

(defun lkk-convert-region (rb re &optional arg looseflag seedonly okuri)
  "Convert region RB RE.  Return t or nil corresponding to conversion result.
Region RB RE have to contain pure yomi of kanji.
The rest of args are all optional:
ARG specifies insert times.
LOOSEFLAG specifies loose dict search.
SEEDONLY specifies inserting yomi only (not kanji) in the `loose' search.
OKURI specifies to search `okuri-ari' area of dictionary."
  ;; arg should be "p"
  (interactive "*r\np")
  (let ((seed (buffer-substring-no-properties rb re))
	(numregex "[0-9]+") numseed numdata record)

    (cond ((string-match numregex seed)
	   (setq numdata (lkk-numeric-convert-seed-crunch seed)
		 numseed (lkk-replace-in-string numregex "#" seed))))

    (setq record (lkk-dict-search 
		  (if numseed numseed seed) looseflag)) ; scan dictionary

    (cond
     (record			; record found.
      (or lkk-conv-beginning
	  (save-excursion (goto-char rb) (lkk-set-conv-beginning nil)))

      (setq lkk-current-record record
	    lkk-kanji-index (if seedonly 0 1))

      (setq lkk-current-display-record
	    (if numseed
		;; Case numeric conversion,
		(mapcar
		 (lambda
		   (cand) (lkk-convert-num-replace cand numdata))
		 record)
	      ;; else non-numeric conversion 
	      record))

      (cond
       ;; Don't display if seedonly (specified by elettronico) and
       ;; display-record is incompletely converted (num mnemonic # is
       ;; includeded).
       ((and seedonly lkk-inhibit-elettronico-num-completion
	     (string-match "#" (nth lkk-kanji-index
				    lkk-current-display-record)))
	())
       (t
	;; Alternative is replacing lkk-conv-beginning to (point)
	;; with (concat insignia0 KANJI), but +() seems like more faster(?)
	;; than concat(). The above lkk-conv-beginning() ensuring code
	;; should not be removed, of course.
	(lkk-replace-buffer-substring (+ (length lkk-insignia0)
					 lkk-conv-beginning)
				      (point)
				      (lkk-parse-cand
				       (nth lkk-kanji-index
					    lkk-current-display-record))
				      arg)
	(lkk-setup-minor-mode 'lkk-J-mode)
	(lkk-update-insignia lkk-insignia1)
	t)))
     (t ; The caller must do the job on conversion failure.
      nil
	))))

;;
;; Further conversion commands.
;;

(defun lkk-reconvert (&optional arg)
  "Henkan-again command. Normally bound to SPC in lkk-J-mode-map."
  (interactive "*P")
  (let ((record (or lkk-current-display-record
		    lkk-current-record)))
    ;; Debugger
    (if lkk-debug-hook (run-hooks 'lkk-debug-hook))

    (cond
     ;; Exceptional hack for elettronico.
     ((or (and (null lkk-kanji-index) (null record))
	  (and lkk-full-auto (integerp lkk-kanji-index) (= lkk-kanji-index 0)))
      ;; We must convert as if it is new yomi (user may edited).
      (lkk-convert arg))

     ;; Ͽ => (lkk-reconvert 3) after "" => Ͽ
     (;; (and arg lkk-kanji-index) ; we have non-nil lkk-kanji-index here
      arg			   ; Just testing arg is good enough.
      (lkk-sit-back)
      (lkk-convert arg)
      )

     ;; Normal reconvert.
     ((and lkk-conv-beginning record lkk-kanji-index)
      (let ((kanji (nth (1+ lkk-kanji-index) record)))
	(cond 
	 (kanji
	  (setq lkk-kanji-index (1+ lkk-kanji-index))
	  (lkk-replace-buffer-substring
	   (+ (length lkk-insignia0) lkk-conv-beginning)
	   (point)
	   (lkk-parse-cand
	    (nth lkk-kanji-index record)))
	  ;; Lousy, just display.
	  (lkk-reconvert-show-cand))
	 (t 
	  (cond (;; lkk-okuri-converting-p() is not suitable here,
		 ;; (lkk-okuri-converting-p)
		 ;; because `lkk-insignia2' is already erased.
		 lkk-okuri
		 
		 (lkk-reconvert-okuri-server-affair record)
		 (lkk-reconvert-okuri-recursive-register record))

		;; server-affair for okuri-nasi reconvert, and registering.
		(t (lkk-reconvert-server-affair record)
		   (lkk-reconvert-recursive-register record)))
	  )))))))

(defun lkk-reconvert-server-affair (record)
  "Server affair handler for okuri-nasi re-conversion.
This function does not know what arg RECORD is all about."
  (cond
   ((and (featurep 'lkk-skkserv)
	 (= lkk-kanji-index (1- (length record)))
	 ;; Suppress no further server search:
	 (null lkk-server-affair))
    (setq lkk-convert-hit-bottom t)
    ;; This adjusts kanji-index
    (lkk-j-search-server (car record)))))

(defun lkk-reconvert-recursive-register (record)
  "Recursive-registering handler for okuri-nasi re-conversion.
This function does not know what arg RECORD is all about."
  (cond
   ((and (or (null (featurep 'lkk-skkserv))
	     ;; Anything non-nil denotes already called server:
	     ;; We should do registering even when 'server-hit.
	     lkk-server-affair)
	 (= lkk-kanji-index (1- (length record))))
    (lkk-recursive-register (car record) nil) )) )

;;
;; Now lkk-j-search-server knows how to deal with okuri-ari reconvert
;;

(defun lkk-reconvert-okuri-server-affair (record)
  "Server affair handler for okuri-ari re-conversion.
This function uses RECORD only for index check."
  (cond
   ((and (featurep 'lkk-skkserv)
	 (= lkk-kanji-index (1- (length record)))
	 ;; Suppress no further server search:
	 (null lkk-server-affair))
    (setq lkk-convert-hit-bottom t)
    ;; In this case seed should not be of car() of disp record.
    (lkk-j-search-server (car lkk-current-record-okuri)))))

(defun lkk-reconvert-okuri-recursive-register (record)
  "Recursive-registering handler for okuri-nasi re-conversion.
This function does not know what arg RECORD is all about."
  (cond
   ((and (or (null (featurep 'lkk-skkserv))
	 ;; Anything non-nil denotes already called server:
	 ;; We should do registering even when 'server-hit.
	 lkk-server-affair)
     (= lkk-kanji-index (1- (length record))))
;;;    (ding)
;;;    (message
;;;     "Sorry, recursive regsitering for okuri-ari is not yet implemented.")

    (lkk-recursive-register-okuri (car lkk-current-record-okuri) nil)
    )))

(defun lkk-reconvert-show-cand ()
  "Show candidate list on echo area if lkk-kanji-index > four."
  (if (> lkk-kanji-index 4)
      (message "%d: %s"
	       lkk-kanji-index
	       (nthcdr lkk-kanji-index 
		       (or lkk-current-display-record
			   lkk-current-record)))))

(defun lkk-unwind-up ()
;;, and clear lkk-J-mode.
  "While in lkk-J-mode, unwind up to YOMI.
Normally bound to C-g in lkk-J-mode-map."
  (interactive)
  (let ((record (or lkk-current-display-record
		    lkk-current-record)))
    (cond
     ((and (integerp lkk-kanji-index)
	   (> lkk-kanji-index 0))
      (setq lkk-kanji-index 0)
      (lkk-replace-buffer-substring
       (+ (length lkk-insignia0) lkk-conv-beginning)
       (point) (nth lkk-kanji-index record))

      (lkk-J-mode-exit)
      (setq lkk-server-affair nil
	    lkk-convert-hit-bottom nil
	    )
      (lkk-update-insignia lkk-insignia0)
      ))))

(defun lkk-prev-cand (&optional arg)
  "Move backward in lkk-current-record.  Decrements lkk-kanji-index."
  (interactive "*P")

  (let ((record (or lkk-current-display-record
		    lkk-current-record)))

    ;; Debugger
    (if lkk-debug-hook (run-hooks 'lkk-debug-hook))

    (cond 
     ;; We must take this keyclick as if it is a part of new seed.
     ((or (and (null lkk-current-record) (null lkk-kanji-index))
	  (and lkk-full-auto (integerp lkk-kanji-index) (= lkk-kanji-index 0)))
      (lkk-j-input (prefix-numeric-value arg)))

     ;; Normal reconvert backward.
     ((and lkk-conv-beginning record
	   (> lkk-kanji-index 0))

      (setq lkk-kanji-index (1- lkk-kanji-index))
      (lkk-replace-buffer-substring
       (+ (length lkk-insignia0) lkk-conv-beginning) (point)
       (lkk-parse-cand
	(nth lkk-kanji-index record)))
      ;; Lousy, just display.
      (lkk-reconvert-show-cand)

      ;; (Hit top) == (yomi is displayed). Turn off lkk-J-mode.
      (cond
       ((= lkk-kanji-index 0)
	(lkk-J-mode-exit)
	(setq lkk-server-affair nil
	      lkk-convert-hit-bottom nil
	      )
	(lkk-update-insignia lkk-insignia0)
	)))

     (t (message "Unknown error in lkk-prev-cand")))

    ;;(lkk-watch 1);; debugger
    ))

(defun lkk-completion-convert (arg)
  "Normally bound to TAB in lkk-j-mode.
Next completion facility is not yet implemented."
  ;; ARG not used yet. Not `*p'. Unusual, because
  ;; I want to use double-M-SPC for lkk toggling even for
  ;; (buffer-read-only => t) buffers.
  (interactive "p")
  (cond
   ((and (null lkk-inhibit-multiple-keyclick)
	 (lkk-multiple-click-crank))
    ())
   (t (barf-if-buffer-read-only)	; read-only checked here
      (cond 
       (lkk-conv-beginning		; else call old map function
	;; (setq lkk-/-converting nil)
	(setq lkk-/-mode nil)
	(lkk-setup-minor-mode 'lkk-J-mode)

	(if (lkk-convert-region
	     (+ (length lkk-insignia0) lkk-conv-beginning) (point) 1 'loose)
	    ;; (lkk-update-insignia lkk-insignia1) ; done in lkk-convert-region()
	    () 
	  ;;
	  ;; Here I should write codes for the case conversion didn't hit.
	  ;;
	  ))
       ;; Or emulate an old emacs command (TAB, M-TAB etc).
       (t (lkk-maneuver-old-keybind))))))

(defun lkk-start-/-convert (arg)
  "Normally bound to / in lkk-j-mode-map."
  (interactive "*p")

  ;; This always sit-back if J-mode.
  (if lkk-J-mode (lkk-sit-back))

  ;; Unless converting, put insignia0
  (or lkk-conv-beginning
      (lkk-set-conv-beginning))

  ;; Input literal (ASCII) via lkk-j-input
  (cond
;;;   ((and lkk-/-converting
;;;	 (null lkk-J-mode))
;;;    (lkk-j-input arg))
   ;; lkk-j-input() refers, and reset by lkk-j-*convert*
   (t
    ;;(setq lkk-/-converting t)
    (setq lkk-/-mode t)
    (force-mode-line-update)
    )))

(defun lkk-backward-delete-char (arg)
  "`delete-backward-char' sentinel in `lkk-j-mode'.
If user tries to delete `lkk-insignia0' or `lkk-insignia1',
delete it backward and then call `lkk-sit-back'.
If user tries to delete `lkk-insignia2',
delete it and then call `lkk-reset-okuri'."
  (interactive "*p")
  (cond
   ;; (Not required with this strict check, but) Okuri status checked first.
   ((and lkk-okuri-beginning
	 (= (point) (+ lkk-okuri-beginning (length lkk-insignia2)))
	 (eq (preceding-char) (string-to-char lkk-insignia2)))
    (backward-delete-char arg)
    (lkk-reset-okuri) (force-mode-line-update))
   ;; then insignia0/1 check.
   ((and lkk-conv-beginning		; how about lkk-J-mode ?
	 (or
	  (and (= (point) (+ lkk-conv-beginning (length lkk-insignia0)))
	       (eq (preceding-char) (string-to-char lkk-insignia0)))
	  (and (= (point) (+ lkk-conv-beginning (length lkk-insignia1)))
	       (eq (preceding-char) (string-to-char lkk-insignia1)))))
    (backward-delete-char arg)
    (lkk-sit-back) (force-mode-line-update))
   (t (lkk-maneuver-old-keybind))))

;; Purge dictionary data under testing.  Not well tested.

;;
;; lkk-purge-it() -- purge current data
;;
;; Fri Mar 24 03:02:53 2000 s.n.
;;

;; This should inevitably use lkk-kanji-index
(defsubst lkk-purge-data (index record)
  "Equiv. \(delq \(nth INDEX RECORD\) RECORD\)."
    (delq (nth index record) record))

(defun lkk-purge-data-okuri (okuri kanji record)
  "Purge data corresponding OKURI and KANJI in okuri-nasi record RECORD."
  (let* ((odata (assoc okuri record)) newodata)
    (cond
     (odata
      ;; Assume the same kanji is not on car() position (allright ?)
      (setq newodata (delete kanji (copy-list odata)))
      (cond
       ;; (length newodata)==1 Shall I purge kanji from klist ? -- NO (in lkk)
       ((null (cdr newodata)) (delq odata record))
       (t (mapcar
	   (lambda (data)
	     (if (eq odata data) newodata data)
	     ) record))))
     ;; Or return as it is.
     (t record))))

(defun lkk-purge-it-nop ()
  "This can be fset to `lkk-purge-it' if a fatal bug found in lkk-purge-it."
  (interactive)
  (message "Not yet \(or forever ?\) implemented.")
  (sit-for 2) (message ""))

(defun lkk-purge-it ()
  "Purge current data. Normally bound to \"X\" in lkk-J-mode-map."
  (interactive)
  (cond
   ;; Never occurs in lkk-J-mode. Index is always > 1 here in this fun.
   ;; ((< lkk-kanji-index 1) ())

   ;; For okuri-ari conversion, we use faked new record to update dictionary
   ;; because lkk-dict-new-record-okuri() generates new okuri data
   ;; if current indexed display record is lacked in lkk-current-record-okuri.
   (;;(and lkk-okuri-h lkk-current-display-record (null lkk-current-record))
    (and lkk-okuri-h lkk-current-display-record lkk-current-record-okuri)
    (let* ((word (nth lkk-kanji-index lkk-current-display-record))
	   (otop (string-match lkk-okuri-h word))
	   (kanji (substring word 0 otop))
	   (okuri (substring word otop))
	   (fakerecord))
      (cond
       ((y-or-n-p
	 (format "Purge \"%s\"+\"%s\" ? " kanji okuri))
	;; Fake whole new record. lkk-current-record-okuri should be
	;; set to the new fake record
	(setq
	 fakerecord (lkk-purge-data-okuri okuri kanji lkk-current-record-okuri)
	 lkk-current-display-record (lkk-convert-okuri-crunch fakerecord)
	 lkk-current-record-okuri fakerecord)

	(lkk-dict-update fakerecord)

	(message "\"%s\"+\"%s\" purged" kanji okuri)
	(lkk-prev-cand)
	))))

   ;; In num-crunch, and plain okuri-nasi conversion,
   ;; lkk-current-record is used to update dictionary.
   (t
    (let ((target (nth lkk-kanji-index lkk-current-record))
	  (fakerecord))
      (cond
       ((y-or-n-p
	 (format "Purge \"%s\" ? " target))
	(setq fakerecord (lkk-purge-data lkk-kanji-index lkk-current-record))

	;; I don't remember why I should fake lkk-kanji-index here: (Maybe)
	;; we have to force lkk-dict-update() work, and disable horizontal
	;; sort, with binding lkk-kanji-index to 1 (?)
	(let ((lkk-kanji-index 1))
	  (lkk-dict-update fakerecord))

	(setq lkk-current-record fakerecord)
	(cond ((string-match "#" (car fakerecord)) ; am doing num-crunch
	       (setq lkk-current-display-record
		     (lkk-purge-data lkk-kanji-index
				     lkk-current-display-record))))
	(message "%s purged" target)
	(lkk-prev-cand)
	)))
    )))

;;
;; Starting input Yomi, and the tangled OKURI-ARI hack
;; Thu Mar  2 11:36:42 2000 s.n.
;;

;;
;; lkk-start-seeding -- Sets kanji or okurigana beginning
;;
;; Started: Tue Feb 29 23:49:59 2000
;;
;; .S
;; .su
;; .  -> auto-conv only ->
;; 
;;

(defun lkk-start-seeding (arg)
  "Normally bound to uppercase char in lkk-j-mode."
  (interactive "*p")
  ;; While partial kana input or not converting yet,
  ;; uppercase is just converted into lower, and do no point fiddle.
  (cond
   ;; == /-CONVERTING, 
   ;; AND  is not yet displayed, accept key without tolower fake.
;;;   ((and lkk-/-converting (null lkk-J-mode))
;;;    (lkk-j-input arg))

   ;; == OKURI-ARI conversion.
   ((and lkk-conv-beginning		; started yomi seq already -- okurigana
	 ;;(null lkk-j-beginning)
	 (null lkk-J-mode))

    ;; Stale romaji left should not be left
    (cond
     ((null lkk-j-beginning)
      (lkk-start-seeding-okuri arg))
     (t					; lkk-j-beginning == non-nil
      (let ((pc (preceding-char))
	    (terminator-p (lkk-r2k-terminator-p
			   ;; last-command-char is not lowercase here.
			   (char-to-string
			     (lkk-tolower last-command-char))))
	    (clobberit (lambda (k h)
			 ;; Is this hacking one char good enough ?
			 ;; (setq lkk-j-beginning nil)
			 (lkk-clear-marker 'lkk-j-beginning)
			 (let (( buffer-undo-list t))
			   (delete-backward-char 1))
			 (insert (if lkk-q-mode k h))
			 (lkk-start-seeding-okuri arg))))
	(cond
	 ((and (null terminator-p)
	       (or (eq (preceding-char) ?n) ; stale `n'
		   (eq (preceding-char) ?m)))
	  (funcall clobberit "" "" ))
	 ((eq (preceding-char)	; stale double consonant
	      (lkk-tolower last-command-char))
	  (funcall clobberit "" ""))
	 ;; Eg. case of stale "A" after "K" in sequence "KA".
	 ;; skk seems to allow it and convert to lower.
	 ;; Further, if this is not trapped, stale lkk-okuri-beginning
	 ;; will be set and that's not what I want.
	 (t (lkk-tolower-input arg))))))) ; No okuri, nor j-beginning reset.

   ;; == START YOMI
   (t 
    (if lkk-J-mode			; implicit kakutei.
	(lkk-sit-back))
    (or lkk-j-beginning
	(progn (lkk-set-conv-beginning)
	       (setq lkk-j-beginning (point-marker))))
    (lkk-tolower-input arg))))

;; Subroutine called by lkk-start-seeding()
(defun lkk-start-seeding-okuri (arg)
  ;; If old one is already there, clear it
  (lkk-okuri-converting-p 'init-all-on-success)

  ;; okurigana hack not yet completed.
  (setq lkk-okuri-beginning (point-marker))
  (insert lkk-insignia2)
  ;; Is it correct to keep in a global -- might be okay,
  ;; non-auto-okuri won't be implemented...
  (setq lkk-okuri (char-to-string (lkk-tolower last-command-char)))
  (lkk-tolower-input arg))

;; Subroutine called by lkk-start-seeding()
(defun lkk-tolower-input (arg)
  (let ((last-command-char (lkk-tolower last-command-char)))
    (lkk-j-input arg)))

;;
;; lkk-convert-okuri() -- Most touchy part
;;
;; Called via lkk-j-input() only -- means auto-okuri only.
;;
;; Started: Wed Mar  1 00:11:34 2000
;;
;; Emulate lkk-current-display-record and lkk-kanji-index so that
;; lkk-reconvert(), lkk-prev-cand() etc. works.
;;
;; Caveat: On dictionary update, lkk-current-display-record can not be used as
;; in okuri-nasi conversion. The entry should be reproduced using
;; lkk-current-record-okuri variable which records real okuri-ari entry.
;;
;; Related Variables:
;; lkk-okuri           : consonant
;; lkk-okuri-h         : hiragana string maybe a char or more
;; lkk-okuri-root      ; root kanji part
;; lkk-okuri-beginning : marker
;; lkk-conv-beginning  : Yomi top
;;
;; If first search-conversion is complete anyway, delete lkk-insignia2 on
;; marker lkk-okuri-beginning.
;;

(defun lkk-convert-okuri (&optional arg)
  "Do okuri-ari conversion, and build lkk-current-display-record and lkk-kanji-index.

Related Variables:
lkk-okuri           : consonant
lkk-okuri-h         : hiragana string maybe a char or more
lkk-okuri-root      : root kanji part
lkk-okuri-beginning : marker
lkk-conv-beginning  : Yomi top"
  (let* ((inhibit-quit t)
	 (dtop 'lkk-okuri-ari-dict-top)
	 (emulated)
	 (seed
	  (concat
;;;	   (buffer-substring-no-properties
;;;	    (+ (length lkk-insignia0) lkk-conv-beginning)
;;;	    lkk-okuri-beginning)
	   lkk-okuri-root		; Build by lkk-convert-okuri-crank()
	   lkk-okuri))			; Maybe extended by lkk-j-input()
	 (record (lkk-dict-search seed nil)))
    (cond
     ((and record (setq emulated (lkk-convert-okuri-crunch record)))

      ;; Branch copy of record. The record should be copied.
      ;; Otherwize subsequent lkk-dict-new-record-okuri() will generate
      ;; broken record when subalist(s) has only one kanji like this:
      ;; ("t" ("") ("" "") ("ä" ""))
      ;; Th destroyed record:
      ;; ("t" (nil "") ("" nil "") ("ä" ""))
      (setq lkk-current-record-okuri (copy-list record))

      (setq lkk-current-display-record emulated
	    lkk-kanji-index 1)
      (lkk-replace-buffer-substring
       (+ (marker-position lkk-conv-beginning) (length lkk-insignia0)) (point)
       (nth lkk-kanji-index lkk-current-display-record)
       1)
      (lkk-setup-minor-mode 'lkk-J-mode)
      (lkk-update-insignia lkk-insignia1))
     (t 
      ;; Server affair,
      (lkk-convert-okuri-server-affair seed record)
      ;; then try the recursive registering.
      (lkk-convert-okuri-recursive-register seed)
      ))))

(defun lkk-convert-okuri-server-affair (seed record)
  (let (emulated)
    (cond
     ((featurep 'lkk-skkserv)
      (setq lkk-convert-hit-bottom nil) ; in lkk-skkserv.el
      ;; This adjusts kanji-index
      (lkk-j-search-server seed 'nodisplay)
      (cond
       ((and (eq lkk-server-affair 'server-hit)
	     (setq record lkk-current-server-record
		   emulated (lkk-convert-okuri-crunch record)))
	;; Branch copy of record. The record should be copied.
	;; See comment in lkk-convert-okuri() above.
	(setq lkk-current-record-okuri (copy-list record)
	      lkk-current-display-record emulated
	      lkk-kanji-index 1)
	(lkk-replace-buffer-substring
	 (+ (marker-position lkk-conv-beginning) (length lkk-insignia0))
	 (point)
	 (nth lkk-kanji-index lkk-current-display-record)
	 1)
	(lkk-setup-minor-mode 'lkk-J-mode)
	(lkk-update-insignia lkk-insignia1)))
      ))))

(defun lkk-convert-okuri-recursive-register (seed)
  (cond ((or (null (featurep 'lkk-skkserv))
	     (eq lkk-server-affair 'server-lost))
	 (lkk-recursive-register-okuri seed 'force)
	 )))

;;
;; lkk-convert-okuri-crunch() -- Get kanji + okurigana string
;;
;; (assoc "Ǥr" lkk-big-dict-okuri)
;; ("Ǥr" ("") ("" "") ("" ""))
;;
;; Related Variables:
;; lkk-okuri           : consonant
;; lkk-okuri-h         : hiragana string maybe a char or more **
;; lkk-okuri-root      ; root kanji part
;; lkk-okuri-beginning : marker
;; lkk-conv-beginning  : Yomi top
;;

;;
;; Wed Mar 22 02:52:05 2000 s.n.
;; Now lkk-convert-okuri-crunch() can generate new display record:
;;
;; x => ("s" (""))
;; (let ((lkk-okuri-root "") (lkk-okuri-h "")) (lkk-convert-okuri-crunch x))
;;  => ("Ȥ" "")
;; (let ((lkk-okuri-root "") (lkk-okuri-h "")) (lkk-convert-okuri-crunch x))
;;  => ("Ȥ" "")
;; y => ("s" ("") ("" "") ("" "") ("" "") ("" ""))
;; (let ((lkk-okuri-root "") (lkk-okuri-h "")) (lkk-convert-okuri-crunch y))
;;  => ("᤽" "")
;;

;; (setq zz '("r" ("" "")))
;; (let ((lkk-okuri-root "") (lkk-okuri-h "")) (lkk-convert-okuri-crunch zz))
;;  => ("Ȥ" "" "")

(defun lkk-convert-okuri-crunch (record)
  "Get a candidate list from okuri-ari RECORD.
Returns a list which can be bound to `lkk-current-display-record'.

Eg., out of a RECORD
  \(\"r\" \(\"\" \"\" \"£\"\) \(\"\" \"£\"\)
a candidate list 
  \(\"\" \"\" \"٤\" \"£\"\)
will be returned.
`lkk-okuri-root' and `lkk-okuri-h' referred for the crunch-gen."
  (let* ((hit (assoc lkk-okuri-h record))
	 (krootmissed
	  (and hit
	       (let (lis)
		 (mapcar (lambda (kroot)
			   (if (null (member kroot hit))
			       (setq lis (cons kroot lis)) ))
			 (nth 1 record)) (nreverse lis)))) )
    (cond
     (hit
      (cons (concat lkk-okuri-root lkk-okuri-h)
	    (mapcar (lambda (k) (concat k
					;;(car hit)
					lkk-okuri-h))
		    ;; (cdr hit)
		    (append (cdr hit) krootmissed)
		    )))
     ;; Generate a new record for display having new okurigana.
     (t
      (cons (concat lkk-okuri-root lkk-okuri-h)
	    (mapcar (lambda (k) (concat k lkk-okuri-h))
		    ;; ("" "" "" "" "" "" "") part
		    (nth 1 record) )))
     )))

;;
;; lkk-dict-new-record-okuri() -- Reproduce new updated record for sit-back.
;;
;; Should not destructvely modify lkk-current-display-record or it's
;; sublist here.
;;

;;
;; Wed Mar 22 02:52:05 2000 s.n.
;; Now lkk-dict-new-record-okuri() can add newly generated okuri entry:
;;
;; y => ("s" ("") ("" "") ("" "") ("" "") ("" ""))
;; (let* ((lkk-kanji-index 1)
;;        (lkk-okuri-root "") (lkk-okuri-h "")
;;        (lkk-current-display-record (lkk-convert-okuri-crunch y)))
;;   (lkk-dict-new-record-okuri))
;; => ("s" ("") ("" "") ("" "") ("" "") ("" "")
;;      ("" ""))
;; x => ("s" (""))
;; (let* ((lkk-kanji-index 1)
;;       (lkk-okuri-root "") (lkk-okuri-h "")
;;       (lkk-current-display-record (lkk-convert-okuri-crunch x)))
;;   (lkk-dict-new-record-okuri))
;; => ("s" ("") ("" ""))
;; (let* ((lkk-kanji-index 1)
;;       (lkk-okuri-root "") (lkk-okuri-h "")
;;       (lkk-current-display-record (lkk-convert-okuri-crunch x)))
;;   (lkk-dict-new-record-okuri))
;; => ("s" ("") ("" ""))
;;

(defun lkk-dict-new-record-okuri ()
  "Return restored okuri-ari record out of lkk-current-display-record.
Called from `lkk-sit-back'."
  (cond
   ((lkk-dict-check-record lkk-current-display-record 'kanjionly)
    (let* ((inhibit-quit t)

	   ;; "礯", the user selection:
	   (hitword (nth lkk-kanji-index lkk-current-display-record))

	   ;; Sub-data in the hit dict record hereunder:
	   ;; ("" "" "" "" "" "" "") part:
	   (klist (car (cdr lkk-current-record-okuri)))

	   ;; nthcdr 2 part of the dict record, THIS MAY BE NIL:
	   (subalist (nthcdr 2 lkk-current-record-okuri))

	   ;; ("" "" "" "" ""),
	   ;; THIS MUST BE NIL IF CRUNCH-GEN DONE:
	   (hitrecord (assoc lkk-okuri-h subalist)) 
	   ;; and its cdr: ("" "" "" ""),
	   ;; THIS MUST BE NIL IF CRUNCH-GEN DONE:
	   (hitrecordcdr (cdr hitrecord))

	   ;; Gonna be ""
	   (hitkroot)
	   )

      ;; get hitkroot out of klist, and sort it.
      (catch 'hit
	(mapcar (lambda (k)
		  (cond ((string-match k hitword)
			 (setq hitkroot k)
			 (throw 'hit hitkroot)
			 ))
		  ) klist))

;;; Some of the copy-list below can be eliminated, but I keep them for
;;; insurance (I don't remember which can be deleted).  Horizontal
;;; list is short, and there won't be so much pain.

      ;; Here is updated new list. Is this sort SKK compatible ?
      (setq klist (cons hitkroot
			    (delete hitkroot (copy-list klist)) ))
      ;; This may not be nil even when crunch-gen done.
      (setq subalist
	    (lkk-dict-alist-delete lkk-okuri-h (copy-list subalist)))

      (cond
       (hitrecord
	;; Complete tail part. Pass copy-list(), or barfs.
	;; Pass copy for delete(), or barfs.
	(setq hitrecordcdr (delete hitkroot (copy-list hitrecordcdr)))
	(setq hitrecord (cons lkk-okuri-h (cons hitkroot hitrecordcdr)))


	(setq subalist (cons hitrecord subalist))
	;; Just return result: (cons klist subalist) will be cdr of valid record.
	(cons (car lkk-current-record-okuri)
	      (cons klist subalist))
	) ; hitrecord

       ;; Or hitrecord == nil. Crunch-gen must have been done
       ;; by lkk-convert-okuri-crunch().
       (t
	;; Is it okay ?
	(cons (car lkk-current-record-okuri)
	      (cons klist
		    (cons (list lkk-okuri-h hitkroot) subalist)))) )
      ))
   (t
    ;; It's allright with just a nil-return, maybe index points yomi.
    ;;      (error "lkk BUG: wrong lkk-kanji-index value %d for %S"
    ;;	       lkk-kanji-index lkk-current-display-record)
    nil
    )))

;;; 
;;; Simple input automaton, and gizmoz
;;;

;; gizmoz first

(defvar lkk-full-auto-hchar-threshold 1
  "Threshold of hiragana character number to start `full-auto' completeion.
Defaults 1 -- meaning complesion is done when you hit 2nd hiragana.")

;; Maybe in big alist method shorter delay is comfortable.
(defvar lkk-elettronico-delay 400
  "Delay in milliseconds before starting lkk-full-auto completion. 
Perhaps default is in 300 to 1000 milliseconds range.")

(defvar lkk-elettronico-delay2 650
  "Delay in milliseconds before lkk-full-auto completion goes into 2nd stage.
Perhaps default is in 500 to 3000 milliseconds range.")

(defvar lkk-elettronico-delay3 300
  "Delay in milliseconds before lkk-full-auto completion goes into 3rd stage.")

(defvar lkk-elettronico-delay4 400
  "Delay in milliseconds before lkk-full-auto completion goes into 4th stage.")

;; Must move to lkk-setup-j-mode-map() when full-auto mode is good enough.
(cond (lkk-j-mode-map
       (define-key lkk-j-mode-map [C-tab] 'lkk-elettronico)))
(cond (lkk-l-mode-map
       (define-key lkk-l-mode-map [C-tab] 'lkk-elettronico)))

(defvar lkk-inhibit-elettronico-in-l-mode nil
  "Inhibit lkk-full-auto-mode in lkk-q-mode if non-nil.")

(defvar lkk-elettronico-/-conv-salt 1
  "Salt added to lkk-full-auto-hchar-threshold while lkk-/-mode.")

;;
;; lkk-modify-dabbrev-syntax() -- dabbrev support in emacs-20
;;
;; mule-2.3(19.28)'s dabbrev.el is assumed to loaded in emacs-20.xx:
;;
;; Original dabbrev.el in (current) emacs-20.[2-6] is brain-damaged at
;; least for Japanese language.  Uses are strongly recommended to
;; install mule-2.3(19.28)'s dabbrev.el (19.34's is also damaged) to
;; use `lkk-elettronico' facility as intended.
;;
;; Wed Mar 29 14:13:56 2000
;;
;; The modify-syntax-entry() technique based on teachings from:
;;
;; Path: miles.smoke.or.jp!nifty!SGU01663
;; Newsgroups: nifty.funix.linux
;; Subject: Re: j-dabbrev-expand
;; Date: 28 Mar 2000 13:57:00 +0900
;; Message-ID: <11318@nifty.funix.mes19> ; Valid only in authours' system
;; References: <11316@nifty.funix.mes19> ; Valid only in authours' system
;;
;; lkk-insignia0 ""
;; lkk-insignia1 ""
;; lkk-insignia2 "."
;;

(defun lkk-modify-dabbrev-syntax ()
  "Fix emacs-20's broken \(?\) syntax mainly for `lkk-elettronico-j-hack'."
  (cond
   ( ;; (and
     (eq (char-syntax ?) ?w)		; Means this is emacs-20.x
	 ;; (lkk-full-auto)
     ;;)

    ;; (message "THE ADVICE called") ; debugger
    (modify-syntax-entry (string-to-char lkk-insignia0) " ")
    (modify-syntax-entry (string-to-char lkk-insignia1) " ")
    (modify-syntax-entry (string-to-char lkk-insignia2) " ") ; okay? or punct?
    (modify-syntax-entry ? ".")
    (modify-syntax-entry ? "."))))

(defadvice dabbrev-expand (before lkk-ad activate)
  (lkk-modify-dabbrev-syntax))

(defun lkk-elettronico-j-hack ()
  "Do sort-of-overkill full-auto yomi completion.
Called at the end of lkk-j-input."
  (let ((stage 0))
    (cond
     ((sit-for 0 lkk-elettronico-delay) ; -- *DO* redisplay here
      (cond
       ;; == STAGE 0 ELETTRONICO: In KANA INPUT, run-hooks level 0 hook.
       ((and (null lkk-conv-beginning)	; j-beginning==0 is already checked.
	     (null lkk-J-mode))		; Might be unnecessary ?
	(while (and (sit-for 0 lkk-elettronico-delay)
		    (lkk-run-hooks 'lkk-elettronico-j-hack-hook0)) t))

       ;; == STAGE 1 ELETTRONICO: in KANJI CONVERTING, Completes YOMI only.
       ((and
	 lkk-conv-beginning		; having conv beginning,
	 (null lkk-j-beginning)		; and kana sequence is complete.
	 (null (or lkk-okuri		; No j-hack, in okuri (by auto-okuri).
		   lkk-okuri-beginning lkk-okuri-root lkk-okuri-h))
	 (> (- (point) lkk-conv-beginning)
	    ;; Should count insignia0 + threshold, but very annoying.
	    (let ((threshold
		   (if ;;lkk-/-converting
		       lkk-/-mode
		       (+ lkk-full-auto-hchar-threshold
			  lkk-elettronico-/-conv-salt)
		     lkk-full-auto-hchar-threshold)))	    
	      (cond ((> lkk-j-bytes 1)	; perhaps mule-2.
		     (cond
		      ( ;;;lkk-/-converting	; size differs due to this...
		       lkk-/-mode
		       (+ (length lkk-insignia0) threshold))
		      (t (* (1+ threshold) (length lkk-insignia0)))))
		    (t (1+ threshold)))) ; (length lkk-insignia0) == 1
	    ))
	(or lkk-J-mode (lkk-setup-minor-mode 'lkk-J-mode))
	(cond
	 ((lkk-convert-region
	   (+ (length lkk-insignia0) lkk-conv-beginning) (point) 1 'loose 'seedonly)
	  (setq stage 1))
	 (t (message "Yomi completion failed.") (sit-for 1) nil)
	 ))
       )))
       
    ;; == STAGE 2 - 4: POST PROCESS, if STAGE 1 completed
    ;; Only STAGE 2 is done by our function, 3 & 4 by hooks.
    (cond
     ((and (> stage 0) (sit-for 0 lkk-elettronico-delay2))
      ;; lkk-reconvert or lkk-convert ?
      ;; (lkk-reconvert nil)
      (cond (lkk-conv-beginning (lkk-convert nil)))
      (setq stage 2)))
    (cond
     ((and (> stage 1) (sit-for 0 lkk-elettronico-delay3))
      (prog1 (lkk-run-hooks 'lkk-elettronico-j-hack-hook3)
	(setq stage 3))))
    (cond
     ((and (> stage 2) (sit-for 0 lkk-elettronico-delay4))
      (prog1				; Tries as much completion possible.
	  (while (and (sit-for 0 lkk-elettronico-delay4)
		      (lkk-run-hooks 'lkk-elettronico-j-hack-hook4)) t)
	(setq stage 4))))
    ))

;;
;; Experimental Full auto mode -- incrementally completes hiragana yomi.
;; Not yet so nice.
;; Related functions: lkk-j-input(), lkk-convert-region().
;;

(defun lkk-elettronico (arg)
  "Toggle lkk-full-auto mode. See below for ARG.

In `lkk-full-auto' mode, the full-auto `yomi' completion is started on
yomi- hiragana\'s exceeding threshold `lkk-full-auto-hchar-threshold'
\(number of hiragana characters\) after conversion starting point is
set.

Explicit numeric argument sets the threshold to the value. For example:

\\[lkk-elettronico]     : Toggle lkk-full-auto \(current threshold kept\)
C-u \\[lkk-elettronico] : Force lkk-full-auto \(current threshold kept\)
M-2 \\[lkk-elettronico] : Turn on lkk-full-auto, and set threshold to 2

Negative value, or zero as the explicit numeric argument is accepted as well.
See also docstring of `lkk-full-auto-hchar-threshold'.
Elettronico in lkk-l-mode can be turned off by
 `lkk-inhibit-elettronico-in-l-mode'."
  (interactive "P")
  (setq lkk-full-auto (or arg (null lkk-full-auto))) ; <=t , or ttoggle 
  (force-mode-line-update)

  (if (and arg (null (listp arg)))	; case explicit numeral
      (setq lkk-full-auto-hchar-threshold 
	    ;; Allow 0, and negative as well.
	    (prefix-numeric-value arg)
	    ))

  (message "lkk-elettronico turned %s"
	   (if lkk-full-auto
	       (format "on with threshold: J-mode:%d l-mode:%s"
		       lkk-full-auto-hchar-threshold
		       (if lkk-inhibit-elettronico-in-l-mode
			   "disabled"
			 (+ lkk-full-auto-hchar-threshold
			    lkk-elettronico-/-conv-salt)))
	     "off")))

(fset 'lkk-toggle-full-auto 'lkk-elettronico)

(defun lkk-elettronico-exit ()
  "Force turn off il elettronico. Mostly for lkk-all-minor-mode-disable()."
  (setq lkk-full-auto nil)
  (force-mode-line-update))

;;
;; lkk-convert-okuri-crank() -- inline called (only) from lkk-j-input
;;

(defsubst lkk-convert-okuri-crank ()
  "Start conversion on okurigana input is completed.
This inline should be called only from `lkk-j-input', and that is the
only righteous thing in lkk."
  (cond
   ((and (null lkk-j-beginning)
	 ;;lkk-okuri lkk-okuri-beginning lkk-conv-beginning
	 (lkk-okuri-converting-p 'reset-vars-on-failure)) ; returns nil if fail
    (setq lkk-okuri-root (buffer-substring-no-properties
			  (+ lkk-conv-beginning (length lkk-insignia0))
			  lkk-okuri-beginning)
	  lkk-okuri-h (buffer-substring-no-properties
		       (+ lkk-okuri-beginning (length lkk-insignia2))
		       ;; If `lkk-inhibit-auto-okuri-conv' is implemented,
		       ;; this point should be stored in a lkk-marker, and
		       ;; that makes things too complicated.
		       (point)
		       ))
    (lkk-convert-okuri))))


;;; Deprecated, because this makes the engine too complicated:
;;; Wee need too implement one more `lkk-okuri-end' sorta lkk-marker,
;;; and that ruins the simplicity of lkk.  Or, just am I toooo lasy ?
;;; (defvar lkk-inhibit-auto-okuri-conv nil
;;;  "Non-nil inhibits conversion on okurigana input.
;;; The conversion is postponed until explicit `lkk-convert' call.")

;;
;; r2k input
;;
;; `C-u 4 ccha' -- do not generate "ääää" like skk, but
;; this just suppresses "äää".  More decent one is in ./experimental/*.el
;;
;; To be simple, clean, and neat:
;;  NEVER use read-char() / read-event().
;;  NEVER push anything back to unread-command-events.
;;  NEVER keep string in global variables -- to assure user's freedom.
;;  NEVER rely on counts of chars/strings if they are not ASCII.
;;    -- the result depends on emacs/mules version
;;

(defun lkk-j-input (arg)
  "Simple Japanese input automaton -- converts romaji sequence to kana."
  (interactive "*p")
  (let ((roma) (kana)
	(key last-command-char)
	(start (point-marker))
	(echo-keystrokes 0))

    (if (or (and lkk-j-beginning
	     (null (lkk-same-line-p lkk-j-beginning (point))))
	    ;; skk-kakutei-early == t equivalent operation.
	    ;; You can't customize this default yet (or forever ?).
	    ;; Further, lkk-full-auto and lkk-J-mode combination should be
	    ;; checked.  Lousy ?
	    (and lkk-J-mode (null lkk-full-auto))
	    (and lkk-J-mode
		 lkk-full-auto
		 (integerp lkk-kanji-index) ; nil -- elettronico compl failed ?
		 (> lkk-kanji-index 0) ; converted string inserted already.
		 ))
	(lkk-sit-back))

    ;; Debugger
    (if lkk-debug-hook (run-hooks 'lkk-debug-hook))

    (if (eq last-command 'lkk-j-input) (lkk-cancel-undo-boundary-subr))

    (or lkk-j-beginning (setq lkk-j-beginning start
			      ;; (copy-marker start) ;; seems okay with ptr.
			      lkk-j-numarg arg))

    ;; This should be symmetric with
    (let ((buffer-undo-list t))
      (insert key)		; Show what user entered anyway.
      )

    (setq roma (buffer-substring-no-properties lkk-j-beginning (point)))

    (cond

     ;; Exceptional ASCII direct input mode after `/' is hit.
     ( ;;; lkk-/-converting ;; already inserted one, so check if gt 1.
      lkk-/-mode
      (while (> arg 1) (insert key) (setq arg (1- arg)))
      (lkk-clear-marker 'lkk-j-beginning)
      (setq ;; lkk-j-beginning nil
	    lkk-j-numarg 1))

     ;; Vowel -- one char sequence. Region start -- (point) contains the `key'.
     ((and (equal lkk-j-beginning start)
	   (setq kana (lkk-get-kana0 roma lkk-r2k-alist-V)))
      (lkk-replace-buffer-substring lkk-j-beginning (point)
				    kana lkk-j-numarg 'noundo)
      (lkk-clear-marker 'lkk-j-beginning)
      (setq ;; lkk-j-beginning nil
	    lkk-j-numarg 1))

     ;; Else consonant hereunder -- already kana sequence is started.

     ;; Try simple romaji sequence matching in lkk-r2k-alist(-vector).
     ((and (lkk-r2k-terminator-p (substring roma -1))
	   (setq kana (lkk-get-kana roma)))
      (lkk-replace-buffer-substring lkk-j-beginning (point)
				    kana lkk-j-numarg 'noundo)
      (lkk-clear-marker 'lkk-j-beginning)
      (setq ;; lkk-j-beginning nil
	    lkk-j-numarg 1))

     ;; Check done only if 2 or more roman chars are there.
     ((>= (- (point) lkk-j-beginning) 2)
      ;; Else invalid or still-partial sequence hereunder.
      ;; lkk-r2k-alist is not yet matched hereunder.
      ;; "" and "" type of exception is checked first.

      (cond
       (;; (and (>= (- (point) lkk-j-beginning) 2)
	(string-equal (substring roma 0 1) (substring roma 1 2));; )
	(lkk-replace-buffer-substring lkk-j-beginning (point)
				      (if lkk-q-mode "" "") nil 'noundo)
	(lkk-clear-marker 'lkk-j-beginning);; added to make the marker void
	(setq lkk-j-beginning (point-marker) lkk-j-numarg 1) ; ignore arg
	(insert (substring roma 1)))

       ;; "nn", "mm" etc. should have been matched already.
       ((and ; (>= (- (point) lkk-j-beginning) 2)
	 (or (string-equal "n" (substring roma 0 1))
	     (string-equal "m" (substring roma 0 1)))
	 (not (lkk-valid-roma-p roma)))
	(lkk-replace-buffer-substring lkk-j-beginning (point)
				      (if lkk-q-mode "" "" ) nil 'noundo)
	(lkk-clear-marker 'lkk-j-beginning) ; added to make the marker void
	(setq lkk-j-beginning (point-marker))
	;;(setq lkk-j-numarg 1)		; ignore arg
	(insert (substring roma 1)))

       ;; Reset "sd" "cccch" sort of invalid sequence as in skk.
       ;; This test must be done at last.
       ((not (lkk-valid-roma-p roma))
	;; Erase including `key' inserted.
	;; (delete-region lkk-j-beginning (point))
	(lkk-replace-buffer-substring lkk-j-beginning (point)
				      nil nil 'noundo)
	;; Fake as if we have just started sequence.
	(lkk-clear-marker 'lkk-j-beginning);; added to make the marker void
	(setq lkk-j-beginning (point-marker) lkk-j-numarg 1)
	(insert key)))

      )  ;; (>= (- (point) lkk-j-beginning) 2)
       
     ) ;;;; The great cond ends here.

    ;; Intended to inhibit the marker to be updated before exiting
    ;; this function, but the marker entity should be seen thru
    ;; lkk-j-beginning after exit.  This code will be needed if
    ;; copy-marker() is used on copying to lkk-j-beginning, though.
    ;; (set-marker start nil)

    ;;
    ;; Well well well, Kana may be ready here, guys.
    ;;

;;; Exceptional. auto-okuri-ari process only supported in lkk
;;;    (or lkk-inhibit-auto-okuri-conv  ; This will never be implemented.
    (lkk-convert-okuri-crank)		; May fail, but it's allright for now.
;;;	)

    ;; Auto-fill-support, do not call handler if kana is not complete.
    (cond ((and lkk-do-auto-fill-in-j-input
		(null lkk-conv-beginning)
		(null lkk-j-beginning))
	   (lkk-do-auto-fill)))

    ;; ELEKITERU
    (or lkk-j-beginning			; Do not complete if kana is not there
	(if lkk-full-auto (lkk-elettronico-j-hack)))

    (if lkk-j-input-hook (run-hooks 'lkk-j-input-hook))
    ))

(defvar lkk-j-input-hook nil		; Currently Mostly for debug.
  "Hook run at the end of `lkk-j-input'.")

;;
;; Load dictionary accessing method
;;

(defsubst lkk-require-or-load (fun feature)
  (let ((feature (if (eq fun 'load) (symbol-name feature) feature)))
    (funcall fun feature)))

(defun lkk-select-right-dict-method (&optional arg)
  "Select dictionary accessing method dynamically according to lkk-dict-method.
This function is assumed to be called after modifying variable lkk-dict-method.
Optional ARG forces to destroy dictionary data in core and to make a required
setup for the new method. You need more interactive version ?"
  (interactive "P")
  (let ((fun (if arg 'load 'require))
	(kill-dict-buf
	 (lambda ()
	   (and (get-buffer lkk-dict-buffer)
		(kill-buffer lkk-dict-buffer)
		(setq lkk-dict-type nil
		      lkk-dict-version nil))))
	(kill-dict-alist
	 (lambda () (if (or lkk-big-dict lkk-big-dict-okuri)
			(setq lkk-big-dict nil lkk-big-dict-okuri nil)))) )

    (cond (arg				; Destroy all dict in core / buffer
	   (funcall kill-dict-alist)
	   (funcall kill-dict-buf)))

    (cond ((or (eq lkk-dict-method 'lkk-skk-dict-buffer)
	       (eq lkk-dict-method 'lkk-dict-buffer))
	   (funcall kill-dict-alist)
	   (if (eq lkk-dict-method 'lkk-skk-dict-buffer)
	       (lkk-require-or-load fun 'lkk-skk-dichead)
	     (lkk-require-or-load fun 'lkk-lkk-dichead)))

	  ((eq lkk-dict-method 'lkk-dict-alist)
	   (funcall kill-dict-buf)
	   (lkk-require-or-load fun 'lkk-lkk-dichead))
	  (t (error "lkk-dict-method value is wrong: %s" lkk-dict-method)))))

;; Select here for the first time.
(lkk-select-right-dict-method nil)

;;
;; lkk-suicide () -- Hooking on kill-emacs
;;
;; The defadvice idea is from skk-10.x as well. Thanks.
;;

(defvar lkk-suicide-hook nil "Hook run on preceding kill-emacs.")

(defun lkk-suicide (&optional arg)
  (interactive "P")
  (cond ((process-status "lkk-skkservd")
	 (process-send-string "lkk-skkservd" "0")
	 (kill-buffer " *.lkk-skkserv*")
	 ))
  ;; Auto-dict-save not yet, but do last save if timer-save is going on.
  (cond (lkk-dump-dict-timer-obj
	 (lkk-dump-dict)		; Do the last one.
	 (lkk-cancel-all-dump-dict-with-timer)
	 ))
  (run-hooks 'lkk-suicide-hook) ; dump-dict etc. can be done.
  )

;; My lkk.el has (require 'lkk-skkserv); so this works.
;;(add-hook 'kill-emacs-hook 'lkk-suicide)
(defadvice save-buffers-kill-emacs (before lkk-ad activate)
  (lkk-suicide))

;; Run load-hook
(run-hooks 'lkk-load-hook)
